/* $NetBSD$ */

/*
 * File "cd_disect.c" is part of the UDFclient toolkit.
 * File $Id: cd_disect.c,v 1.33 2005/01/11 15:47:38 imago Exp $ $Name:  $
 *
 * Copyright (c) 2003, 2004 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udftoolkit/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <assert.h>
#include <sys/types.h>
#include <inttypes.h>

#include "scsi_sense.h"
#include "scsilib.h"


/* globals */
char *profile_name[0x30], *write_types[16], *data_block_types[16], *session_formats[0x100];
char *inactivities[16];
struct udf_discinfo disc;


/* helper functions; these could be shared away */
int read_cd_hex2(int val) {
	int nl, nh;

	nl = val & 15;
	nh = val >> 4;
	if (nl >= 'A') nl -= 'A' + 10;
	if (nh >= 'A') nh -= 'A' + 10;

	return (nh*16) + nl;
}


int read_cd_bcd(int val) {
	int nl, nh;

	nl = (val & 15) - '0';
	nh = (val >> 4) - '0';
	if ((nl < 0 || nl > 9) || (nh < 0 || nh > 9)) return val;

	return nh*10 + nl;
}


int cd_msf2lba(int m, int s, int f) {
	return 4500*m + 75*s + f - 150;
}


void init_tables(void) {
	profile_name[0x00]	= "Unknown[0] profile";
	profile_name[0x01]	= "Non removeable disc";
	profile_name[0x02]	= "Removable disc";
	profile_name[0x03]	= "Magneto Optical with sector erase";
	profile_name[0x04]	= "Magneto Optical write once";
	profile_name[0x05]	= "Advance Storage Magneto Optical";
	profile_name[0x08]	= "CD-ROM";
	profile_name[0x09]	= "CD-R recordable";
	profile_name[0x0a]	= "CD-RW rewritable";
	profile_name[0x10]	= "DVD-ROM";
	profile_name[0x11]	= "DVD-R sequential";
	profile_name[0x12]	= "DVD-RAM rewritable";
	profile_name[0x13]	= "DVD-RW restricted overwrite";
	profile_name[0x14]	= "DVD-RW sequential";
	profile_name[0x1a]	= "DVD+RW rewritable";
	profile_name[0x1b]	= "DVD+R recordable";
	profile_name[0x20]	= "DDCD readonly";
	profile_name[0x21]	= "DDCD-R recodable";
	profile_name[0x22]	= "DDCD-RW rewritable";

	write_types[0x00]	= "Packet/Incremental";
	write_types[0x01]	= "Track-at-once";
	write_types[0x02]	= "Session-at-one";
	write_types[0x03]	= "Raw";

	data_block_types[ 0]	= "raw data 2352 bytes";
	data_block_types[ 1]	= "raw data 2368 bytes with P and Q channel";
	data_block_types[ 2]	= "raw data 2352 bytes (+96) P-W subchannel appended";
	data_block_types[ 3]	= "raw data 2352 bytes (+96) raw P-W subchannel appended";
	data_block_types[ 8]	= "ISO mode 1 with 2048 bytes data";
	data_block_types[ 9]	= "ISO mode 2 with 2336 bytes data, formless";
	data_block_types[10]	= "ISO mode 2 with 2048 bytes data (CDROM-XA, form 1), subheader from write parameters";
	data_block_types[11]	= "ISO mode 2 with 2048 bytes data (CDROM-XA, form 1), 8 bytes for subheader first";
	data_block_types[12]	= "ISO mode 2 with 2324 bytes data (CDROM-XA, form 2), subheader from write parameters";
	data_block_types[13]	= "ISO mode 2 with 2332 bytes data (CDROM-XA, form 1 or 2 or mixed), 8 bytes for subheader first";

	session_formats[0x00] = "CD-DA, CD-ROM or other data discs";
	session_formats[0x10] = "CD-I disc";
	session_formats[0x20] = "CD-ROM XA disc or DDCD disc";

	inactivities[0x0]	= "Vendor specific";
	inactivities[0x1]	= "125 ms";
	inactivities[0x2]	= "250 ms";
	inactivities[0x3]	= "500 ms";
	inactivities[0x4]	= "1 sec";
	inactivities[0x5]	= "2 sec";
	inactivities[0x6]	= "4 sec";
	inactivities[0x7]	= "8 sec";
	inactivities[0x8]	= "16 sec";
	inactivities[0x9]	= "32 sec";
	inactivities[0xa]	= "1 min";
	inactivities[0xb]	= "2 min";
	inactivities[0xc]	= "4 min";
	inactivities[0xd]	= "8 min";
	inactivities[0xe]	= "16 min";
	inactivities[0xf]	= "32 min";

}


int scsi_call(int flags, struct udf_discinfo *disc, void *cmd, int cmdlen, void *data, int datalen) {
	return scsi_command(disc, cmd, cmdlen, data, datalen, 30000, flags);
}


/* start of discect functions */
void dump_recorded_capacity(void) {
	scsicmd cmd;
	uint8_t buf[36];
	uint32_t lba, blk_len;
	int error;

	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x25;	/* CD READ RECORDED CAPACITY */
	error = scsi_call(SCSI_READCMD, &disc, cmd, 10, buf, 8);
	if (error) fprintf(stderr, "Read recorded capacity SCSI call returned : %s\n", strerror(error));

	lba     = buf[3] | (buf[2]<<8) | (buf[1]<<16) | (buf[0]<<24);
	blk_len = buf[7] | (buf[6]<<8) | (buf[5]<<16) | (buf[4]<<24);
	printf("\nCD recorded capacity is LBA %d (%d blocks), blk_len %d\n",
			lba, lba/4, blk_len);
}


char *get_profile_name(int profile) {
	static char scrap[255];

	if (profile >=0 && profile < 0x30) return profile_name[profile];
	sprintf(scrap, "Reserved profile %d", profile);

	return scrap;
}


void dump_feature(char *fpos) {
	uint32_t	 feature, cnt, profile;
	uint32_t	 feature_len, feature_ver, feature_cur, feature_pers;
	uint32_t	 interface, blocking, log_blk_size, last_log_blk_addr;
	uint32_t	 datablock_types, num_link_sizes, cuesheet_len;
	uint32_t	 num_vol_levels, dcb_entry;
	uint32_t	 century, year, day, month, hour, minute;
	uint8_t		*pos;

	feature     =  fpos[1] | (fpos[0] << 8);
	feature_ver = (fpos[2] >> 2) & 15;
	feature_cur = (fpos[2] & 1);
	feature_pers= (fpos[2] & 2);
	feature_len =  fpos[3];
	printf("\t\tFeature  0x%04x (%2d bytes) version %2d; persistent %s; currently active %s\n",
			feature, feature_len, feature_ver, feature_pers?"yes":" no", feature_cur?"yes":" no");

	pos = &fpos[4];
	switch (feature) {
		case 0x0000:
			printf("\t\tProfile list; supporting profiles\n");
			for (cnt=0; cnt < feature_len; cnt += 4) {
				profile = pos[1] | (pos[0] << 8);
				printf("\t\t\t%s %s\n", pos[2] & 1 ?"ACTIVE  ":"inactive", get_profile_name(profile));
				pos += 4;
			};
			break;
		case 0x0001:
			interface = pos[3] | (pos[2] << 8) | (pos[1] << 16) | (pos[0] << 24);
			printf("\t\tCore features : physical interface standard commands for `");
			switch (interface) {
				case 0      : printf("Unspecified"); break;
				case 1      : printf("SCSI family"); break;
				case 2      : printf("ATAPI"); break;
				case 3      : printf("IEEE 1394 - 1995"); break;
				case 4      : printf("IEEE 1394A"); break;
				case 5      : printf("Fibre channel"); break;
				case 0xffff : printf("Vendor specific"); break;
				default     : printf("<reserved (0x%04x)>", interface); break;
			};
			printf("'\n");
			break;
		case 0x0002:
			printf("\t\tMorphing command set; %ssupport for ASYNC\n", (pos[0] & 1)?"":"no ");
			break;
		case 0x0003:
			printf("\t\tRemovable medium features\n");
			if (feature_ver > 1) {
				printf("\t\t\tUnknown flags\n");
				break;
			};
			if (pos[0] & 1) printf("\t\t\tDevice can be locked against removal\n");
			printf("\t\t\tDevice will go into the %slocked state by default\n", (pos[0] & 4)?"":"un-");
			printf("\t\t\tDevice %s eject the media with START/STOP with eject bit\n", (pos[0] & 8)?"can":"can't");
			printf("\t\t\tLoading mechanism : ");
			switch (pos[0] >> 5) {
				case 0  : printf("Caddy");  break;
				case 1  : printf("Tray");   break;
				case 2  : printf("Pop-up"); break;
				case 4  : printf("Embedded changer with seperate discs"); break;
				case 5  : printf("Embedded changes using magazines"); break;
				default : printf("Unknown/Reserved");
			};
			printf("\n");
			break;
		case 0x0004:
			printf("\t\tWrite protect features :");
			if (pos[0] & 2) printf(" SPWP");
			if (pos[0] & 1) printf(" SSWPP");
			printf("\n");
			break;
		case 0x0010:
			printf("\t\tRandom readable\n");
			log_blk_size = pos[3] | (pos[2] << 8) | (pos[1] << 16) | (pos[0] << 24);
			blocking     = pos[5] | (pos[4] << 8);
			printf("\t\t\tLogical block size %u bytes\n", log_blk_size);
			printf("\t\t\tDevice blocking number %d logical blocks\n", blocking);
			if (pos[6] & 1) printf("\t\t\tHas RW error recovery mode page\n");
			break;
		case 0x001d:
			printf("\t\tMulti-read; The Logical Unit can read all CD media types; based on OSTA MultiRead\n");
			break;
		case 0x001e:
			printf("\t\tAbility to read CD specific structures\n");
			printf("\t\t\tDevice does %ssupport C2 error pointers\n", (pos[0] & 2)?"":"NOT ");
			printf("\t\t\tDevice does %ssupport CD-Text\n", (pos[0] & 1)?"":"NOT ");
			break;
		case 0x001f:
			printf("\t\tAbility to read DVD specific structures\n");
			break;
		case 0x0020:
			printf("\t\tRandom writable\n");
			last_log_blk_addr = pos[3] | (pos[2] << 8) | (pos[1] << 16) | (pos[0] << 24);
			log_blk_size      = pos[7] | (pos[6] << 8) | (pos[5] << 16) | (pos[4] << 24);
			blocking          = pos[9] | (pos[8] << 8);
			printf("\t\t\tLast writable logic block address is %u ", last_log_blk_addr);
			printf("(approx %u Mb)\n", (uint32_t) ((uint64_t) last_log_blk_addr*log_blk_size/(1024*1024)));
			printf("\t\t\tLogical block size %u bytes\n", log_blk_size);
			printf("\t\t\tDevice blocking number %d logical blocks\n", blocking);
			if (pos[10] & 1) printf("\t\t\tHas RW error recovery mode page\n");
			break;
		case 0x0021:
			printf("\t\tIncremental streaming writable\n");
			datablock_types = pos[1] | (pos[0] << 8);
			num_link_sizes = pos[3];
			printf("\t\t\tDevice is %scapable of `Zero loss linking'\n", (pos[2] & 1)?"":"NOT ");
			printf("\t\t\tDevice supported data types (bitfield) 0x%04x\n", datablock_types);
			printf("\t\t\tDevice supports %d link size%s :", num_link_sizes, (num_link_sizes>1)?"s":"");
			for (cnt=0; cnt < num_link_sizes; cnt++) {
				printf(" %d", pos[4+cnt]);
			};
			printf("\n");
			break;
		case 0x0022:
			printf("\t\tCan erase/support for erasing media (OBSOLETE)\n");
			break;
		case 0x0023:
			printf("\t\tSupport for formatting media\n");
			break;
		case 0x0024:
			printf("\t\tHas defect handling; i.e. apparently defect-free space\n");
			printf("\t\t\tDevice does %ssupport read `Space Area Information' (DVD)\n", (pos[0] & 128)?"":"NOT ");
			break;
		case 0x0025:
			printf("\t\tSupport for writing any unrecorded logical block on write once media in random order\n");
			log_blk_size = pos[3] | (pos[2] << 8) | (pos[1] << 16) | (pos[0] << 24);
			blocking     = pos[5] | (pos[4] << 8);
			printf("\t\t\tLogical block size %u bytes\n", log_blk_size);
			printf("\t\t\tDevice blocking number %d logical blocks\n", blocking);
			if (pos[6] & 1) printf("\t\t\tHas RW error recovery mode page\n");
			break;
		case 0x0026:
			printf("\t\tSupport for restricted overwrite; i.e. on blocking boundaries only\n");
			break;
		case 0x0027:
			printf("\t\tCD-RW CAV Write; The ability to write high speed CD-RW media\n");
			/* parameters all reserved */
			break;
		case 0x0028:
			printf("\t\tMRW formatted media support\n");
			printf("\t\t\tDevice %swrite/format media in MRW but can read it\n", (pos[0] & 1)?"can ":"can't");
			break;
		case 0x002a:
			printf("\t\tDVD+RW media reading/writing support\n");
			if (pos[0] & 1) {
				printf("\t\t\tDevice can write and background format DVD+RW media\n");
				if (pos[1] & 1) printf("\t\t\tDevice only supports read compatibility format stop\n");
			} else {
				printf("\t\t\tDevice can only recognise and read DVD+RW\n");
			};
			break;
		case 0x002b:
			printf("\t\tThe ability to read/write DVD+R recorded media\n");
			if (pos[0] & 1) {
				printf("\t\t\tDevice can write DVD+R media\n");
			} else {
				printf("\t\t\tDevice can only recognise and read DVD+R\n");
			};
			break;
		case 0x002c:
			printf("\t\tSupport for rigid restricted overwrite only (CD-RW)\n");
			if (pos[0] & 8) printf("\t\t\tDevice can generate direct status data during formatting\n");
			if (pos[0] & 4) printf("\t\t\tDevice can read the defect status data recorded on media\n");
			if (pos[0] & 2) printf("\t\t\tDevice allows writing on immediate state sessions and quick formatting\n");
			printf("\t\t\tDevice does %ssupport blanking of media\n", (pos[0] & 1)?"":"NOT ");
			break;
		case 0x002d:
			printf("\t\tTrack at once recording support\n");
			datablock_types = pos[3] | (pos[2] << 8);
			if (pos[0] & 64) printf("\t\t\tDevice is capable of zero-loss linking\n");
			if (pos[0] & 16) printf("\t\t\tDevice is capable of writing R-W Sub code in RAW mode\n");
			if (pos[0] &  8) printf("\t\t\tDevice is capable of writing R-W Sub code in Packet mode\n");
			printf("\t\t\tDevice does %ssupport test writes (i.e. laser off)\n", (pos[0] & 4)?"": "NOT ");
			if (pos[0] &  2) printf("\t\t\tDevice supports overwriting a track using track at once\n");
			if (pos[0] &  1) printf("\t\t\tDevice supports writing R-W Sub channels with user data\n");
			printf("\t\t\tTrack at once data types supported (bitfield) 0x%04x\n", datablock_types);
			break;
		case 0x002e:
			printf("\t\tSession at once or RAW writing support; CD Mastering\n");
			cuesheet_len = pos[3] | (pos[2] << 8) | (pos[1] << 16);
			if (pos[0] & 64) printf("\t\t\tDevice is capable of zero-loss linking\n");
			printf("\t\t\tDevice can %swrite using the Session at Once write type\n", (pos[0] & 32)?"":"CANT ");
			if (pos[0] & 16) printf("\t\t\tDevice is capable of writing multi-session in RAW mode\n");
			printf("\t\t\tDevice can %swrite using the raw write type\n", (pos[0] & 8)?"":"CANT ");
			printf("\t\t\tDevice does %ssupport test writes (i.e. laser off)\n", (pos[0] & 4)?"": "NOT ");
			if (pos[0] &  2) printf("\t\t\tDevice supports overwriting previous recorded media\n");
			if (pos[0] &  1) printf("\t\t\tDevice supports writing R-W Sub channels with user data\n");
			printf("\t\t\tMaximum cue sheet length %d bytes\n", cuesheet_len);
			break;
		case 0x002f:
			printf("\t\tDVD-R/-RW Write; The ability to write DVD specific structures\n");
			if (pos[0] & 64) printf("\t\t\tDevice is capable of zero-loss linking\n");
			printf("\t\t\tDevice does %ssupport test writes (i.e. laser off)\n", (pos[0] & 4)?"": "NOT ");
			if (pos[0] &  2) printf("\t\t\tDevice supports overwriting previous recorded media\n");
			break;
		case 0x0030:
			printf("\t\tCan read DDCD user data\n");
			break;
		case 0x0031:
			printf("\t\tCan write and read DDCD-R media\n");
			printf("\t\t\tDevice does %ssupport test writes (i.e. laser off)\n", (pos[0] & 4)?"": "NOT ");
			break;
		case 0x0032:
			printf("\t\tCan write and read DDCD-RW media\n");
			if (pos[0] & 2) printf("\t\t\tDevice allows writing on immediate state sessions and quick formatting\n");
			printf("\t\t\tDevice does %ssupport blanking of media\n", (pos[0] & 1)?"":"NOT ");
			break;
		case 0x0100:
			printf("\t\tPower management support\n");
			break;
		case 0x0101:
			printf("\t\tS.M.A.R.T. (Self-Monitoring Analysis and Reporting Technology support)\n");
			printf("\t\t\tDevice does %sprovide Fault/Failure Reporting Mode Page\n", (pos[0] & 1)?"":"NOT ");
			break;
		case 0x0102:
			printf("\t\tEmbedded changer feature\n");
			printf("\t\t\tDevice is %scapable of switching media sides\n", (pos[0] & 16)?"":"NOT ");
			printf("\t\t\tDevice can %sreport disc presence after a reset of magazine change\n", (pos[0] & 4)?"":"NOT ");
			break;
		case 0x0103:
			printf("\t\tAbility to play audio CDs via the Logical Unit's own analog output\n");
			num_vol_levels = pos[3] | (pos[2] << 8);
			printf("\t\t\tDevice is does %ssupport the SCAN command\n", (pos[0] & 4)?"":"NOT ");
			printf("\t\t\tDevice %s mute seperate channels\n", (pos[0] & 4)?"can":"can't");
			printf("\t\t\tDevice %s set volumes for each channel seperately\n", (pos[0] & 4)?"can":"can't");
			printf("\t\t\tDevice has %d audio volume levels\n", num_vol_levels);
			break;
		case 0x0104:
			printf("\t\tAbility for the device to accept new microcode via the interface\n");
			break;
		case 0x0105:
			printf("\t\tAbility to respond to all commands within a specific time\n");
			break;
		case 0x0106:
			printf("\t\tAbility to perform DVD CSS/CPPM authentication and RPC\n");
			printf("\t\t\tSupporting CCS version %d\n", pos[3]);
			break;
		case 0x0107:
			printf("\t\tAbility to read and write using Initiator requested performance parameters; realtime streaming\n");
			printf("\t\t\tDevice does %shave the `read buffer capacity' command \n", (pos[0] & 16)?"":"NOT ");
			printf("\t\t\tDevice CD speed can %sbe set up\n", (pos[0] & 8)?"":"NOT ");
			printf("\t\t\tDevice write speed performance can %sbe queried\n", (pos[0] & 4)?"":"NOT ");
			printf("\t\t\tDevice does %shave a `set streaming' command\n", (pos[0] & 2)?"":"NOT ");
			printf("\t\t\tDevice does %ssupport stream recording operation\n", (pos[0] & 1)?"":"NOT ");
			break;
		case 0x0108:
			printf("\t\tThis device has an unique serialnumber : \"");
			for (cnt=0; cnt < feature_len; cnt++) {
				printf("%c", pos[cnt]);
			};
			printf("\"\n");
			break;
		case 0x010a:
			printf("\t\tThe ability to read and/or write Disc Control Blocks (DVD only)\n");
			for (cnt=0; cnt < feature_len; cnt+=4, pos+=4) {
				dcb_entry = pos[3] | (pos[2] << 8) | (pos[1] << 16);
				printf("\t\t\t\tSupported Disc Control Blocks content descriptor %d\n", dcb_entry);
			};
			printf("\n");
			break;
		case 0x010b:
			printf("\t\tThe Logical Unit supports DVD CPRM authentication\n");
			printf("\t\t\tSupporting CPRM version %d\n", pos[3]);
			break;
		case 0x01ff:
			printf("\t\tFirmware creation date report :\n");
			century = pos[ 1] | (pos[ 0] << 8);
			year	= pos[ 3] | (pos[ 2] << 8);
			month	= pos[ 5] | (pos[ 4] << 8);
			day	= pos[ 7] | (pos[ 6] << 8);
			hour	= pos[ 9] | (pos[ 8] << 8);
			minute	= pos[11] | (pos[10] << 8);
			printf("\t\t\tFirmware version date %d-%d-%2d%2d at time %02d:%02d\n", day, month, century, year, hour, minute);
			break;
		default:
			break;
	};
	printf("\n");
}


void dump_drive_configuration(int feature_current) {
	scsicmd		cmd;
	uint8_t		features[1024+100], *fpos;
	uint32_t	features_len, feat_tbl_len;
	uint32_t	current_profile, feature, pos;
	uint32_t	feature_cur, feature_len;
	int		error;

	feat_tbl_len = 1024;

	printf("Getting drive configuration\n");
	feature = 0;

	do {
		bzero(cmd, SCSI_CMD_LEN);
		cmd[0] = 0x46;			/* Get configuration */
		cmd[1] = 0;			/* RT=0 -> all independent of current setting */
		cmd[2] = (feature) >> 8;	/* MSB feature number */
		cmd[3] = (feature) & 0xff;	/* LSB feature number */
		cmd[7] = (feat_tbl_len) >> 8;	/* MSB buffersize */
		cmd[8] = (feat_tbl_len) & 0xff;	/* LSB buffersize */
		cmd[9] = 0;			/* control */
		error = scsi_call(SCSI_READCMD, &disc, cmd, 10, features, feat_tbl_len+8);
		if (error) {
			fprintf(stderr, "While reading feature table : %s\n", strerror(error));
			return;
		};

		features_len    = features[3] | (features[2]<<8) | (features[1]<<16) | (features[0]<<24);
		current_profile = features[7] | (features[6]<<8);

		printf("\tFeatures table length %d bytes\n", features_len);
		printf("\tCurrent profile       0x%04x `%s'\n", current_profile, get_profile_name(current_profile));
		pos = 8;
		while (pos < features_len) {
			fpos = &features[pos];
			feature_cur = (fpos[2] & 1);
			feature_len =  fpos[3];
			if (feature_cur == feature_current) dump_feature(fpos);

			assert((feature_len & 3) == 0);
			pos += 4 + feature_len;
		};
	} while (features_len >= 0xffff);
}


void dump_disc_information(void) {
	char		*disc_status_txt[]  = {"empty disc", "incomplete (appendable)", "full (not appendable)", "random writable"};
	char		*last_session_txt[] = {"empty", "incomplete", "reserved/damaged", "complete/closed disc"};

	scsicmd		 cmd;
	uint8_t		 di[10000], ti[10000];
	uint32_t	 di_len = 10000, ti_len = 10000;

	int		 disc_status, last_session_state, eraseable;
	int		 num_sessions;
	int		 first_track, first_track_last_session, last_track_last_session;

	int		 data_length, is_track_number, is_session_number, is_track_mode, is_data_mode;
	int		 is_copy, is_damage, is_fixed_packet, is_packet_or_inc, is_blank, is_reserved;
	int		 nwa_valid, lra_valid;
	int		 track_start, next_writable_addr, free_blocks, packet_size, track_size, last_recorded_addr;

	int		 track;

	int		 error;

	bzero(di, di_len);

	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x51;				/* Read disc information */
	cmd[7] = di_len >> 8;			/* MSB allocation length */
	cmd[8] = di_len & 0xff;			/* LSB allocation length */
	cmd[9] = 0;				/* control */
	error = scsi_call(SCSI_READCMD, &disc, cmd, 10, di, di_len);
	if (error) {
		fprintf(stderr, "While reading disc information : %s\n", strerror(error));
		return;
	}

	di_len = di[0] | (di[1]<<8);
	printf("\tDrive returned %d bytes of data\n", di_len);

	disc_status		 =  di[2]       &  3;
	last_session_state	 = (di[2] >> 2) &  3;
	eraseable		 =  di[2]       & 16;
	num_sessions		 =  di[4] | (di[ 9] << 8);
	first_track		 =  di[3];
	first_track_last_session =  di[5] | (di[10] << 8);
	last_track_last_session  =  di[6] | (di[11] << 8);

	printf("\tDisc status           : %s\n", disc_status_txt[disc_status]);
	printf("\tKind of disc          : %s\n", eraseable ? "eraseable (rewritable disc)": "NOT eraseable (recordable disc)");
	printf("\tFirst track number    : %d\n", first_track);
	printf("\tNumber of sessions    : %d\n", num_sessions);
	printf("\tLast session information :\n");
	printf("\t\tState : %s\n", last_session_txt[last_session_state]);
	printf("\t\tFrom track %d to track %d (including hidden)\n", first_track_last_session, last_track_last_session);
	printf("\tDisc is %sin 'Ristricted Use Mode'\n", (di[7] & 32) ? "": "NOT ");
	switch (di[7] & 3) {
		case 0 : printf("\tBackground formatting not applicable"); break;
		case 1 : printf("\tDisc formatting was interrupted and not running now"); break;
		case 2 : printf("\tDisc is being formatted in the background"); break;
		case 3 : printf("\tDisc is formatted"); break;
	};
	printf("\n");
	printf("\tDisc has %s valid 'Disc ID'\n", (di[7] & 128) ? "a ": "NO ");
	printf("\tDisc has %s valid 'Disc Bar Code'\n", (di[7] & 64) ? "a ": "NO");

	printf("\nReading track and session information; track by track\n");
	for(track = first_track; track <= last_track_last_session; track++) {
		bzero(cmd, SCSI_CMD_LEN);
		bzero(ti, ti_len);
		cmd[0] = 0x52;				/* Read track information */
		cmd[1] = 1;				/* indexed on track       */
		cmd[4] = track >> 8;			/* track number 0-0xff ?  */
		cmd[5] = track & 0xff;
		cmd[7] = ti_len >> 8;
		cmd[8] = ti_len & 0xff;
		cmd[9] = 0;
		error = scsi_call(SCSI_READCMD, &disc, cmd, 10, ti, ti_len);
		if (error) {
			fprintf(stderr, "While reading track info : %s\n", strerror(error));
			break;
		}
		data_length		= ti[1] | (ti[0]  << 8);
		is_track_number		= ti[2] | (ti[32] << 8);
		is_session_number	= ti[3] | (ti[33] << 8);
		is_track_mode		= ti[5] & 15;
		is_copy			= ti[5] & 16;
		is_damage		= ti[5] & 32;
		is_fixed_packet		= ti[6] & 16;
		is_packet_or_inc	= ti[6] & 32;
		is_blank		= ti[6] & 64;
		is_reserved		= ti[6] & 128;
		is_data_mode		= ti[6] & 15;
		nwa_valid		= ti[7] & 1;
		lra_valid		= ti[7] & 2;
		track_start		= ti[11] | (ti[10]<<8) | (ti[ 9]<<16) | (ti[ 8]<<24);
		next_writable_addr	= ti[15] | (ti[14]<<8) | (ti[13]<<16) | (ti[12]<<24);
		free_blocks		= ti[19] | (ti[18]<<8) | (ti[17]<<16) | (ti[16]<<24);
		packet_size		= ti[23] | (ti[22]<<8) | (ti[21]<<16) | (ti[20]<<24);
		track_size		= ti[27] | (ti[26]<<8) | (ti[25]<<16) | (ti[24]<<24);
		last_recorded_addr	= ti[31] | (ti[30]<<8) | (ti[29]<<16) | (ti[28]<<24);

		if (data_length <= 30) {
			/* 8 bits track and session numbers returned; last_recordable is invalid */
			is_track_number   &= 0xff;
			is_session_number &= 0xff;
			last_recorded_addr = 0;
		};

		printf("\tTrack %d of session %d\n", is_track_number, is_session_number);
		if (data_length < 27 && lra_valid) printf("\t\tLast recorded addres valid but not returned\n");
		printf("\t\tStart at       : %d\n", track_start);
		printf("\t\tLength         : %d\n", track_size);
		if (free_blocks)      printf("\t\tFree blocks    : %d\n", free_blocks);
		if (packet_size)      printf("\t\tPacket size    : %d\n", packet_size);
		if (nwa_valid)        printf("\t\tNext writable  : %d\n", next_writable_addr);
		if (lra_valid)        printf("\t\tLast recorded  : %d\n", last_recorded_addr);
		if (is_copy)	      printf("\t\tTrack is a copy\n");
		if (is_reserved)      printf("\t\tReserved or Complete track\n");
		if (is_blank)	      printf("\t\tBlank track\n");
		if (is_packet_or_inc) printf("\t\tPacket mode track\n");
		if (is_fixed_packet)  printf("\t\tFixed packet sizes track\n");
	}
}


void dump_toc_pma_atip_info(void) {
	scsicmd		 cmd;
	uint8_t		 toc[10000];
	int		 toc_len = 10000;
	int		 pos, data_length;
	int		 first_track, last_track, first_session, last_session;
	int		 cur_session, session, cntrl, addr, tno, point;
	int		 min, sec, frame, pmin, psec, pframe, nwa, extent;
	int		 error;

	printf("\n\nTrack information gathering using TOC/PMA/ATIP\n");

	/* just TOC */
	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x43;			/* READ TOC/PMA/ATAP INFO	*/
	cmd[1] = 0;			/* no times please but LBA's	*/
	cmd[2] = 0;			/* format 0; TOC		*/
	cmd[6] = 0xAA;			/* go for track 0xAA for length	*/
	cmd[7] = toc_len >> 8;
	cmd[8] = toc_len & 0xff;
	cmd[9] = 0;
	error = scsi_call(SCSI_READCMD, &disc, cmd, 10, toc, toc_len);
	if (error) {
		fprintf(stderr, "TOC reading of tracks failed : %s\n", strerror(error));
		return;
	};
	first_track = read_cd_hex2(toc[2]);
	last_track  = read_cd_hex2(toc[3]);
	printf("\tFirst track   : %d\n", first_track);
	printf("\tLast track    : %d\n", last_track);

	cmd[2] = 1;			/* format 1 this time		*/
	cmd[6] = 0;			/* obligatory zero		*/
	error = scsi_call(SCSI_READCMD, &disc, cmd, 10, toc, toc_len);
	if (error) {
		fprintf(stderr, "TOC reading of sessions failed : %s\n", strerror(error));
		return;
	}
	first_session = read_cd_hex2(toc[2]);
	last_session  = read_cd_hex2(toc[3]);
	printf("\tFirst session : %d\n", first_session);
	printf("\tLast session  : %d\n", last_session);

	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x43;			/* READ TOC/PMA/ATAP INFO	*/
	cmd[1] = 0;			/* no times please but LBA's	*/
	cmd[2] = 2;			/* format 2; full TOC		*/
	cmd[6] = first_session;		/* start at first session	*/
	cmd[7] = toc_len >> 8;
	cmd[8] = toc_len & 0xff;
	cmd[9] = 0;
	error = scsi_call(SCSI_READCMD, &disc, cmd, 10, toc, toc_len);
	if (error) {
		fprintf(stderr, "TOC reading of full TOC failed : %s\n", strerror(error));
		return;
	}

	data_length =toc[1] | (toc[0] << 8);
	pos = 4; cur_session = 0;
	while (pos < data_length + 2) {
		session = read_cd_bcd(toc[pos+ 0]);
		cntrl   = toc[pos+1] & 15;
		addr    = toc[pos+1] >> 4;
		tno     = read_cd_bcd(toc[pos+ 2]);
		point   = read_cd_bcd(toc[pos+ 3]);
		min     = read_cd_bcd(toc[pos+ 4]);
		sec     = read_cd_bcd(toc[pos+ 5]);
		frame   = read_cd_bcd(toc[pos+ 6]);
		pmin    = read_cd_bcd(toc[pos+ 8]);
		psec    = read_cd_bcd(toc[pos+ 9]);
		pframe  = read_cd_bcd(toc[pos+10]);
		if (session != cur_session) {
			if (session) printf("\tSession %d\n", session);
			cur_session = session;
		};
		if (tno == 0 && session) {
			if (addr == 1) {
				printf("\t\t");
				switch (point) {
					case 0xa0 :
						printf("First track number  %d, disctype %d\t", pmin, psec);
						break;
					case 0xa1 :
						printf("Last track number   %d\t\t\t", pmin);
						break;
					case 0xa2 :
						printf("Lead out            %02d:%02d.%02d (LBA %8d)", pmin, psec, pframe, cd_msf2lba(pmin, psec, pframe));
						break;
					default :
						printf("Track start         %02d:%02d.%02d (LBA %8d)", pmin, psec, pframe, cd_msf2lba(pmin, psec, pframe));
				};
				if (min || sec || frame) printf(" ATIME (%02d:%02d.%02d)", min, sec, frame);
			} else if (addr == 5) {
				printf("\t\t");
				switch (point) {
					case 0xb0 :
						nwa    = cd_msf2lba( min,  sec,  frame);
						extent = cd_msf2lba(pmin, psec, pframe);
						printf("Next writable at    %02d:%02d.%02d (LBA %8d)", min, sec, frame, nwa);
						printf("\n\t\t");
						printf("Lead out max at     %02d:%02d.%02d (LBA %8d)", pmin, psec, pframe, cd_msf2lba(pmin, psec, pframe));
						// printf("\n\t\t");
						// printf("Disc surface used for %2d %%\t\t", (int) (100.0 * (float) nwa / (float) extent));
						break;
					case 0xc0 :
						printf("Start first lead-in %02d:%02d.%02d (LBA %8d)", pmin, psec, pframe, cd_msf2lba(pmin, psec, pframe));
						printf("\n\t\t");
						printf("Optimum recording power %d\t\t", min);
						break;
					default :
						printf("<undumped %d, %d, %02x pair>\t\t", addr, cntrl, point);
				};
			} else {
				printf("\t\t<undumped %d, %d, %02x pair>\t", addr, cntrl, point);
			}
			printf("\t (");
			if ((cntrl & 12) == 4) {
				printf("data track");
				if (cntrl & 1) printf("; incremental"); else printf("; uninterrupted");
				if (cntrl & 2) printf("; copy prohibited");
			} else {
				printf("non data");
			};
			printf(")\n");
		};
		pos += 11;
	}
}


char *prchange(val) {
	if (val) return "[changable]";
	return          "[fixed]    ";
}


void dump_parameter_page(uint8_t *pos, uint8_t *change) {
	int page_code, page_length;
	int write_type, track_mode, data_block_type, link_size;
	int testwrite, linksize_valid, zerolinking;
	int appl_code, session_format, packet_size, audio_pause;
	int multi_session, fixed_packets, copybit;
	int i, recovery_time_limit, inactiv_time_mult, S_per_MSF, F_per_MSF;
	int idle_timer_value, standby_timer_value, group1_timeouts, group2_timeouts;

	page_code    = pos[0] & 63;
	page_length  = pos[1];

	printf("\tGot page 0x%02x (%spersistent) : %2d bytes for ", page_code, pos[0] & 128? "":"non ", page_length);
	switch (page_code) {
		case 0x01 :
			printf("read/write error recovery page\n");
			printf("\t\tError recovery parameter %s 0x%02x\n", prchange(change[2]), pos[2]);
			printf("\t\tRead retry count         %s %d\n", prchange(change[3]), pos[3]);
			printf("\t\tWrite retry count        %s %d\n", prchange(change[8]), pos[8]);
			recovery_time_limit = pos[11] + (pos[10] << 8);
			printf("\t\tRecovery time limit      %s %d\n", prchange(change[11]), recovery_time_limit);
			printf("\n");
			break;
		case 0x05 :
			printf("write parameter page\n");
			write_type	= pos[2] & 15;
			track_mode	= pos[3] & 15;
			data_block_type	= pos[4] & 15;
			testwrite	= pos[2] & 16;
			linksize_valid	= pos[2] & 32;
			zerolinking	= pos[2] & 64;
			multi_session   = pos[3] >> 6;
			fixed_packets	= pos[3] & 32;
			copybit		= pos[3] & 16;
			link_size	= pos[5];
			appl_code	= pos[7] & 63;
			session_format  = pos[8];
			packet_size	= pos[13] + (pos[12] << 8) + (pos[11] << 16) + (pos[10] << 24);
			audio_pause	= pos[15] + (pos[14] << 8);
			printf("\t\tWrite type          %s %s\n", prchange(change[2] & 15), write_types[write_type]);
			printf("\t\tTrack mode          %s %d (mode 1 Q subchannel control nibble)\n", prchange(change[3] & 15), track_mode);
			printf("\t\tData block type     %s %s\n", prchange(change[4] & 15), data_block_types[data_block_type]);
			printf("\t\tTestwriting         %s %s\n", prchange(change[2] & 16), testwrite ? "On":"Off");
			printf("\t\tLinksize            %s %d sectors (%s)\n", prchange(change[5]), link_size, linksize_valid ? "valid": "not valid (7)");
			printf("\t\tApplication code    %s %d\n", prchange(change[7] & 63), appl_code);
			printf("\t\tSession format      %s %s\n", prchange(change[8]), session_formats[session_format]);
			printf("\t\tUsing fixed packets %s %s\n", prchange(change[3] & 32), fixed_packets ? "On":"Off");
			printf("\t\tCopybit             %s %s\n", prchange(change[3] & 16), copybit ? "On":"Off");
			printf("\t\tMulti session field %s %d\n", prchange(change[3] >> 6), multi_session);
			printf("\t\tPacket size         %s %d\n", prchange(change[13]), packet_size);
			printf("\t\tAudio pause length  %s %d\n", prchange(change[15]), audio_pause);
			printf("\t\tMedia catalog number                 %s 0x", prchange(change[31]));
			for (i=16; i<=31; i++) printf("%02x", pos[i]);
			printf("\n");
			printf("\t\tInitiational standard recording code %s 0x", prchange(change[47]));
			for (i=32; i<=47; i++) printf("%02x", pos[i]);
			printf("\n");
			printf("\t\tSubheader byte 0    %s 0x%02x\n", prchange(change[48]), pos[48]);
			printf("\t\tSubheader byte 1    %s 0x%02x\n", prchange(change[49]), pos[49]);
			printf("\t\tSubheader byte 2    %s 0x%02x\n", prchange(change[50]), pos[50]);
			printf("\t\tSubheader byte 3    %s 0x%02x\n", prchange(change[51]), pos[51]);
			printf("\n");
			break;
		case 0x07 :
			printf("verify error recovery mode page\n");
			break;
		case 0x0b :
			printf("media t, strerror(error)ypes supported page\n");
			break;
		case 0x0d :
			printf("cd device parameters page\n");
			inactiv_time_mult = pos[3] & 15;
			S_per_MSF = pos[5] + (pos[4] << 8);
			F_per_MSF = pos[7] + (pos[6] << 8);
			printf("\t\tHold track inactivity time  %s value %d (%s)\n", prchange(change[3] & 15), inactiv_time_mult, inactivities[inactiv_time_mult]);
			printf("\t\tSeconds per MSF             %s %d\n", prchange(change[5]), S_per_MSF);
			printf("\t\tFrames  per MSF             %s %d\n", prchange(change[7]), F_per_MSF);
			printf("\n");
			break;
		case 0x0e :
			printf("cd audio control page\n");
			printf("\t\tStop on track crossing      %s %s\n", prchange(change[ 2] & 2 ), pos[2] & 2 ? "On":"Off");
			printf("\t\tCDDA output port 0 channel  %s %d\n", prchange(change[ 8] & 31), pos[ 8] & 31);
			printf("\t\tOutput port 0 volume        %s %d\n", prchange(change[ 9]     ), pos[ 9]);
			printf("\t\tCDDA output port 1 channel  %s %d\n", prchange(change[10] & 31), pos[10] & 31);
			printf("\t\tOutput port 1 volume        %s %d\n", prchange(change[11]     ), pos[11]);
			printf("\t\tCDDA output port 2 channel  %s %d\n", prchange(change[12] & 31), pos[12] & 31);
			printf("\t\tOutput port 2 volume        %s %d\n", prchange(change[13]     ), pos[13]);
			printf("\t\tCDDA output port 3 channel  %s %d\n", prchange(change[14] & 31), pos[14] & 31);
			printf("\t\tOutput port 3 volume        %s %d\n", prchange(change[15]     ), pos[15]);
			printf("\n");
			break;
		case 0x1a :
			printf("power condition page\n");
			idle_timer_value    = pos[ 7] + (pos[ 6] << 8) + (pos[ 5] << 16) + (pos[ 4] << 24);
			standby_timer_value = pos[11] + (pos[10] << 8) + (pos[ 9] << 16) + (pos[ 8] << 24);
			printf("\t\tIdle timer activate         %s %s\n", prchange(change[3] & 2  ), pos[3] & 2 ? "On":"Off");
			printf("\t\tStandby timer active        %s %s\n", prchange(change[3] & 1  ), pos[3] & 1 ? "On":"Off");
			printf("\t\tIdle timer start value      %s %d * 100ms\n", prchange(change[ 7] & 1 ), idle_timer_value);
			printf("\t\tStandby timer start value   %s %d * 100ms\n", prchange(change[11] & 1 ), standby_timer_value);
			printf("\n");
			break;
		case 0x1c :
			printf("fault/failure reporting page\n");
			break;
		case 0x1d :
			printf("timeout and protect page\n");
			group1_timeouts = pos[7] + (pos[6] << 8);
			group2_timeouts = pos[9] + (pos[8] << 8);
			printf("\t\tTimeouts commands enabled   %s %s\n", prchange(change[4] & 4), pos[4] & 4 ? "On":"Off");
			printf("\t\tDisable device till powerd. %s %s\n", prchange(change[4] & 2), pos[4] & 2 ? "On":"Off");
			printf("\t\tSoftware write protect      %s %s\n", prchange(change[4] & 1), pos[4] & 1 ? "On":"Off");
			if (page_length >= 8)
				printf("\t\tGroup 1 minimum time outs   %s %d\n", prchange(change[7]), group1_timeouts);
			if (page_length >= 10)
				printf("\t\tGroup 2 minimum time outs   %s %d\n", prchange(change[9]), group2_timeouts);
			printf("\n");
			break;
		case 0x2a :
			printf("CD/DVD capabilities and mechanical status page\n");
			printf("\t\tNot dumped yet\n");
			printf("\n");
			break;
		case 0x2c :
			printf("MRW mode page\n");
			break;
		default :
			printf("Reserved/vendor specific\n");
			break;
	};
}


void dump_parameter_pages(void) {
	scsicmd cmd;
	int     blk_len = 60000, val_length, changable_length;
	uint8_t val[blk_len], changable[blk_len], zeros[256];
	int	voff, choff, vpage_code, vpage_length, chpage_code, chpage_length;
	int	error;

	bzero(val, blk_len);
	bzero(changable, blk_len);
	bzero(zeros, 256);

	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x5A;			/* MODE SENSE (10)		*/
	cmd[1] = 0;			/* allow multiple blocks	*/
	cmd[2] = 0x3F;			/* return all current values	*/
	cmd[7] = blk_len >> 8;		/* MSB block length 		*/
	cmd[8] = blk_len & 0xff;	/* LSB block length 		*/
	cmd[9] = 0;			/* control			*/
	error = scsi_call(SCSI_READCMD, &disc, cmd, 10, val, blk_len);
	if (error) {
		fprintf(stderr, "While reading all parameter pages : %s\n", strerror(error));
		return;
	};
	val_length = val[1] | (val[2] << 8);
	printf("\tSCSI mode sense for all pages returned %d bytes of data\n", val_length);

	cmd[2] = 64 | 0x3F;		/* get all changable values	*/
	error = scsi_call(SCSI_READCMD, &disc, cmd, 10, changable, blk_len);
	if (error) {
		fprintf(stderr, "While reading all changable parameter pages : %s\n", strerror(error));
	};
	changable_length = changable[1] | (changable[2] << 8);
	printf("\tSCSI mode sense changable parameters list returned %d bytes of data\n", changable_length);

	/* read the mode sense block and process page by page               */
	/* XXX assumption for now that val and changable are the same order */
	voff = 8;
	while (voff < val_length) {
		vpage_code    = val[voff    ] & 63;
		vpage_length  = val[voff + 1];

		if (vpage_code) {
			choff = 8; chpage_code = -1;
			while (choff < changable_length) {
				chpage_code   = changable[choff    ] & 63;
				chpage_length = changable[choff + 1];
				if (chpage_code == vpage_code) break;
				choff += chpage_length + 2;
			};
			if (chpage_code == vpage_code) {
				dump_parameter_page(val + voff, changable + choff);
			} else {
				dump_parameter_page(val + voff, zeros);
			};
		};
		voff += vpage_length + 2;
	};
}


int main(int argc, char **argv) {
	scsicmd cmd;
	uint8_t buf[36];
	struct scsi_addr saddr;
	int error;

	bzero(&disc, sizeof(disc));
	init_tables();

	if (argc != 2) {
		printf("Usage : %s devicename\n", argv[0]);
		return 1;
	};

	/* Open the device */
	disc.dev_name = strdup(argv[1]);
	printf("Opening device %s\n", disc.dev_name);
	if (scsi_open(&disc) != 0) {
		exit(1);
	};

	error = scsi_check_for_scsi(&disc);
	if (error) {
		fprintf(stderr, "sorry, not a SCSI device : %s\n", strerror(error));
		exit(1);
	};

	error = scsi_identify(&disc, &saddr);
	if (error) {
		fprintf(stderr, "SCSI identify returned : %s\n", strerror(error));
		exit(1);
	};

	printf("\n\nDevice dentifies itself as : ");
	if (saddr.type == 0) {
		printf("SCSI   busnum = %d, target = %d, lun = %d\n", saddr.addr.scsi.scbus, saddr.addr.scsi.target, saddr.addr.scsi.lun);
	} else {
		printf("ATAPI  busnum = %d, drive = %d\n", saddr.addr.atapi.atbus, saddr.addr.atapi.drive);
	};


	printf("\n\nTest unit ready\n");
	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0;	/* test unit ready */
	error = scsi_call(SCSI_READCMD, &disc, cmd, 1, buf, 36);
	if (error) perror("SCSI test unit ready returned : ");


	printf("\n\nCURRENT features\n");
	dump_drive_configuration(1);

	printf("\n\nCAPAPLE features\n");
	dump_drive_configuration(0);


	printf("\n\nRead recorded capacity\n");
	dump_recorded_capacity();

	printf("\n\nDisc information\n");
	dump_disc_information();

	printf("\n\nReading TOC\n");
	dump_toc_pma_atip_info();

	printf("\n\nReading SCSI 'mode sense' parameter page\n");
	dump_parameter_pages();

	scsi_close(&disc);

	return 0;
};

