/* $NetBSD$ */

/*
 * File "udf.c" is part of the UDFclient toolkit.
 * File $Id: udf_allocentries.c,v 1.2 2005/01/07 17:53:07 imago Exp $ $Name:  $
 *
 * Copyright (c) 2003, 2004 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udftoolkit/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/* XXX strip list to bare minimum XXX */
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>
#include <dirent.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <time.h>
//#include <math.h>	/* TEMP! */

#include "scsi_sense.h"
#include "scsilib.h"


/* for locals */
#include "udf.h"
#include "udf_bswap.h"
#include "udf_discop.h"
#include "uio.h"
#include <pthread.h>


/* for scsilib */
extern const char *dvname;


#ifndef MAX
#	define MAX(a,b) ((a)>(b)?(a):(b))
#	define MIN(a,b) ((a)<(b)?(a):(b))
#endif


//#define DEBUG(a) { a; };
#define DEBUG(a) if (0) { a; };



/******************************************************************************************
 *
 * Unallocated/freed space management and allocation entry operations
 *
 ******************************************************************************************/


void udf_merge_allocentry_queue(struct udf_alloc_entries *queue, uint32_t lb_size) {
	struct udf_allocentry *alloc_entry, *next_alloc;
	uint64_t this_end, next_start;
	int merge;

	TAILQ_FOREACH(alloc_entry, queue, next_alloc) {
		do {
			merge = 0;

			/* only non busy */
			if (alloc_entry->flags == UDF_SPACE_FREED) break;

			next_alloc = TAILQ_NEXT(alloc_entry, next_alloc);
			if (next_alloc) {
				if (next_alloc->flags != alloc_entry->flags) break;

				if (alloc_entry->flags == UDF_SPACE_ALLOCATED) {
					/* merge on virtual/physical lb_num base; they are automatically adjacent on offset */
					if (next_alloc->vpart_num != alloc_entry->vpart_num) break;
					this_end   = alloc_entry->lb_num * lb_size + alloc_entry->len;
					next_start = next_alloc->lb_num * lb_size;
	
					if (this_end != next_start) break;
				};

				/* Only merge if merge would result in a legal UDF allocation size */
				if (((uint64_t) alloc_entry->len + (uint64_t) next_alloc->len) > ((uint64_t) 1<<30)-1) break;

				/* merge! */
				alloc_entry->len = alloc_entry->len + next_alloc->len;
				TAILQ_REMOVE(queue, next_alloc, next_alloc);
				free(next_alloc);
				merge = 1;
			};
		} while (merge);
	}; /* foreach */
}


/* Splits up allocated pieces so that there is a break on `data_offset' and on `data_offset + data_length' */
int udf_cut_allocentry_queue(struct udf_alloc_entries *queue, uint32_t lb_size, uint64_t offset) {
	struct udf_allocentry *entry, *new_entry;
	uint64_t cur_offset;
	uint64_t total_length, extra_length, max_slot, new_length;
	uint64_t entry_offset;

	total_length = 0;
	TAILQ_FOREACH(entry, queue, next_alloc) {
		total_length += entry->len;
	};

	//printf("Cutting up at offset %ld (lb %ld)\n", offset, offset / lb_size);
	if (offset < total_length) {
		/* split */
		cur_offset = 0;
		TAILQ_FOREACH(entry, queue, next_alloc) {
			if ((offset >= cur_offset) && (offset < cur_offset + entry->len)) {
				/* overlap */
				entry_offset = offset - cur_offset;
				entry_offset = (entry_offset / lb_size) * lb_size;
				assert(entry_offset % lb_size == 0);
				if (entry_offset == 0) return 0;

				/* clone the current space */
				new_entry = calloc(1, sizeof(struct udf_allocentry));
				if (!new_entry) return ENOMEM;
				memcpy(new_entry, entry, sizeof(struct udf_allocentry));

				/* split! */
				//printf("split up lb %d + lb %d due to lb offset = %ld\n", entry->lb_num, entry->len / lb_size, entry_offset);

				entry->len           = entry_offset;
				new_entry->len      -= entry_offset;
				new_entry->lb_num   += entry_offset / lb_size;
				TAILQ_INSERT_AFTER(queue, entry, new_entry, next_alloc);
				DEBUG(printf("split up due to lb offset = %d\n", (int) entry_offset));
				return 0;
			};
			cur_offset += entry->len;
		};
		printf("Sanity check: i can't be here\n");
		exit(1);
	};

	/* no use to do more if we're there */
	if (offset == total_length) return 0;

	/*
	 * Glue extra piece on the queue (auto-extending)
	 * see if we reached our `goal'; see if we can just extent the last
	 * allocation entry but NEVER more or equal to one block size for that
	 * would alter semantics.
	 */
	entry = TAILQ_LAST(queue, udf_alloc_entries);
	if (!TAILQ_EMPTY(queue)) {
		extra_length = (uint64_t) lb_size*(((uint64_t) entry->len + lb_size -1) / lb_size) - entry->len;
		extra_length = MIN(extra_length, (offset - total_length));
		/* keep semantics: only meant for extending upto blocksize */
		if (extra_length < lb_size) {
			entry->len   += extra_length;
			total_length += extra_length;
		};
	};

	max_slot = ((((uint64_t) 1<<30)-1) / lb_size) * lb_size;
	while (offset > total_length) {
		/* grow queue by adding difference as a zero unallocated space */
		new_length  = offset - total_length;
		new_length  = MIN(max_slot, new_length);

		new_entry = calloc(1, sizeof(struct udf_allocentry));
		if (!new_entry) return ENOMEM;

		new_entry->len   = new_length;
		new_entry->flags = UDF_SPACE_FREE;
		TAILQ_INSERT_TAIL(queue, new_entry, next_alloc);

		total_length += new_entry->len;
	}; /* while */

	return 0;
}


/* Splits up allocated pieces so that there is a break on `data_offset' and on `data_offset + data_length' */
int udf_splitup_allocentry_queue(struct udf_alloc_entries *queue, uint32_t lb_size, uint64_t data_offset, uint64_t data_length, struct udf_allocentry **res_firstae, struct udf_allocentry **res_lastae) {
	struct udf_allocentry *entry, *prev_entry;
	uint64_t cur_offset, len;
	int error;

	entry = prev_entry = NULL;
#if 0
	printf("Split %ld + %ld : \n", data_offset, data_length);
	printf("PRE SPLIT block = lb %d + lb %d (%ld bytes)\n", (int) (data_offset / lb_size), (int) (data_length / lb_size), data_length);
	TAILQ_FOREACH(entry, queue, next_alloc) {
		printf("\t(lb %08d + lb %08d[+ %d]) flag %d\n", entry->lb_num, entry->len/lb_size, entry->len % lb_size, entry->flags);
	};
	printf("END PRE\n");
#endif

	/* cut the string at the specified places */
	error = udf_cut_allocentry_queue(queue, lb_size, data_offset);
	error = udf_cut_allocentry_queue(queue, lb_size, data_offset + data_length);

#if 0
	printf("POST SPLIT\n");
	TAILQ_FOREACH(entry, queue, next_alloc) {
		printf("\t(lb %08d + lb %08d[+ %d]) flag %d\n", entry->lb_num, entry->len/lb_size, entry->len % lb_size, entry->flags);
	};
	printf("END POST\n\n");
#endif

	if ((res_firstae == NULL) && (res_lastae == NULL)) return 0;

	if (res_firstae) *res_firstae = NULL;
	if (res_lastae)  *res_lastae  = NULL;

	DEBUG(printf("SEARCH SPLIT block = %"PRIu64" + %"PRIu64"\n", (data_offset / lb_size), (data_length / lb_size)));
	/* search the element-range this splitting induced */
	cur_offset = 0;
	entry = TAILQ_FIRST(queue);
	while (entry) {
		len = entry->len;

		DEBUG(printf("\t(%d + %d) flag %d\n", (int) entry->lb_num, (int) entry->len, (int) entry->flags));
		if (cur_offset + len > data_offset) {
			if (res_firstae) *res_firstae = entry;
			DEBUG(printf("\t\tReturned as first\n"));
			break;
		};
		/* advance */
		cur_offset += len;
		entry = TAILQ_NEXT(entry, next_alloc);
	};
	prev_entry = entry;
	while (entry) {
		len = entry->len;

		if (cur_offset + len > data_offset + data_length) {
			break;
		};
		DEBUG(printf("\t(%d + %d) flag %d\n", (int) entry->lb_num, (int) entry->len, (int) entry->flags));

		/* advance */
		cur_offset += len;
		prev_entry  = entry;
		entry = TAILQ_NEXT(entry, next_alloc);
	};
	if (res_lastae)  *res_lastae = prev_entry;
	DEBUG(printf("\t\tReturned as last\n\t<skip>\n"));

	DEBUG(printf("END POST\n\n"));

	if (res_firstae) assert(*res_firstae);
	if (res_lastae)  assert(*res_lastae);
	
	return 0;
}


/* tested OK */
int udf_mark_allocentry_queue(struct udf_alloc_entries *queue, uint32_t lb_size, uint64_t data_offset, uint64_t data_length, int mark, struct udf_allocentry **res_firstae, struct udf_allocentry **res_lastae) {
	struct udf_allocentry *alloc_entry, *first_alloc_entry, *last_alloc_entry;
	int error;

	DEBUG(printf("mark %d\n", mark));
	/* first split up so we don't have to worry about boundaries */
	error = udf_splitup_allocentry_queue(queue, lb_size, data_offset, data_length, &first_alloc_entry, &last_alloc_entry);
	assert(error == 0);

	alloc_entry = first_alloc_entry;
	/* inclusive last_alloc_entry */
	last_alloc_entry = TAILQ_NEXT(last_alloc_entry, next_alloc);
	while (alloc_entry != last_alloc_entry) {
		DEBUG(printf("marking %d + %d into type %d\n", alloc_entry->lb_num, alloc_entry->len/lb_size, mark);)
		alloc_entry->flags = mark;
		alloc_entry = TAILQ_NEXT(alloc_entry, next_alloc);
	};

	if (res_firstae) *res_firstae = first_alloc_entry;
	if (res_lastae)  *res_lastae  = last_alloc_entry;

	return 0;
} 


/* get at most req_lbs from the partition and mark the area used */
int udf_partmap_get_lbs(struct udf_log_vol *udf_log_vol, uint16_t vpart_num, uint32_t req_lbs, int is_metadata, uint32_t *res_start_lb, uint32_t *res_num_lbs) {
	struct udf_partition     *udf_partition;
	struct udf_part_mapping  *udf_part_mapping;
	struct udf_allocentry    *alloc_entry, *chosen;
	struct udf_alloc_entries *queue;
	uint64_t	 start;
	uint32_t	 lb_size;
	uint32_t	 req_size, size;
	int		 error;

	lb_size = udf_rw32(udf_log_vol->log_vol->lb_size);
	error = udf_logvol_vpart_to_partition(udf_log_vol, vpart_num, &udf_part_mapping, &udf_partition);
	if (error) return error;

	switch (udf_part_mapping->udf_part_mapping_type) {
		case UDF_PART_MAPPING_PHYSICAL :
		case UDF_PART_MAPPING_SPARABLE :
			/* sparable is just like the physical mapping */
			req_size = lb_size * req_lbs;

			/*
			 * Until proper ICB clustering is implemented all
			 * ICB's are sheduled at the end of free space to keep
			 * out of the filestream's way.
			 */
			if (is_metadata) {
				queue  = &udf_partition->unalloc_space_queue;
				udf_merge_allocentry_queue(queue, lb_size);

				start = (uint64_t) lb_size * (udf_rw32(udf_partition->partition->part_len));
				assert(start != 0);

				TAILQ_FOREACH_REVERSE(alloc_entry, queue, udf_alloc_entries, next_alloc) {
					start = start - alloc_entry->len;
					if (alloc_entry->flags == UDF_SPACE_FREE) {
						chosen = alloc_entry;
						size = MIN(chosen->len, req_size);

						/* round to an integral number of blocks */
						size = lb_size * (size / lb_size);
						if (size == 0) continue;

						/* we can only give out on lb_size */
						assert(chosen->len % lb_size == 0);
						assert(size % lb_size == 0);

						/* we want the end of the chunk */
						start = start + chosen->len - size;

						/* we can only start on lb_size */
						assert(start % lb_size == 0);

						*res_start_lb = start / lb_size;
						if (res_num_lbs) *res_num_lbs  = size / lb_size;

						udf_partition->free_unalloc_space -= size;
						error = udf_mark_allocentry_queue(queue, lb_size, start, size, UDF_SPACE_ALLOCATED, NULL, NULL);
						return error;
					};
				};
				printf("YIKES! out of space?\n");
				return ENOSPC;
			};

/* XXX mutex!! ?? XXX */
			/* look for the most suiting piece; XXX allocation algoritm XXX */
			/* first pick, smallest fit */
			chosen = NULL;
			DEBUG(printf("Start alloc: requested %d lbnums\n", req_lbs));
			queue  = &udf_partition->unalloc_space_queue;
			udf_merge_allocentry_queue(queue, lb_size);

			alloc_entry = TAILQ_FIRST(queue);
			start = alloc_entry->lb_num;
			assert(start == 0);

			TAILQ_FOREACH(alloc_entry, queue, next_alloc) {
				if (alloc_entry->flags == UDF_SPACE_FREE) {
					/* nothing -> choose first */
					if (!chosen) chosen = alloc_entry;

					/* Algorithm choice: first come, first served allocation */
					break;
#if 0
					/* Algorithm choice: seach for best fit (wastfull, leaves small gaps around */
					/* too big but smaller than choosen one? -> choose */
					if ((alloc_entry->len >= req_size) && (chosen->len < chosen->len)) {
						chosen = alloc_entry;
					};
					/* if too small, choose biggest */
					if ((chosen->len < req_size) && (alloc_entry->len > chosen->len)) {
						chosen = alloc_entry;
					};
					DEBUG(printf("\tcurrent chosen->lb_num = %8d, size = %d (%d lb)\n", chosen->lb_num, chosen->len, chosen->len/lb_size));

					/* if exact fit return this size */
					if (chosen->len == req_size) break;
#endif
				};
				start = start + alloc_entry->len;
			};
			if (!chosen) return ENOSPC;

			DEBUG(printf("Alloc chosen!\n"));
			size = MIN(chosen->len, req_size);
			size = lb_size * (size / lb_size);

			/* return chosen entry */
			*res_start_lb = start / lb_size;
			if (res_num_lbs) *res_num_lbs  = size / lb_size;

			/* mark allocation */
			udf_partition->free_unalloc_space -= size;
			error = udf_mark_allocentry_queue(queue, lb_size, (uint64_t) (*res_start_lb) * lb_size, size, UDF_SPACE_ALLOCATED, NULL, NULL);

			return error;
		case UDF_PART_MAPPING_VIRTUAL  :
			/* strict increasing virtual adress giveout */
			printf("UDF: get lbs from virtual partition mapping not implemented yet\n");
			return EBADF;
		case UDF_PART_MAPPING_META     :
			/* select blobs from the metadata file */
			printf("UDF: get lbs from metadata partition mapping not implemented yet\n");
			break;
	};
	return 0;
}


/* inverse of udf_partmap_get_lbs */
int udf_partmap_free_lbs(struct udf_log_vol *udf_log_vol, uint16_t vpart_num, uint32_t lb_num, uint32_t size) {
	struct udf_partition     *udf_partition;
	struct udf_part_mapping  *udf_part_mapping;
	struct udf_alloc_entries *queue;
	uint32_t	 lb_size;
	int		 error;

	if (!udf_log_vol) return 0;

	lb_size = udf_rw32(udf_log_vol->log_vol->lb_size);
	error = udf_logvol_vpart_to_partition(udf_log_vol, vpart_num, &udf_part_mapping, &udf_partition);
	if (error) return error;

	/* space can only be freed in lb_size chuncks by definition */
	size = lb_size * ((size + lb_size -1) / lb_size);
	switch (udf_part_mapping->udf_part_mapping_type) {
		case UDF_PART_MAPPING_PHYSICAL :
		case UDF_PART_MAPPING_SPARABLE :
			/* sparable is just like the physical mapping */
/* XXX mutex!! ?? XXX */
			queue = &udf_partition->unalloc_space_queue;	/* TODO or freed- for MO media */
			error = udf_mark_allocentry_queue(queue, lb_size, (uint64_t) lb_num * lb_size, size, UDF_SPACE_FREE, NULL, NULL);
			udf_partition->free_unalloc_space += size;
			return error;
		case UDF_PART_MAPPING_VIRTUAL  :
			/* strict increasing virtual adress giveout */
			printf("UDF: get lbs from virtual partition mapping not implemented yet\n");
			return EBADF;
		case UDF_PART_MAPPING_META     :
			/* select blobs from the metadata file */
			printf("UDF: get lbs from metadata partition mapping not implemented yet\n");
			break;
	};
	return 0;

}


int udf_request_free_logblks(struct udf_log_vol *udf_log_vol, int is_metadata, uint32_t req_lbs, char *what, uint16_t *res_vpart_num, uint32_t *res_start_lb, uint32_t *res_num_lbs) {
	struct udf_partition *udf_partition;
	struct udf_part_mapping *udf_part_mapping;
	uint32_t	 lb_size, num_lbs;
	uint16_t	 vpart_num;
	int		 ok;
	int		 error;

	lb_size     = udf_rw32(udf_log_vol->log_vol->lb_size);
	assert(udf_log_vol);

	/* select udf partition to write to depending on the contents */
	vpart_num = is_metadata ? udf_log_vol->metadata_vpart : udf_log_vol->data_vpart;

	DEBUG(printf("UDF request %d free logblks for node `%s`\n", req_lbs, what));
	do {
		DEBUG(printf("do: vpart_num = %d\n", vpart_num));
		error = udf_logvol_vpart_to_partition(udf_log_vol, vpart_num, &udf_part_mapping, &udf_partition);
		ok = (udf_partition != NULL) && (udf_part_mapping != NULL);
		if (ok) ok = (udf_partition->free_unalloc_space + udf_partition->free_freed_space >= lb_size);
		if (ok && !is_metadata) ok = udf_part_mapping->data_writable;
		if (ok &&  is_metadata) ok = udf_part_mapping->metadata_writable;
		if (!ok) {
			DEBUG(printf("advance vpart\n"));
			vpart_num = is_metadata ? ++(udf_log_vol->metadata_vpart) : ++(udf_log_vol->data_vpart);
			if (vpart_num >= udf_log_vol->num_part_mappings) {
				udf_log_vol->metadata_vpart = udf_log_vol->data_vpart = 0;
				printf("UDF: logvol discs full ?\n");
				return ENOSPC;
			};
		};
	} while (!ok);
	DEBUG(printf("Got vpart_num = %d\n", vpart_num));

	/* got a partition to snoop lb's from */
	*res_vpart_num = vpart_num;
	error = udf_partmap_get_lbs(udf_log_vol, vpart_num, req_lbs, is_metadata, res_start_lb, &num_lbs);
	if (res_num_lbs) *res_num_lbs = num_lbs;

	DEBUG(printf("get_free_logic_blocks : req %d, got %d\n", req_lbs, num_lbs));
	return 0;
}


int udf_node_request_free_logblks(struct udf_node *udf_node, int req_lbs, uint16_t *res_vpart_num, uint32_t *res_start_lb, uint32_t *res_num_lbs) {
	int is_metadata;

	is_metadata = (udf_node->udf_filetype != UDF_ICB_FILETYPE_RANDOMACCESS);
	return udf_request_free_logblks(udf_node->udf_log_vol, is_metadata, req_lbs, udf_node->dirent.d_name,
			res_vpart_num, res_start_lb, res_num_lbs);
}


int udf_filepart_populate_allocentry(struct udf_node *udf_node, struct udf_allocentry *alloc_entry, struct udf_allocentry **next_entry) {
	struct udf_allocentry *cur_entry, *new_alloc_entry;
	int32_t len, lb_size, lbs_needed, start_lb, num_lbs;
	uint16_t vpart_num;
	int      error;

	if (alloc_entry->flags != UDF_SPACE_FREE) {
		printf("UDF warning: filepart_populate_allocentry called on alloc entry with flags = %d\n", alloc_entry->flags);
		return 0;
	};

	/* populate non allocated entry */
	lb_size    = udf_rw32(udf_node->udf_log_vol->log_vol->lb_size);

	len = alloc_entry->len;
	lbs_needed = (len + lb_size -1) / lb_size;

	*next_entry     = NULL;
	cur_entry       = alloc_entry;
	new_alloc_entry = cur_entry;
	while (lbs_needed) {
		error = udf_node_request_free_logblks(udf_node, lbs_needed, &vpart_num, &start_lb, &num_lbs);
		if (error) return error;

		/* fill in results */
		new_alloc_entry->len       = MIN(len, num_lbs * lb_size);	/* dont forget to count last block */
		new_alloc_entry->lb_num    = start_lb;
		new_alloc_entry->vpart_num = vpart_num;
		new_alloc_entry->flags     = UDF_SPACE_FREED;			/* mark allocated but not written out yet */

		len        -= new_alloc_entry->len;
		lbs_needed -= num_lbs;

		if (new_alloc_entry != cur_entry) {
			TAILQ_INSERT_AFTER(&udf_node->alloc_entries, cur_entry, new_alloc_entry, next_alloc);
		};
		cur_entry = new_alloc_entry;	/* move along */
		if (lbs_needed) {
			new_alloc_entry = calloc(1, sizeof(struct udf_allocentry));
			if (!new_alloc_entry) return ENOMEM;
		};
	};

	*next_entry = cur_entry;
	return 0;
}


/*
 * Differs from mark_allocentry_queue in that its interpreting for the
 * different media types
 */
int udf_filepart_mark_extent(struct udf_node *udf_node, uint64_t data_offset, uint64_t data_length, int mark) {
	struct udf_allocentry *first_alloc_entry, *last_alloc_entry;
	struct udf_allocentry *alloc_entry;
	uint64_t lb_size, len;
	int      rewritable, error;

	rewritable = udf_node->udf_log_vol->primary->udf_session->disc->rewritable;
	lb_size    = udf_rw32(udf_node->udf_log_vol->log_vol->lb_size);

	DEBUG(
		printf("udf_filepart_mark_extent %8d + %8d (%d bytes) with mark %d\n", (int) (data_offset/lb_size), (int) (data_length/lb_size), (int) data_length, mark);
		if (0) {
			printf("PRE MARKING %d :\n", mark);
			TAILQ_FOREACH(alloc_entry, &udf_node->alloc_entries, next_alloc) {
				printf("\t(%d, %d) flag %d\n", (int) alloc_entry->lb_num, (int) alloc_entry->len, (int) alloc_entry->flags);
			};
			printf("END PRE\n");
		};
	);

	error = udf_splitup_allocentry_queue(&udf_node->alloc_entries, lb_size, data_offset, data_length, &first_alloc_entry, &last_alloc_entry);
	assert(error == 0);

	/*
	 * we now have a nicely split up allocation entry queue at the
	 * breakpoints we need. Processing of the intermediate allocation
	 * entries is dependent on contents and the disc type
	 */
	if (rewritable) {
		/* just make sure all is allocated */
		alloc_entry = first_alloc_entry;
		/* inclusive last_alloc_entry entry */
		last_alloc_entry = TAILQ_NEXT(last_alloc_entry, next_alloc);
		while (alloc_entry != last_alloc_entry) {
			if (alloc_entry->flags == UDF_SPACE_FREE) {
				/* non allocated entry on a rewriteble -> allocate */
				error = udf_filepart_populate_allocentry(udf_node, alloc_entry, &alloc_entry);
				if (error) return error;
			};

			alloc_entry->flags = mark;
			alloc_entry = TAILQ_NEXT(alloc_entry, next_alloc);
		};
	} else {
		/* write once */
		/* TODO test this out */
		if (mark == UDF_SPACE_FREED) {
			/* patch up the first allocation entry to completely hold the content */
			alloc_entry = first_alloc_entry;
	
			/* skip the current one to not lose the insertion point	   */
			/* accumulate deleted length to make sure we don't goof up */
			alloc_entry = TAILQ_NEXT(alloc_entry, next_alloc);
			len = first_alloc_entry->len;
	
			/* delete all intermediate spaces and allocate anew */
			while (alloc_entry != TAILQ_NEXT(last_alloc_entry, next_alloc)) {
				len += alloc_entry->len;
				TAILQ_REMOVE(&udf_node->alloc_entries, alloc_entry, next_alloc);
	
				alloc_entry = TAILQ_NEXT(alloc_entry, next_alloc);
			};
			alloc_entry = first_alloc_entry;
			alloc_entry->len   = len;
			alloc_entry->flags = UDF_SPACE_FREE;		/* mark non allocated */

			/* allocate conglomorated alloc_entry */
			error = udf_filepart_populate_allocentry(udf_node, alloc_entry, &alloc_entry);
			if (error) return error;
		};
		if (mark == UDF_SPACE_ALLOCATED) {
			alloc_entry = first_alloc_entry;
			while (alloc_entry != TAILQ_NEXT(last_alloc_entry, next_alloc)) {
				assert(alloc_entry->flags != UDF_SPACE_FREE);

				alloc_entry->flags = UDF_SPACE_ALLOCATED;
				alloc_entry = TAILQ_NEXT(alloc_entry, next_alloc);
			};
		};
	};

	DEBUG(
		if (0) {
			printf("MARKING %d result :\n", mark);
			TAILQ_FOREACH(alloc_entry, &udf_node->alloc_entries, next_alloc) {
				printf("\t(%d, %d) flag %d\n", (int) alloc_entry->lb_num, (int) alloc_entry->len, (int) alloc_entry->flags);
			};
			printf("END MARKING\n");
		};
	);
	
	return 0;
}



