/* $NetBSD$ */

/*
 * File "udf_discinfo.c" is part of the UDFclient toolkit.
 * File $Id: udf_discop.c,v 1.31 2005/01/14 02:26:57 imago Exp $ $Name:  $
 *
 * Copyright (c) 2003, 2004 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udftoolkit/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/* XXX strip this XXX */
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>
#include <dirent.h>
#include <string.h>
#include <strings.h>
#include <limits.h>

#include "scsi_sense.h"
#include "scsilib.h"


/* for locals */
#include "udf_discop.h"
#include "udf_bswap.h"


#ifndef MAX
#	define MAX(a,b) ((a)>(b)?(a):(b))
#	define MIN(a,b) ((a)<(b)?(a):(b))
#endif


//#define DEBUG(a) { a; };
#define DEBUG(a) if (0) { a; };


/* globals */
extern int udf_verbose;
//extern int error;


/******************************************************************************************
 *
 * Tables and helper functions section
 *
 ******************************************************************************************/

int read_cd_hex2(int val) {
	int nl, nh;

	nl = val & 15;
	nh = val >> 4;
	if (nl >= 'A') nl -= 'A' + 10;
	if (nh >= 'A') nh -= 'A' + 10;

	return (nh*16) + nl;
}


int read_cd_bcd(int val) {
	int nl, nh;

	nl = (val & 15) - '0';
	nh = (val >> 4) - '0';
	if ((nl < 0 || nl > 9) || (nh < 0 || nh > 9)) return val;

	return nh*10 + nl;
}


int cd_msf2lba(int m, int s, int f) {
	return 4500*m + 75*s + f - 150;
}


/******************************************************************************************
 *
 * SCSI/ATAPI abstraction section; could be cleaned up a bit
 *
 ******************************************************************************************/


int scsi_call(int flags, struct udf_discinfo *disc, void *cmd, int cmdlen, void *data, int datalen) {
#ifdef SCSI
	return scsi_command(disc, cmd, cmdlen, data, datalen, 30000, flags);
#else
	return ENODEV;
#endif
}


#ifndef SCSI
int scsi_check_for_scsi(struct udf_discinfo *disc) {
	return ENODEV;
}
#endif


/******************************************************************************************
 *
 * Disc level operations 
 *
 ******************************************************************************************/
int udf_discinfo_is_cd_or_dvd(struct udf_discinfo *disc) {
	/* check device type */
	switch (disc->devdrv_class & UDF_DEVDRV_CLASS) {
		case UDF_DEVDRV_CLASS_FILE :
		case UDF_DEVDRV_CLASS_DISC :
			/* not nessisary */
			return 0;
		case UDF_DEVDRV_CLASS_CD   :
		case UDF_DEVDRV_CLASS_DVD  :
			/* it really is */
			return 1;
		default :
			break;
	};
	return ENODEV;
}


int udf_discinfo_set_recording_parameters(struct udf_discinfo *discinfo, int testwriting) {
	scsicmd cmd;
	int     blk_len = 10000, val_len, packet;
	uint8_t res[blk_len], *pos;
	int	error;

	if (!udf_discinfo_is_cd_or_dvd(discinfo)) return 0;

	/* Set up CD/DVD recording parameters */
	if (!discinfo->recordable) return 0;

	DEBUG(printf("Setting device's recording parameters\n"));

	/* all recordables are not packet based.... */
	packet = 0;
	/* except rewritables! */
	if (discinfo->rewritable) packet = 1;

	val_len = 0x32+2+8;
	bzero(res, val_len);

	pos = res + 8;

	bzero(cmd, SCSI_CMD_LEN);
	if (!packet) {
		pos[ 0] = 0x05;			/* page code 5 : cd writing					*/
		pos[ 1] = 0x32;			/* length in bytes						*/
		pos[ 2] = 64 + 0;		/* BUFE + write type 1 : track at once				*/
		if (testwriting) pos[ 2] += 16;
		pos[ 3] = (3<<6) | 5;		/* next session OK, data packet, rec. incr. var packets		*/
		pos[ 4] = 8;			/* ISO mode 1							*/
		pos[ 8] = 0;			/* normal CD-DA/CD-ROM or data disc				*/
		DEBUG(printf("\tsetting up for sequential writing\n"));
	} else {
		pos[ 0] = 0x05;			/* page code 5 : cd writing					*/
		pos[ 1] = 0x32;			/* length in bytes						*/
		pos[ 2] = 0;			/* write type 0 : packet/incremental				*/
		if (testwriting) pos[ 2] += 16;
		pos[ 3] = (3<<6) | 32 | 5;	/* next session OK, data packet, rec. incr. fixed packets	*/
		pos[ 4] = 10;			/* ISO mode 2; XA form 1					*/
		pos[ 8] = 0x20;			/* CD-ROM XA disc or DDCD disc */
		pos[10] = 0;			/* MSB packet size 						*/
		pos[11] = 0;
		pos[12] = 0;
		pos[13] = 32;			/* LSB packet size in SECTORS 					*/
		DEBUG(printf("\tsetting up for packet writing with packet size 32\n"));
	};

	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x55;			/* MODE SELECT (10)		*/
	cmd[1] = 16;			/* PF format			*/
	cmd[7] = val_len >> 8;		/* length of blob		*/
	cmd[8] = val_len & 0xff;
	cmd[9] = 0;			/* control			*/

	error = scsi_call(SCSI_WRITECMD, discinfo, cmd, 10, res, val_len);
	if (error) {
		perror("While WRTITING parameter page 5");
		return error;
	};

	/* Set CD/DVD speed on 'optimal' for it doesnt seem to do it automatically */
	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0xBB;				/* Set CD speed */
	cmd[ 1] = 1;				/* select CAV (1) or CLV (0) recording */
	cmd[ 2] = 0xff;
	cmd[ 3] = 0xff;				/* max read performance speed */
	cmd[ 4] = 0xff;
	cmd[ 5] = 0xff;				/* max write performance speed; applic? */
	cmd[11] = 0;				/* control */
	error = scsi_call(SCSI_WRITECMD, discinfo, cmd, 12, NULL, 0);
	if (error) {
		perror("While setting speed");
		return error;
	};

	/* flag OK */
	return 0;
}


/*
 * important: it must be called when stopping; it can be called at start too
 * just in case
 */
int udf_discinfo_finish_writing(struct udf_discinfo *discinfo) {
	scsicmd cmd;
	int	error;

	if (!udf_discinfo_is_cd_or_dvd(discinfo)) return 0;

	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x35;				/* Synchrnise cache	*/
	cmd[ 9] = 0;				/* control		*/
	error = scsi_call(SCSI_WRITECMD, discinfo, cmd, 10, NULL, 0);
	if (error) {
		perror("While synchronising write cache");
	};

	return error;
}


int udf_discinfo_reserve_track_in_logic_units(struct udf_discinfo *discinfo, uint32_t logic_units) {
	scsicmd cmd;
	int	error;

	if (!udf_discinfo_is_cd_or_dvd(discinfo)) return ENODEV;

	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x53;				/* reserve track	*/
	cmd[ 5] = (logic_units  >> 24) & 0xff;	/* size MSB		*/
	cmd[ 6] = (logic_units  >> 16) & 0xff;
	cmd[ 7] = (logic_units  >>  8) & 0xff;
	cmd[ 8] = (logic_units       ) & 0xff;	/* size LSB		*/
	cmd[ 9] = 0;				/* control		*/
	error = scsi_call(SCSI_WRITECMD, discinfo, cmd, 10, NULL, 0);

	return error;
}


int udf_discinfo_close_track(struct udf_discinfo *discinfo, uint16_t trackno) {
	scsicmd cmd;
	int	error;

	if (!udf_discinfo_is_cd_or_dvd(discinfo)) return ENODEV;

	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x5B;				/* close session/track	*/
	cmd[ 2] = 1;				/* track 		*/
	cmd[ 4] = (trackno      >>  8) & 0xff;	/* specify trackno MSB	*/
	cmd[ 5] = (trackno           ) & 0xff;	/*         trackno LSB	*/
	cmd[ 9] = 0;				/* control		*/
	error = scsi_call(SCSI_WRITECMD, discinfo, cmd, 10, NULL, 0);

	return error;
}


int udf_discinfo_repair_track(struct udf_discinfo *discinfo, uint16_t trackno) {
	scsicmd cmd;
	int	error;

	if (!udf_discinfo_is_cd_or_dvd(discinfo)) return ENODEV;

	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x58;				/* repair track		*/
	cmd[ 4] = (trackno      >>  8) & 0xff;	/* specify trackno MSB	*/
	cmd[ 5] = (trackno           ) & 0xff;	/*         trackno LSB	*/
	cmd[ 9] = 0;				/* control		*/
	error = scsi_call(SCSI_WRITECMD, discinfo, cmd, 10, NULL, 0);

	return error;
}


/*
 * This routine 'get disc type' is a bit 'magic' in that it tries to figure out what
 * kind of device/disc we are dealing with without going into obscure functions to
 * let it be sorted out for us; ONLY fills in devdrv_class and MMC profile.
 */

int udf_discinfo_get_disc_type(struct udf_discinfo *disc) {
	struct stat stat;
	scsicmd	    cmd;
	uint8_t	    buf[256];
	int         error;

	/* assume generic CD-ROM with no known MMC profile */
	disc->devdrv_class = UDF_DEVDRV_CLASS_CD;
	disc->mmc_profile  = 0;

	/* check if its a regular file */
	fstat(disc->fhandle, &stat);
	if (S_ISREG(stat.st_mode)) {
		UDF_VERBOSE(printf("UDF device %s is a regular file\n", disc->dev_name));
		disc->devdrv_class   = UDF_DEVDRV_CLASS_FILE;
		return 0;
	};

	/* check if its a ATAPI/SCSI device */
	error = scsi_check_for_scsi(disc);
	if (error) {
		/* obviously no ATAPI/SCSI device -> has to be IDE disc or other but no file either */
		disc->devdrv_class   = UDF_DEVDRV_CLASS_DISC;
		UDF_VERBOSE(printf("Got error executing SCSI command, assuming IDE disc\n"));
		return 0;
	};

	/* get MMC profile */
	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x46;				/* Get configuration */
	cmd[ 8] = 32;				/* just a small buffer size */
	cmd[ 9] = 0;				/* control */
	error = scsi_call(SCSI_READCMD, disc, cmd, 10, buf, 32);
	if (!error) {
		disc->mmc_profile = buf[7] | (buf[6] << 8);
	} else {
		disc->mmc_profile = 0;	/* mark unknown MMC profile */
	};
	UDF_VERBOSE_MAX(printf("Device has MMC profile 0x%02x\n", disc->mmc_profile));

	/* try to be smart */
	switch (disc->mmc_profile) {
		case 0x01 : case 0x02 : case 0x03 : case 0x04 : case 0x05 :
			/* SCSI discs class; treat like normal discs */
			disc->devdrv_class   = UDF_DEVDRV_CLASS_DISC;
			UDF_VERBOSE(printf("SCSI disc detected; treating like normal disc device\n"));
			return 0;
		/* 0x00, 0x08, 0x09, 0x0a : different types of CD-ROM devices like CD-R/RW etc. */
		case 0x00 :
			disc->devdrv_class = UDF_DEVDRV_TYPE_CD;	/* not allways clear */
			break;
		case 0x08 :
			disc->devdrv_class = UDF_DEVDRV_TYPE_CD;
			break;
		case 0x09 :
			disc->devdrv_class = UDF_DEVDRV_TYPE_CD_R;
			break;
		case 0x0a :
			disc->devdrv_class = UDF_DEVDRV_TYPE_CD_RW;
			break;
		/* 0x10...0x14 DVD-ROM and DVD_MIN devices */
		case 0x10 :	/* DVD-ROM */
		case 0x11 :	/* DVD-R   */
		case 0x12 :	/* DVD-RAM */
		case 0x13 :	/* DVD-RW (restricted overwrite) */
		case 0x14 :	/* DVD-RW (sequential)           */
			disc->devdrv_class = UDF_DEVDRV_TYPE_DVD;
			break;
		/* 0x1a..0x1b DVD_PLUS devices */
		case 0x1a :
			disc->devdrv_class = UDF_DEVDRV_TYPE_DVD_PLUS_R;
			break;
		case 0x1b :
			disc->devdrv_class = UDF_DEVDRV_TYPE_DVD_PLUS_RW;
			break;
		default :
			break;
	};

	/* We are now sure its either a CD or a DVD ; but we might still be on a legacy CD-ROM drive */
	return 0;
}


/* 10000 is arbitrary */
int udf_get_disc_info(struct udf_discinfo *disc) {
	scsicmd		 cmd;
	struct stat	 stat;
	uint32_t	 res_len = 10000, val_len;
	uint8_t		 res[res_len];
	uint32_t	 first_track, last_track;
	uint32_t	 first_session, last_session;
	uint32_t	 first_track_last_session, last_track_last_session;
	uint32_t	 next_writable_addr, packet_size;
	uint32_t	 track, session;
	uint32_t	 cntrl, addr, tno, point, min, sec, frame, pmin, psec, pframe;
	uint32_t	 data_length, pos;
	int		 nwa_valid;
	off_t		 track_start, track_end, track_size, disc_size, free_blocks;
	int		 error;

	error = udf_discinfo_get_disc_type(disc);
	if (error) {
		perror("Error during classification of disc; skipping disc\n");
		return error;
	};

	if (disc->devdrv_class == UDF_DEVDRV_CLASS_FILE) {
		fstat(disc->fhandle, &stat);
		disc->sequential         = 0;					/* full r/w		*/
		disc->recordable         = 1;					/* assuming rw access	*/
		disc->rewritable         = 1;
		disc->sector_size	 = UDF_SECTOR_SIZE;			/* take a value; not applicable	really */
		disc->link_size		 = 0;					/* no link lossage	*/
		disc->disc_state         = DISC_STATE_NOT_SERIAL;
		disc->last_session_state = SESSION_STATE_INCOMPLETE;
		disc->num_sessions       = 1;
		disc->session_start [0]  = 0;
		disc->session_end   [0]  =  stat.st_size    / UDF_SECTOR_SIZE;	/* sectors right? */
		disc->next_writable [0]  = (stat.st_size    / UDF_SECTOR_SIZE) + 1;
		disc->packet_size   [0]  =  stat.st_blksize / UDF_SECTOR_SIZE;
		return 0;
	};

	if (disc->devdrv_class == UDF_DEVDRV_CLASS_DISC) {
		disc->sequential         = 0;					/* full r/w		*/
		disc->recordable         = 1;
		disc->rewritable         = 1;
		disc->sector_size	 = DISC_SECTOR_LENGTH;	/* XXX normal on Unix systems; assumption!!! XXX */
		disc->link_size		 = 0;					/* no link lossage	*/
		disc->disc_state         = DISC_STATE_NOT_SERIAL;
		disc->last_session_state = SESSION_STATE_COMPLETE;
		disc->num_sessions       = 1;
		disc->session_start [0]  = 0;
		disc->session_end   [0]  = 0;	/* XXX how to determine? 'BIG Q XXX */
		fprintf(stderr," UDF: no support for disc devices yet\n");
		exit(1);

		fprintf(stderr, "UDF: warning... reading/writing on 'disc' device\n");
		fprintf(stderr, "UDF: warning... assuming sector_size; of %d bytes\n", DISC_SECTOR_LENGTH);
		return 0;
	}

	/* classes CD and DVD remain; we can be on a DVD/CD recordable or on a legacy CD-ROM */

	/* Set CD/DVD speed on 'optimal' for it doesnt seem to do it automatically */
	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0xBB;				/* Set CD speed */
	cmd[ 1] = 0;				/* select CLV recording */
	cmd[ 2] = 0xff;
	cmd[ 3] = 0xff;				/* max read performance speed */
	cmd[ 4] = 0xff;
	cmd[ 5] = 0xff;				/* max write performance speed; applic? */
	cmd[11] = 0;				/* control */
	scsi_call(SCSI_WRITECMD, disc, cmd, 12, NULL, 0);


	/* Get track information */
	val_len = 12;
	bzero(res, val_len);
	bzero(cmd, SCSI_CMD_LEN);

	cmd[0] = 0x51;				/* Read disc information */
	cmd[7] = val_len >> 8;			/* MSB allocation length */
	cmd[8] = val_len & 0xff;		/* LSB allocation length */
	cmd[9] = 0;				/* control */
	error = scsi_call(SCSI_READCMD, disc, cmd, 10, res, val_len);

	if (!error) {
		/* we are a MMC3+ device! */
		disc->disc_state		 =  res[2]       &  3;		/* just 'happends' to be the same as we use */
		disc->last_session_state	 = (res[2] >> 2) &  3;		/* ditto */
		disc->recordable		 =  1;				/* XXX this might be proven wrong one day XXX */
		disc->rewritable		 =  res[2]       & 16;		/* erasable */
		disc->sequential		 =  disc->recordable && !(disc->rewritable);	/* XXX might be proven wrong one day XXX */
		disc->sector_size		 =  CD_SECTOR_LENGTH;
		disc->num_sessions		 =  res[4] | (res[ 9] << 8);
		first_track			 =  res[3];
		first_track_last_session	 =  res[5] | (res[10] << 8);	/* to build up the last 'session' */
		last_track_last_session  	 =  res[6] | (res[11] << 8);

		/* XXX MMC profile parsing for recordable, rewritable and sequential XXX */
	
		/* Initialise the sessions to be taken as overspanning tracks */
		for(session = 0; session < disc->num_sessions; session++) {
			disc->session_start[session] = INT_MAX;
			disc->session_end  [session] = 0;
		}

		for (track = first_track; track <= last_track_last_session; track++) {
			/* each track record is 36 bytes long */
			val_len = 36;
			bzero(res, val_len);

			bzero(cmd, SCSI_CMD_LEN);
			cmd[0] = 0x52;				/* Read track information */
			cmd[1] = 1;				/* indexed on track       */
			cmd[4] = track >> 8;			/* track number 0-0xff ?  */
			cmd[5] = track & 0xff;
			cmd[7] = val_len >> 8;			/* MSB length resultbuf	  */
			cmd[8] = val_len & 0xff;		/* LSB ,,		  */
			cmd[9] = 0;
			error = scsi_call(SCSI_READCMD, disc, cmd, 10, res, val_len);
			if (error) {
				perror("While reading track info");
				break;
			}
			data_length		= res[1] | (res[0]  << 8);
			// track_number		= res[2] | (res[32] << 8);	  /* why? */
			session			= res[3] | (res[33] << 8);
			// is_track_mode	= res[5] & 15;
			// is_copy		= res[5] & 16;
			// is_damage		= res[5] & 32;
			// is_fixed_packet	= res[6] & 16;
			// is_packet_or_inc	= res[6] & 32;
			// is_blank		= res[6] & 64;
			// is_reserved		= res[6] & 128;
			// is_data_mode		= res[6] & 15;
			nwa_valid		= res[7] & 1;
			// lra_valid		= res[7] & 2;
			track_start		= res[11] | (res[10]<<8) | (res[ 9]<<16) | (res[ 8]<<24);
			next_writable_addr	= res[15] | (res[14]<<8) | (res[13]<<16) | (res[12]<<24);
			free_blocks		= res[19] | (res[18]<<8) | (res[17]<<16) | (res[16]<<24);
			packet_size		= res[23] | (res[22]<<8) | (res[21]<<16) | (res[20]<<24);
			track_size		= res[27] | (res[26]<<8) | (res[25]<<16) | (res[24]<<24);
			// last_recorded_addr	= res[32] | (res[30]<<8) | (res[29]<<16) | (res[28]<<24);
			track_end = track_start + track_size;

			if (data_length <= 30) session &= 0xff;

			disc->session_start[session-1] = MIN(disc->session_start[session-1], track_start);
			disc->session_end  [session-1] = MAX(disc->session_end  [session-1], track_end);
			disc->free_blocks  [session-1] = free_blocks;
			disc->packet_size  [session-1] = packet_size;
			if (nwa_valid) disc->next_writable[session-1] = next_writable_addr;
		};
		if (disc->session_start[disc->num_sessions-1] == INT_MAX) {
			if (disc->disc_state == DISC_STATE_FULL) {
				if (disc->last_session_state == SESSION_STATE_COMPLETE) disc->num_sessions--;
			};
		};

		/* XXX
		 * initialise default link size to zero; only set DEFAULT
		 * link lossage for CD-R we are using `Burn Free' for writing
		 * but if it fails its good to know the penalty for recovery
		 * XXX
		 */
		disc->link_size = 0;
		if (disc->mmc_profile == 0x09) {
			disc->link_size = 7;
		};
		
		return 0;
	}

	/* Start with legacy CD-ROM .... trying to get as much info from it as possible */
	disc->recordable         = 0;
	disc->rewritable         = 0;
	disc->sequential	 = 0;
	disc->sector_size	 = CD_SECTOR_LENGTH;
	disc->disc_state         = DISC_STATE_FULL;
	disc->last_session_state = SESSION_STATE_COMPLETE;
	disc->link_size          = 7;	/* DEFAULT link lossage for CDs */


	/* Read session range */
	val_len = 4;			/* only head */
	bzero(res, val_len);
	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x43;			/* READ TOC/PMA/ATAP INFO	*/
	cmd[1] = 0;			/* no times please but LBA's	*/
	cmd[2] = 2;			/* format 2 also this time	*/
	cmd[6] = 0;			/* obligatory zero		*/
	cmd[7] = val_len >> 8;
	cmd[8] = val_len & 0xff;
	cmd[9] = 0;
	error = scsi_call(SCSI_READCMD, disc, cmd, 10, res, val_len);
	if (error) {
		perror("TOC reading of sessions failed");
		return error;
	}
	first_session = read_cd_hex2(res[2]);
	last_session  = read_cd_hex2(res[3]);

	disc->num_sessions = last_session - first_session + 1;

	/* Initialise the sessions to be taken as overspanning tracks */
	for(session = 0; session <= disc->num_sessions; session++) {
		disc->session_start[session] = INT_MAX;
		disc->session_end  [session] = 0;
	}

	/* calculate how big the result buffer ought to be to get the whole TOC */
	val_len = res[1] + (res[0] << 8) + 2;
	assert(val_len < res_len);

	/* Read the complete TOC and extract information */
	bzero(res, val_len);
	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x43;			/* READ TOC/PMA/ATAP INFO	*/
	cmd[1] = 0;			/* no times please but LBA's	*/
	cmd[2] = 2;			/* format 2; full TOC		*/
	cmd[6] = first_session;		/* start at first session	*/
	cmd[7] = val_len >> 8;
	cmd[8] = val_len & 0xff;
	cmd[9] = 0;
	error = scsi_call(SCSI_READCMD, disc, cmd, 10, res, val_len);
	if (error) {
		perror("TOC reading of full TOC failed");
		return error;
	}

	data_length =res[1] | (res[0] << 8);
	if (data_length != val_len) {
		fprintf(stderr, "Warning: device didn't return all data requested when reading full TOC\nn");
	};
	pos = 4;
	track_start = INT_MAX; track_end = 0;

	while (pos < data_length + 2) {
		session = read_cd_bcd(res[pos+ 0]);
		cntrl   = res[pos+1] & 15;
		addr    = res[pos+1] >> 4;
		tno     = read_cd_bcd(res[pos+ 2]);
		point   = read_cd_bcd(res[pos+ 3]);
		min     = read_cd_bcd(res[pos+ 4]);
		sec     = read_cd_bcd(res[pos+ 5]);
		frame   = read_cd_bcd(res[pos+ 6]);
		pmin    = read_cd_bcd(res[pos+ 8]);
		psec    = read_cd_bcd(res[pos+ 9]);
		pframe  = read_cd_bcd(res[pos+10]);

		/* extract information; explicit writeout. See SCSI docs */
		if (tno == 0 && session && addr == 1) {
			switch (point) {
				case 0xa0 : first_track = pmin; break;
				case 0xa1 : last_track  = pmin; break;
				case 0xa2 :
					track_end   = cd_msf2lba(pmin, psec, pframe);
					disc->session_end  [session-1] = MAX(disc->session_end  [session-1], track_end);
					break;
				default   :
					track_start = cd_msf2lba(pmin, psec, pframe);
					disc->session_start[session-1] = MIN(disc->session_start[session-1], track_start);
					break;
			};
		};
		if (tno == 0 && session && addr == 5) {
			if (point == 0xb0) {
				next_writable_addr = cd_msf2lba(min, sec, frame);
				disc_size = cd_msf2lba(pmin, psec, pframe);
				printf("UDF: ignoring B0 Q channel : next writeable address; pre MMC3 device; fix me\n");
			}

		};
		pos += 11;
	};

	/* Last session information is notoriously flawed in TOC format so update it */
	bzero(res, 8);
	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x25;		/* CD READ RECORDED CAPACITY */
	error = scsi_call(SCSI_READCMD, disc, cmd, 10, res, 8);
	if (error) {
		perror("Can't read CD recorded capacity; last session's length can be too big");
		return 0;
	};

	session = disc->num_sessions-1;
	disc->session_end[session] = res[3] | (res[2]<<8) | (res[1]<<16) | (res[0]<<24);
	return 0;
}


int udf_open_disc(char *dev_name, struct udf_discinfo **discptr) {
	struct udf_discinfo *disc;

	if (!discptr) return EINVAL;
	*discptr = NULL;

	/* determine what kind of file/device we are dealing with */
	disc = calloc(1, sizeof(struct udf_discinfo));
	if (!disc) return ENOMEM;

	/* fill in the name */
	disc->dev_name = strdup(dev_name);

	if (scsi_open(disc) != 0) {
		perror("Failure to open device or file");
		free(disc);
		return ENODEV;
	};

	/* return the pointer */
	*discptr = disc;

	/* get disc info */
	if (udf_get_disc_info(disc)) {
		fprintf(stderr, "Can't get disc info");
		scsi_close(disc);
		free(disc);
	};

	/* set recording parameters */
	udf_discinfo_set_recording_parameters(disc, 0);	/* no testwrite */

	/* initialise threads and associated conditional variables */
	pthread_mutex_init(&disc->rwthread_lock,   NULL);
	pthread_cond_init( &disc->rwthread_signal, NULL);

	pthread_mutex_init(&disc->processed_lock,  NULL);
	pthread_cond_init(&disc->processed_signal, NULL);

	/* XXX ... external to this module ... XXX */
	udf_start_disc_thread(disc);

	return 0;
}


int udf_close_disc(struct udf_discinfo *disc) {
	if (!disc) return 0;

	udf_stop_disc_thread(disc);
	scsi_close(disc);

	printf("Disc access statistics\n");
	printf("\tsector reads   %8d  (%d Kbyte)\n", disc->sectors_read,    (disc->sectors_read    * UDF_SECTOR_SIZE) / 1024);
	printf("\tsector written %8d  (%d Kbyte)\n", disc->sectors_written, (disc->sectors_written * UDF_SECTOR_SIZE) / 1024);
	printf("\tswitches       %8d\n", disc->switchings);

	return 0;
}


/******************************************************************************************
 *
 * Sector readers and writers
 *
 ******************************************************************************************/


/* read an extent of sectors in the `result' buffer */
int udf_read_physical_sectors(struct udf_discinfo *disc, off_t sector, uint32_t num_sectors, char *what, uint8_t *result) {
	scsicmd		cmd;
	uint32_t	size, size_read, sect_mult, chunk;
	int		error;

	/* read one UDF sector at a physical address specified in UDF_SECTOR size units */
	size = num_sectors * UDF_SECTOR_SIZE;
	size_read = 0;
	bzero(result, size);	/* just in case */		/* has to go? */

	/* XXX clean up XXX */
	assert(disc->sector_size);
	sect_mult = UDF_SECTOR_SIZE / disc->sector_size;
	sector      *= sect_mult;
	num_sectors *= sect_mult;
	assert(num_sectors * disc->sector_size == size);
	assert(num_sectors <= 0xffff);

	error = 0;
	DEBUG(printf("%08d : %s; read %d bytes\n", (int) sector, what, size));
	while (num_sectors) {
		switch (disc->devdrv_class & UDF_DEVDRV_CLASS) {
			case UDF_DEVDRV_CLASS_CD  :
			case UDF_DEVDRV_CLASS_DVD :
				/*
				 * Use a SCSI command to read it so we can go past the last session;
				 * allthough READ (12) is available, some older CD-ROM
				 * devices only want to do READ (10).
				 */

				/* limited by MAXPHYS, taking 64kb as limitation */
				chunk     = MIN(64*1024/UDF_SECTOR_SIZE, num_sectors);
				size_read = chunk * UDF_SECTOR_SIZE;	/* definition */

				bzero(cmd, SCSI_CMD_LEN);
				cmd[0] = 0x28;				/* READ (10) command		*/
				cmd[1] = 0;				/* normal access		*/
				cmd[2] = (sector >> 24) & 0xff;
				cmd[3] = (sector >> 16) & 0xff;
				cmd[4] = (sector >>  8) & 0xff;
				cmd[5] = (sector      ) & 0xff;
				cmd[6] = 0;				/* reserved */
				cmd[7] = (chunk >>  8) & 0xff;		/* MSB transfer */
				cmd[8] = (chunk      ) & 0xff;		/* number of logical block(s) */
				cmd[9] = 0;				/* control */
				error = scsi_call(SCSI_READCMD, disc, cmd, 10, result, size_read);
	
				if (error) return ENOENT;
				break;
			default :
				/* XXX first and only reference to {p}read() XXX */
				if (sector>=0)
					size_read = pread(disc->fhandle, result, (uint64_t) num_sectors*UDF_SECTOR_SIZE, sector*UDF_SECTOR_SIZE);
				break;
		};
		/* statistics */
		disc->sectors_read += size_read / UDF_SECTOR_SIZE;
		if (disc->am_writing) disc->switchings++;
		disc->am_writing = 0;

		/* advance */
		num_sectors -= size_read / UDF_SECTOR_SIZE;
		sector      += size_read / UDF_SECTOR_SIZE;
		result      += size_read;
		if (size_read <= 0) {
			UDF_VERBOSE_MAX(
				if (what) {
					printf("Can't read sectors %d+%d for %s\n", (int) sector, num_sectors, what);
				}
			);
			if (size_read == 0) return ENOENT;
			return error;
		};
	};

	return 0; /* flag ok */
}


/* write an extent of sectors to disc */
int udf_write_physical_sectors(struct udf_discinfo *disc, off_t sector, uint32_t num_sectors, char *what, uint8_t *source) {
	scsicmd    cmd;
	uint32_t   trans_length, size, size_written, sect_mult, chunk;
	int        error;

//	if (!disc->udf_recording) return ENODEV;

	/* write one UDF sector at a physical address specified in UDF_SECTOR size units */
	size = num_sectors * UDF_SECTOR_SIZE;
	size_written = 0;

	/* XXX clean up XXX */
	assert(disc->sector_size);
	sect_mult = UDF_SECTOR_SIZE / disc->sector_size;
	sector      *= sect_mult;
	num_sectors *= sect_mult;
	assert(num_sectors * disc->sector_size == size);
	assert(num_sectors <= 0xffff);					/* compatible with WRITE (10)? */

	DEBUG(printf("%08d : %s ;WRITE %d bytes\n", (int) sector, what, size));

	error = 0;
	while (num_sectors) {
		switch (disc->devdrv_class & UDF_DEVDRV_CLASS) {
			case UDF_DEVDRV_CLASS_CD  :
			case UDF_DEVDRV_CLASS_DVD :
				/*
				 * Use WRITE (12) command to write to the disc; we
				 * might have to downgrade later to using WRITE (10)
				 * on older discs though :-/
				 */

				/* limited by MAXPHYS, taking 64kb as limitation */
				chunk        = MIN(64*1024/UDF_SECTOR_SIZE, num_sectors);	/* in sectors */
				trans_length = chunk;						/* in sectors */
				bzero(cmd, SCSI_CMD_LEN);
				cmd[ 0] = 0xAA;				/* WRITE (12)	*/
				cmd[ 1] = 0;				/* no force unit access */
				cmd[ 2] = (sector       >> 24) & 0xff;
				cmd[ 3] = (sector       >> 16) & 0xff;
				cmd[ 4] = (sector       >>  8) & 0xff;
				cmd[ 5] = (sector            ) & 0xff;
				cmd[ 6] = (trans_length >> 24) & 0xff;	/* MSB */
				cmd[ 7] = (trans_length >> 16) & 0xff;
				cmd[ 8] = (trans_length >>  8) & 0xff;
				cmd[ 9] = (trans_length      ) & 0xff;	/* LSB  */
				cmd[10] = 0;				/* no streaming */
				cmd[11] = 0;				/* control */
				error = scsi_call(SCSI_WRITECMD, disc, cmd, 12, source, trans_length*UDF_SECTOR_SIZE);
	
				size_written = trans_length*UDF_SECTOR_SIZE;			/* sectors->bytes */
				if (error) size_written = 0;
				break;
			default :
				/* XXX first and only reference to {p}write() XXX */
				DEBUG(printf("udf_discop: pwrite %"PRIu64" + %d\n", (uint64_t) sector*UDF_SECTOR_SIZE, (int) num_sectors*UDF_SECTOR_SIZE));
				size_written = pwrite(disc->fhandle, source, (uint64_t) num_sectors*UDF_SECTOR_SIZE, sector*UDF_SECTOR_SIZE);
				break;
		};
		/* statistics */
		disc->sectors_written += size_written / UDF_SECTOR_SIZE;
		if (!disc->am_writing) disc->switchings++;
		disc->am_writing = 1;

		num_sectors -= size_written / UDF_SECTOR_SIZE;
		sector      += size_written / UDF_SECTOR_SIZE;
		source      += size_written;
		if ((size_written < size) || error) {
			DEBUG(if (error) printf("Writing %s at sectors %d+%d failed\n", what, (int) sector, num_sectors));
			return EIO;
		};
	};

	return 0;
}


/* End of udf_discinfo.c */

