/* $NetBSD$ */

/*
 * File "udf_readwrite.c" is part of the UDFclient toolkit.
 * File $Id: udf_readwrite.c,v 1.27 2005/01/07 17:53:08 imago Exp $ $Name:  $
 *
 * Copyright (c) 2003, 2004 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udftoolkit/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/* XXX strip list to bare minimum XXX */
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>
#include <dirent.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <time.h>

#include "scsi_sense.h"
#include "scsilib.h"


/* for locals */
#include "udf.h"
#include "udf_bswap.h"
#include "udf_discop.h"
#include "uio.h"
#include <pthread.h>


#ifndef MAX
#	define MAX(a,b) ((a)>(b)?(a):(b))
#	define MIN(a,b) ((a)<(b)?(a):(b))
#endif


//#define DEBUG(a) { a; };
#define DEBUG(a) if (0) { a; };


/******************************************************************************************
 *
 * Sector readers and writers 
 *
 ******************************************************************************************/


/* read sectors at a physical address using session offsets */
int udf_read_session_sectors(uint8_t *result, struct udf_session *udf_session, off_t sector, uint32_t num_sectors, char *what) {
	/*
	 * read_physical_sectors will read any extent even when its too big
	 * for the transportlayer
	 */
	if (((long) result) & 3) {
		printf("Unaligned read of sector : possible panic() on some systems avoided\n");
		return EIO;
	};

	return udf_read_physical_sectors(udf_session->disc, sector + udf_session->session_offset, num_sectors, what, result);
}



/* try to write sectors but session local independent */
int udf_write_session_sectors(struct udf_session *udf_session, off_t sector, uint32_t num_sectors, char *what, uint8_t *source) {
	/*
	 * write_physical_sectors will write any extent even when its too big
	 * for the transportlayer
	 */
	if (((long) source) & 3) {
		printf("Unaligned write of sector : possible panic() on some systems avoided\n");
		return EIO;
	};

	return udf_write_physical_sectors(udf_session->disc, sector + udf_session->session_offset, num_sectors, what, source);
}


/******************************************************************************************
 *
 * Start of cache reading/writing functions
 *
 ******************************************************************************************/

int udf_init_session_caches(struct udf_session *udf_session) {
	UDF_MUTEX_INIT(&udf_session->session_cache_lock);

	TAILQ_INIT(&udf_session->data_cache.lru_chain);
	TAILQ_INIT(&udf_session->data_cache.cache_lines);

	UDF_MUTEX_INIT(&udf_session->data_cache.cache_lock);
	udf_session->data_cache.max_lru_length  = UDF_DCACHE_READC_LENGTH;
	udf_session->data_cache.lru_length      = 0;
	udf_session->cache_line_read  = malloc(UDF_READWRITE_LINE_LENGTH * UDF_SECTOR_SIZE);
	udf_session->cache_line_write = malloc(UDF_READWRITE_LINE_LENGTH * UDF_SECTOR_SIZE);
	assert(udf_session->cache_line_read);
	assert(udf_session->cache_line_write);

	return 0;
}


void udf_cache_refered_to_entry(struct udf_cache *cache, struct udf_cache_entry *element) {
	/* only update it if its in the LRU chain */
	UDF_MUTEX_LOCK(&cache->cache_lock);

	TAILQ_REMOVE(&cache->lru_chain, element, next_lru);
	TAILQ_INSERT_TAIL(&cache->lru_chain, element, next_lru);

	UDF_MUTEX_UNLOCK(&cache->cache_lock);
}


/* allways gets us a cache line; and its returned back locked */
int udf_lookup_cache_line(struct udf_session *udf_session, uint32_t sector, struct udf_cache_line **lkd_cache_line) {
	struct udf_cache	*cache;
	struct udf_cache_line	*cache_line, *first_line, *last_line, *prev_line;
	uint32_t		 line_start;

	line_start = UDF_READWRITE_LINE_LENGTH * (sector / UDF_READWRITE_LINE_LENGTH);

	cache = &udf_session->data_cache;

	/* check for presence of cache line */
	/* lock queue to prevent it from changing beneath us */
	UDF_MUTEX_LOCK(&cache->cache_lock);

	TAILQ_FOREACH(cache_line, &cache->cache_lines, next_line) {
		if (cache_line->line_start == line_start) break;
	};

	if (!cache_line) {
		/* try to recycle one */
		TAILQ_FOREACH(cache_line, &cache->cache_lines, next_line) {
			if (!(cache_line->present || cache_line->request_read || cache_line->cache_line_lock.locked)) {
				TAILQ_REMOVE(&cache->cache_lines, cache_line, next_line);
				break;
			};
		};

		if (!cache_line) {
			/* get us a new cache line */
			cache_line = calloc(1, sizeof(struct udf_cache_line)); assert(cache_line);
			UDF_MUTEX_INIT(&cache_line->cache_line_lock);
		};

		/* init cache line */
		cache_line->line_start = line_start;
		first_line = last_line = NULL;

		/* try to get the first and the last cache line */
		first_line = TAILQ_FIRST(&cache->cache_lines);
		if (first_line) last_line  = TAILQ_LAST(&cache->cache_lines, udf_cache_line_chain);

		/* insert cacheline in sorted list */
		if (!first_line || (first_line && (first_line->line_start > line_start))) {
			TAILQ_INSERT_HEAD(&cache->cache_lines, cache_line, next_line);
		} else {
			assert(last_line);
			if (last_line->line_start > line_start) {
				TAILQ_FOREACH_REVERSE(prev_line, &cache->cache_lines, udf_cache_line_chain, next_line) {
					if (prev_line->line_start < line_start) break;
				};
				assert(prev_line);
				TAILQ_INSERT_AFTER(&cache->cache_lines, prev_line, cache_line, next_line);
			} else {
				TAILQ_INSERT_TAIL(&cache->cache_lines, cache_line, next_line);
			};
		};
	};

	/* we don't need the lock anymore; we are not modifying the queue anymore */
	UDF_MUTEX_UNLOCK(&cache->cache_lock);
/*XXX*/	UDF_MUTEX_LOCK(&cache_line->cache_line_lock);

	*lkd_cache_line = cache_line;

	return 0;
}


struct udf_cache_entry *udf_recycle_cache_entry(struct udf_cache *cache) {
	struct udf_cache_line  *cache_line;
	struct udf_cache_entry *entry;
	uint64_t bit;

	/* lock the whole cache to prevent concurrency problems */
	UDF_MUTEX_LOCK(&cache->cache_lock);

	entry = NULL;
	if (cache->lru_length >= cache->max_lru_length) {
		TAILQ_FOREACH(entry, &cache->lru_chain, next_lru) {
			/* a cache entry is free for reuse iff :
			 *   - its present
			 *   - its not modified
			 *   - we are not recycling from a line that needs readin
			 */
			cache_line  = entry->cache_line;

			if (!(cache_line->request_read) && !(cache_line->modified) && (!cache_line->cache_line_lock.locked)) {
				/* let it disappear */
/* XXX */			UDF_MUTEX_LOCK(&cache_line->cache_line_lock);
					bit         = ((uint64_t) 1 << entry->line_offset);

					assert(cache_line->present & bit);
					cache_line->present = cache_line->present & (~bit);
					cache_line->entries[entry->line_offset] = NULL;
					TAILQ_REMOVE(&cache->lru_chain, entry, next_lru);
					cache->lru_length--;
				UDF_MUTEX_UNLOCK(&cache_line->cache_line_lock);

				/* wipe */
				bzero(entry, sizeof(struct udf_cache_entry) + (size_t) UDF_SECTOR_SIZE);
				break;	/* break off lru_loop */
			};
		}; /* lru loop */
	} else {
		entry = calloc(1, sizeof(struct udf_cache_entry) + (size_t) UDF_SECTOR_SIZE);
		assert(entry);
	};

	/* return result (can be NULL) */
	UDF_MUTEX_UNLOCK(&cache->cache_lock);

	return entry;
}


/* sorry for the extra memcpy */
int udf_read_session_cache_sector(struct udf_session *udf_session, uint32_t sector, char *what, uint8_t *result_blob, int flags) {
	struct udf_discinfo	*disc;
	struct udf_cache_line	*cache_line;
	struct udf_cache_entry	*entry;
	uint32_t bit;
	uint32_t line_offset;

	assert(result_blob);

	disc = udf_session->disc;
	udf_lookup_cache_line(udf_session, sector, &cache_line);
	assert(cache_line);

	/*
	 * do whatever is nessisary; i.e. mark stuff that needs to be read in;
	 * we hold the cache line lock
	 */

	line_offset = sector - cache_line->line_start;
	bit     = (1 << line_offset);
	if (!(cache_line->present & bit)) {
		/* Please ask the r/w-thread to read the requested sector */
		cache_line->request_read |= bit;

		/* release lock on cache line so r/w-thread can access it */
		UDF_MUTEX_UNLOCK(&cache_line->cache_line_lock);

		while (!(cache_line->present & bit)) {
			/* signal a request is made */
			disc->rwthread_kicked = 1;
			pthread_cond_signal(&disc->rwthread_signal);

			/* and wait for a response */
			pthread_mutex_lock(&disc->processed_lock);
			pthread_cond_wait(&disc->processed_signal, &disc->processed_lock);
			pthread_mutex_unlock(&disc->processed_lock);
		};

		/* re-lock the cache line */
		UDF_MUTEX_LOCK(&cache_line->cache_line_lock);

		cache_line->request_read = cache_line->request_read & (~bit);
	};
	assert(cache_line->present & bit);

	/* it's here now, return the sector; it only goes OK by `refering to it' */
	entry = cache_line->entries[line_offset];
	assert(entry);
	entry->what = what;

	memcpy(result_blob, entry->buffer, UDF_SECTOR_SIZE);

	udf_cache_refered_to_entry(&udf_session->data_cache, entry);

	/* release lock on cache line */
	UDF_MUTEX_UNLOCK(&cache_line->cache_line_lock);

	return 0;
}


int udf_write_session_cache_sector(struct udf_session *udf_session, uint32_t sector, char *what, uint8_t *source, int flags, struct udf_wrcallback *wrcallback) {
	struct udf_discinfo	*disc;
	struct udf_cache_line	*cache_line;
	struct udf_cache_entry	*entry;
	struct udf_cache        *cache;
	uint64_t bit;
	uint32_t line_offset;

	disc = udf_session->disc;
	cache = &udf_session->data_cache;

	udf_lookup_cache_line(udf_session, sector, &cache_line);
	assert(cache_line);

	/*
	 * do whatever is nessisary; get us an entry to write in and if there
	 * is no space entry left, wait for the new entries.
	 * we hold the cache line lock
	 */

	line_offset = sector - cache_line->line_start;
	bit   = ((uint64_t) 1 << line_offset);
	entry = cache_line->entries[line_offset];
	if (cache_line->present & bit) {
		if (cache_line->modified & bit) {
			/* call anulate */
			entry = cache_line->entries[line_offset];
			assert(entry);

			if (entry->wrcallback.function) {
				(entry->wrcallback.function)(UDF_WRCALLBACK_REASON_ANULATE, &entry->wrcallback, 0, entry->buffer);
			};
		};
	};

	if (!entry) {
		/* not here... hmm... get an entry; this can block... */
		do {
			entry = udf_recycle_cache_entry(&udf_session->data_cache);
			if (!entry) {
				/* release lock on cache line so r/w-thread can access it */
				UDF_MUTEX_UNLOCK(&cache_line->cache_line_lock);

				/* wait for new entries */
				disc->rwthread_kicked = 1;
				pthread_cond_signal(&disc->rwthread_signal);

				/* wait for a response and directly re-lock the cache line before unlocking the lock*/
				pthread_mutex_lock(&disc->processed_lock);
				pthread_cond_wait(&disc->processed_signal, &disc->processed_lock);

				UDF_MUTEX_LOCK(&cache_line->cache_line_lock);
				pthread_mutex_unlock(&disc->processed_lock);
			};
		} while (!entry);

		/* Bugalert:
		 * Due to read-in requests and/or fill-up reading, look if the
		 * reqested sector hasn't `popped-up' suddenly by the
		 * read/write thread. If it does, and its entry is not equal
		 * to the entry we have, free our now obsolete entry. Note
		 * that the entry we find CAN't be marked `modified'.
		 * XXX what to do with multiple writers... XXX
		 */

		if (entry && (cache_line->present & bit)) {
			if (entry != cache_line->entries[line_offset]) free(entry);

			entry = cache_line->entries[line_offset];
			assert(entry);
			assert(!(cache_line->modified & bit));
		} else {
			UDF_MUTEX_LOCK(&cache->cache_lock);
				TAILQ_INSERT_TAIL(&cache->lru_chain, entry, next_lru);
				cache->lru_length++;
			UDF_MUTEX_UNLOCK(&cache->cache_lock);
		};
	};

	/* got a cache_entry; fill in */
	UDF_MUTEX_LOCK(&cache->cache_lock);
		entry->cache_line  = cache_line;
		entry->what        = what;
		entry->line_offset = line_offset;
		memcpy(entry->buffer, source, UDF_SECTOR_SIZE);
		if (wrcallback) memcpy(&entry->wrcallback, wrcallback, sizeof(struct udf_wrcallback));

		cache_line->entries[line_offset] = entry;
		cache_line->present  |= bit;
		cache_line->modified |= bit;
	UDF_MUTEX_UNLOCK(&cache->cache_lock);

	udf_cache_refered_to_entry(&udf_session->data_cache, entry);

	if (entry->wrcallback.function) {
		(entry->wrcallback.function)(UDF_WRCALLBACK_REASON_PENDING, &entry->wrcallback, 0, entry->buffer);
	};

	/* release lock on cache line */
	UDF_MUTEX_UNLOCK(&cache_line->cache_line_lock);

	return 0;
}


/*
 * All reading and writing including `fill-up' reading is done in the
 * specialised read/write-thread; this is to prevent intermixing of readers
 * and writers. All reading will thus be clustered as will all writing by
 * alternating between a read- and a write-phase.
 * 
 * In the read-phase, fill-up of incomplete dirty packets is done so no
 * reading in is required while in the write-phase. This significantly
 * improves writing speed since switching between te modes is extremely
 * expensive in terms of time. A `normal' write-seek time for a DVD drive is
 * about 500 ms(!) compared to a harddisc's 10-15 ms.
 *
 * Note that fillup-reading is for rewritables only; recordables are not
 * supposed to have such gaps and ought to be filled in with `zero' blocks
 */


/* returns nonzero if there is still work to do */
int udf_readwriter_process_session(struct udf_session *udf_session) {
	struct udf_discinfo    *disc;
	struct udf_cache       *cache;
	struct udf_cache_line  *cache_line;
	struct udf_cache_entry *cache_entry;
	uint64_t  all_present, readin, bit;
	uint32_t  cache_sector, offset, num_sectors;
	uint8_t  *cache_line_read, *cache_line_write, *sector;
	int32_t   num_write_lines;
	int       error, again;

	disc  = udf_session->disc;
	cache = &udf_session->data_cache;

	again = 0;	/* we ought to be finished after one pass */

	/* READ in */
	all_present = ((uint64_t) 1 << UDF_READWRITE_LINE_LENGTH) - 1;

	TAILQ_FOREACH(cache_line, &cache->cache_lines, next_line) {
		if ((cache_line->present != all_present) && ((cache_line->request_read) || (cache_line->modified))) {
			UDF_MUTEX_LOCK(&cache_line->cache_line_lock);

			/* read in missing or requested sectors */
			cache_sector    = cache_line->line_start;
			num_sectors     = UDF_READWRITE_LINE_LENGTH;
			cache_line_read = udf_session->cache_line_read;	 assert(cache_line_read);

			/* try to read a cacheline in one go, if it fails, keep a holdback time */
DEBUG(printf("read cacheline from %08d + %d\n", cache_sector, num_sectors));
			error  = cache_line->error;
			if (!error) {
				/* only read in complete line if there was no error the last time we read in this cache line */
				/* TODO create a flag that disables cache line reading */
				error  = udf_read_session_sectors(cache_line_read, udf_session, cache_sector, num_sectors, "cache line");
			};
			readin = all_present;
			if (error) {
				DEBUG(printf("Failed to read in cache line\n"));
DEBUG(printf("Failed to read in cache line\n"));
				readin = 0;
				for (offset = 0; offset < UDF_READWRITE_LINE_LENGTH; offset++) {
					bit = ((uint64_t) 1 << offset);
					if (cache_line->request_read & bit) {
						/* fall back to reading one sector at a time */
						assert(!(cache_line->modified & bit));
						sector = cache_line_read + UDF_SECTOR_SIZE * offset;
						error = udf_read_session_sectors(sector, udf_session, cache_sector + offset, 1, "cache sector");
						if (error) {
							DEBUG(printf("Failed to read in sector %d (offset %d)\n", (uint32_t) cache_sector + offset, offset));
							cache_line->error |= bit;
							bzero(sector, UDF_SECTOR_SIZE);
						};
						readin |= bit;
					};
				};
			};

DEBUG(printf("Processing "));
			/* use ``readin'' to see what we've actually read in */
			for (offset = 0; offset < UDF_READWRITE_LINE_LENGTH; offset++) {
				bit = ((uint64_t) 1 << offset);
				/* NEVER overwrite present sectors */
				if (!(cache_line->present & bit) && (readin & bit)) {
					sector = cache_line_read + UDF_SECTOR_SIZE * offset;

					/* get us a cache entry */
					cache_entry = udf_recycle_cache_entry(cache);
					if (!cache_entry) {
						/* no recycle candidate found or just not enough entries yet */
						cache_entry = calloc(1, sizeof(struct udf_cache_entry) + (size_t) UDF_SECTOR_SIZE);
						assert(cache_entry);
					};

					UDF_MUTEX_LOCK(&cache->cache_lock);
						memcpy(cache_entry->buffer, sector, UDF_SECTOR_SIZE);
						cache_entry->cache_line      = cache_line;
						cache_entry->what            = "unknown contents";
						cache_entry->line_offset     = offset;

						cache_line->entries[offset]  = cache_entry;
						cache_line->present |= bit;

						TAILQ_INSERT_TAIL(&cache->lru_chain, cache_entry, next_lru);
						cache->lru_length++;
					UDF_MUTEX_UNLOCK(&cache->cache_lock);
				};
			};

			/* signal there is new space available */
			pthread_cond_signal(&disc->processed_signal);
			UDF_MUTEX_UNLOCK(&cache_line->cache_line_lock);
DEBUG(printf("\n"));
		};
	};
DEBUG(printf("UDF readwriter : READ loop end\n"));


pthread_cond_signal(&disc->processed_signal);


	num_write_lines = 0;
	again      = 0;
	TAILQ_FOREACH(cache_line, &cache->cache_lines, next_line) {
		if ((cache_line->present == all_present) && (cache_line->modified)) {
			num_write_lines++;
			again = 1;
		};
		if ((cache_line->request_read & ~cache_line->present)) {
			again = 1;
		};
	};

	if (!disc->finish_rw_thread) {
		/* check if we could wait for more data */
		if (num_write_lines < 2 /* UDF_DCACHE_WRITEP_TRESHOLD */) return 0;
	};

	/* WRITE out */
	TAILQ_FOREACH(cache_line, &cache->cache_lines, next_line) {
		if (cache_line->modified && cache_line->present != all_present) {
			again = 1;	/* we are not finished yet */
			continue;		/* cache_line */
		};
		if (cache_line->modified) {
			UDF_MUTEX_LOCK(&cache_line->cache_line_lock);

			assert((cache_line->present | cache_line->error) == all_present);
			cache_sector    = cache_line->line_start;
			num_sectors     = UDF_READWRITE_LINE_LENGTH;
			cache_line_write = udf_session->cache_line_write;	 assert(cache_line_write);

			/* build up cache line  */
			for (offset = 0; offset < UDF_READWRITE_LINE_LENGTH; offset++) {
				bit    = ((uint64_t) 1 << offset);
				sector = cache_line_write + UDF_SECTOR_SIZE * offset;
				cache_entry = cache_line->entries[offset];
if (!(cache_line->present & bit)) {
	printf("bit %d clear (entry=%p)\n", offset, cache_entry);
	abort();
};

				if (cache_line->error & bit) {
					bzero(sector, UDF_SECTOR_SIZE);
				} else {
					memcpy(sector, cache_entry->buffer, UDF_SECTOR_SIZE);
				};
			};

DEBUG(printf("UDF readwriter : writing cache line %08d + %d\n", cache_line->line_start, UDF_READWRITE_LINE_LENGTH));
			/* write out cache line */
			error = udf_write_session_sectors(udf_session, cache_sector, num_sectors, "cache line", cache_line_write);
			for (offset = 0; offset < UDF_READWRITE_LINE_LENGTH; offset++) {
				bit = ((uint64_t) 1 << offset);
				cache_entry = cache_line->entries[offset];
				if ((cache_line->modified & bit) && cache_entry->wrcallback.function)
					(cache_entry->wrcallback.function)(UDF_WRCALLBACK_REASON_WRITTEN, &cache_entry->wrcallback, error, cache_entry->buffer);
			};
DEBUG(printf("UDF readwriter : fin writing cache line and upcalls\n"));
			if (!error) cache_line->modified = 0;

			UDF_MUTEX_UNLOCK(&cache_line->cache_line_lock);
		};

		/* signal there is new space available */
		pthread_cond_signal(&disc->processed_signal);
	};
DEBUG(printf("UDF readwriter : WRITE loop end\n"));

	/* signal there is new space available */
	pthread_cond_signal(&disc->processed_signal);

	return again;
}


int udf_sync_caches(struct udf_log_vol *udf_log_vol) {
	/* TODO Virtual caches sync etc; what about the session caches ? */
	return 0;
}


void *udf_readwriter(void *arg) {
	struct udf_discinfo	*disc = (struct udf_discinfo *) arg;
	struct udf_session	*udf_session;
	int finished;

	printf("UDF read/write thread initialising for file/disc/device `%s'\n", disc->dev_name);
	while(1) {
		DEBUG(printf("UDF read/write thread: waiting for lock\n"));
		/*
		 * If we are not asked to finished up our writing, block to
		 * wait for more data. Signal the reader to continue just in
		 * case it is still stuck.
		 */
		if (!disc->finish_rw_thread) {
			do {
				/* ask for more requests */
				pthread_cond_signal(&disc->processed_signal);
				pthread_mutex_lock(&disc->rwthread_lock);
				pthread_cond_wait(&disc->rwthread_signal, &disc->rwthread_lock);
				pthread_mutex_unlock(&disc->rwthread_lock);
				if (!disc->rwthread_kicked)
					printf("\n\nIEEE!!!!!!! read/write thread condwait woke up without being kicked\n\n");
			} while (!disc->rwthread_kicked && !disc->finish_rw_thread);
		};
		disc->rwthread_kicked = 0;

		DEBUG(printf("UDF read/write thread: got activate\n"));

		/* now look in all udf_sessions; lock sessions list (?!) */
		do {
			finished = 1;
			/* multiple writable sessions? */
			STAILQ_FOREACH(udf_session, &disc->sessions, next_session) {
				/* check if there are cache lines waiting to be written out */
				DEBUG(
					printf("UDF read/write thread: processing file/disc/device `%s', udf_session %d\n",
						disc->dev_name, udf_session->session_num);
				);
				if (udf_readwriter_process_session(udf_session)) finished = 0;
			};
		} while (!finished);

		/* if asked to quit break out of the loop */
		if (disc->finish_rw_thread) break;
	};

	printf("UDF read/write thread exiting\n");
	pthread_exit(0);	/* join */

	/* not reached */
	return NULL;
}


int udf_start_disc_thread(struct udf_discinfo *disc) {
	/*
	 * start up read/write thread and crudely kick it into
	 * existence.
	 */
	DEBUG(printf("Starting up read/write thread\n"));

	disc->thread_active = 1;
	pthread_create(&disc->rwthread_id, NULL, udf_readwriter, disc);
	sleep(1);

	DEBUG(printf("\n\n"));
	return 0;
}


int udf_stop_disc_thread(struct udf_discinfo *disc) {
	/* stop all associated threads */
	UDF_VERBOSE(printf("Stopping read/write thread\n"));

	if (disc->recordable) {
		disc->rwthread_kicked  = 1;
		disc->finish_rw_thread = 1;
		pthread_cond_signal(&disc->rwthread_signal);
		pthread_join(disc->rwthread_id, NULL);			/* wait for join */
	};

	UDF_VERBOSE(printf("\n\n"));

	disc->thread_active = 0;
	return 0;
}


int udf_writequeue_kick(struct udf_session *udf_session, char *why) {
	/*
	 * Kick the writequeue into existence in case its not active and wait
	 * for it to signal there is space left. We could set a variable in
	 * the writequeue to make this unnessisary but first try this out.
	 */

	DEBUG(printf("\nwritequeue kick! because of %s\n", why));
	udf_session->disc->rwthread_kicked = 1;
	pthread_cond_signal(&udf_session->disc->rwthread_signal);

	return 0;
}


/******************************************************************************************
 *
 * Start of uio based reading/writing functions
 *
 ******************************************************************************************/


/* sector is in UDF_SECTOR_SIZE units */
int udf_read_session_bytes_uio(struct udf_session *udf_session, uint32_t sector, uint32_t offset, uint64_t length, char *what, struct uio *uio, int flags) {
	uint8_t	   *some_space;
	int	    error = 0;
	uint32_t    copied;

	/* XXX malloc .... XXX */
	some_space = malloc(UDF_SECTOR_SIZE);
	if (!some_space) return ENOMEM;

	while (length) {
		copied = 0;
		UDF_MUTEX_LOCK(&udf_session->session_cache_lock);
			error = udf_read_session_cache_sector(udf_session, sector, what, some_space, flags);
			if (!error) {
				copied = MIN(length, UDF_SECTOR_SIZE-offset);
				uiomove(some_space + offset, copied, uio);
			};
		UDF_MUTEX_UNLOCK(&udf_session->session_cache_lock);
		if (error) return error;

		length -= copied;
		offset  = 0;
		sector++;
	};

	return error;
}


int udf_write_session_bytes_uio(struct udf_session *udf_session, uint32_t sector, uint32_t offset, uint64_t length, char *what, struct uio *uio, int flags, struct udf_wrcallback *wrcallback) {
	uint8_t	   *some_space;
	int	    error = 0;
	uint32_t    copied;

	/* XXX malloc .... XXX */
	some_space = malloc(UDF_SECTOR_SIZE);
	if (!some_space) return ENOMEM;

	/* partial written sectors need to be read in to be updated unless `flags' tell different */
	if (offset > 0) {
		copied = 0;
		UDF_MUTEX_LOCK(&udf_session->session_cache_lock);
			/* read in start sector */
			error = udf_read_session_cache_sector(udf_session, sector, what, some_space, flags);

			if (!error) {
				/* copy in old contents and write packaged data in it */
				copied = MIN(length, UDF_SECTOR_SIZE-offset);
				uiomove(some_space + offset, copied, uio);

				wrcallback->length = copied;
				error = udf_write_session_cache_sector(udf_session, sector, what, some_space, flags, wrcallback);
			};
		UDF_MUTEX_UNLOCK(&udf_session->session_cache_lock);
		if (error) {
			free(some_space);
			return error;
		};

		/* advance */
		wrcallback->offset += copied;
		length -= copied;
		sector++;
	};

	/* mid part */
	if (length >= UDF_SECTOR_SIZE) {
		while (length >= UDF_SECTOR_SIZE) {
			uiomove(some_space, UDF_SECTOR_SIZE, uio);

			wrcallback->length  = UDF_SECTOR_SIZE;
			error = udf_write_session_cache_sector(udf_session, sector, what, some_space, flags | UDF_CACHEFLAG_NOREADIN, wrcallback);
			if (error) {
				free(some_space);
				return error;
			};

			wrcallback->offset += UDF_SECTOR_SIZE;
			length -= UDF_SECTOR_SIZE;
			sector++;
		};
	};
	/* partial written sectors need to be read in to be updated unless `flags' tell different */
	if (length > 0) {
		DEBUG(printf("\nnon sector boundary extent : flags = %d\n", flags));
		copied = 0;
		UDF_MUTEX_LOCK(&udf_session->session_cache_lock);
			error = 0;
			/* read in start sector */
			if ((flags & UDF_CACHEFLAG_LASTSECTOR) == 0) {
				/*
				 * This rare event happens when we
				 * write/update a part of a directory `file'
				 * that is not in cache anymore.
				 *
				 * TODO: Avoid this completely when using recordable media
				 */
				error = udf_read_session_cache_sector(udf_session, sector, what, some_space, flags);
			} else {
				bzero(some_space, UDF_SECTOR_SIZE);
			};
			if (!error) {
				/* copy in */
				copied = length;
				uiomove(some_space, length, uio);

				/*
				 * If this extent contains the last sector in the blob to be
				 * written, then reading in the last sector for the possible
				 * remaining bytes is not nessisary. It'll only slow down the
				 * writing process considerably.
				 */
				if (flags & UDF_CACHEFLAG_LASTSECTOR) flags |= UDF_CACHEFLAG_NOREADIN;

				wrcallback->length = length;
				error = udf_write_session_cache_sector(udf_session, sector, what, some_space, flags, wrcallback);
			};
		UDF_MUTEX_UNLOCK(&udf_session->session_cache_lock);
		if (error) {
			free(some_space);
			return error;
		};

		/* advance */
		wrcallback->offset += copied;
		length -= copied;
		sector++;
	};
	free(some_space);

	return 0;
};


int udf_read_logvol_bytes_uio(struct udf_log_vol *udf_log_vol, uint32_t vpart_num, uint32_t lb_num, uint64_t offset, uint64_t length, char *what, struct uio *uio, int flags) {
	struct udf_partition	 *udf_partition;
	struct udf_part_mapping  *udf_part_mapping;
	struct udf_session	 *udf_session;
	struct udf_queue_entry	 *mcentry;
	uint64_t 		  ses_addr, trans_valid_len;
	uint64_t 		  trans_length;
	uint32_t		  lb_size;
	uint32_t		  ses_sector, ses_offset;
	int			  error, resolved;

DEBUG(
	printf("Read logvol space for %s, from vpart %d, lb_num %d at offset %d, for %d bytes\n",
		what, vpart_num, (int)lb_num, (int) offset, (int) length);
);

	error = udf_logvol_vpart_to_partition(udf_log_vol, vpart_num, &udf_part_mapping, &udf_partition);
	if (error) return error;

	/* get the offset (in bytes) in the partition */
	lb_size     = udf_rw32(udf_log_vol->log_vol->lb_size);
	offset     += (uint64_t) lb_num * lb_size;

	/* Virtual partitions have a write-back buffer cache; it holds a list of virtual sectors */
	/* Its more a still-to-allocate virtual cache to allow easy recording on CD-R/DVD-R/DVD+R etc */
	udf_session = udf_partition->udf_session;
	do {
		trans_length = length;
		resolved     = 0;
		mcentry      = NULL;
		ses_sector   = 0;

		/* no use if minicache/writequeue */
		error = udf_translate_addr(udf_part_mapping, udf_partition, offset, &ses_addr, &trans_valid_len);
		if (error) break;
		ses_sector = ses_addr / UDF_SECTOR_SIZE;
		ses_offset = ses_addr % UDF_SECTOR_SIZE;

		/* determine the translated address and its translation validity length */
		trans_length = MIN(length, trans_valid_len);
		error = udf_read_session_bytes_uio(udf_session, ses_sector, ses_offset, trans_length, what, uio, flags);
		if (error) break;

		offset	  += trans_length;
		length    -= trans_length;

		if (length == 0) return error;
	} while (length && !error);

	return EFAULT;
}


int udf_write_logvol_bytes_uio(struct udf_log_vol *udf_log_vol, uint32_t vpart_num, uint32_t lb_num, uint64_t offset, uint64_t length, char *what, struct uio *uio, int flags, struct udf_wrcallback *wrcallback) {
	struct udf_partition	 *udf_partition;
	struct udf_part_mapping  *udf_part_mapping;
	struct udf_session	 *udf_session;
	uint64_t 		  ses_addr, trans_valid_len;
	uint64_t 		  trans_length;
	uint32_t		  lb_size;
	uint32_t		  ses_sector, ses_offset;
	int			  error, resolved;

DEBUG(
	printf("Write logvol space for %s, from vpart %d, lb_num %d at offset %d, for %d bytes\n",
		what, vpart_num, lb_num, (int) offset, (int) length);
);

	udf_open_logvol(udf_log_vol);

	error = udf_logvol_vpart_to_partition(udf_log_vol, vpart_num, &udf_part_mapping, &udf_partition);
	if (error) return error;

	/* get the offset (in bytes) in the partition */
	lb_size     = udf_rw32(udf_log_vol->log_vol->lb_size);
	offset     += (uint64_t) lb_num * lb_size;

	udf_session = udf_partition->udf_session;

	/* Virtual partitions have a write-back buffer cache; it holds a list of virtual sectors */
	/* Its more a still-to-allocate virtual cache to allow easy recording on CD-R/DVD-R/DVD+R etc */
	do {
		trans_length = length;
		resolved     = 0;

		/* determine the translated address and its translation validity length */
		error = udf_translate_addr(udf_part_mapping, udf_partition, offset, &ses_addr, &trans_valid_len);
		if (error) break;

		trans_length = MIN(length, trans_valid_len);
		ses_sector = ses_addr / UDF_SECTOR_SIZE;
		ses_offset = ses_addr % UDF_SECTOR_SIZE;

		error = udf_write_session_bytes_uio(udf_session, ses_sector, ses_offset, trans_length, what, uio, flags, wrcallback);
		if (error) break;

		offset	  += trans_length;
		length    -= trans_length;

		if (length == 0) return error;
	} while (length && !error);

	return EFAULT;
}


/* end of udf_readwrite.c */

