/* $NetBSD$ */

/*
 * File "udfclient.c" is part of the UDFclient toolkit.
 * File $Id: udfclient.c,v 1.67 2005/01/14 21:24:40 imago Exp $ $Name:  $
 *
 * Copyright (c) 2003, 2004 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udftoolkit/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include "udf.h"
#include "udf_bswap.h"

/* switches */

//#define DEBUG(a) (a)
#define DEBUG(a) if (0) { a; };


#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)<(b)?(a):(b))


/* include the dump parts ... in order to get a more sane splitting */
extern void udf_dump_alive_sets(void);


/* globals */
extern int udf_verbose;
extern int scsilib_verbose;
int read_only;


struct curdir {
	char			*name;
	struct udf_mountpoint	*mountpoint;			/* foreign */
	struct udf_node		*udf_node;			/* foreign */
	struct hash_entry	*udf_mountpoint_entry;		/* `current' mountpoint entry */
} curdir;



/*
 * XXX
 * FTP client; de volumes vooraan zetten in de VFS ; evt. in meerdere subdirs.
 * general file name format 
 *
 * /volset:pri:log/udfpath
 *     of
 * /volset/pri/log/udfpath/
 * 
 * XXX
 */

int udfclient_readdir(struct udf_node *udf_node, struct uio *result_uio, int *eof_res) {
	struct dirent          entry;
	struct udf_mountpoint *mountable;

	assert(result_uio);
	if (!udf_node) {
		/* mountables */
		/* XXX result_uio needs to be long enough to hold all mountables!!!! XXX */
		SLIST_FOREACH(mountable, &udf_mountables, all_next) {
			strcpy(entry.d_name, mountable->mount_name);
			entry.d_type = DT_DIR;
			uiomove(&entry, sizeof(struct dirent), result_uio);
		};
		if (eof_res) *eof_res = 1;
		return 0;
	};

	/* intree nodes : pass on to udf_readdir */
	return udf_readdir(udf_node, result_uio, eof_res);
}


/* VOP_LOOKUP a-like */
int udfclient_lookup(struct udf_node *dir_node, struct udf_node **vnode, char *name) {
	struct udf_mountpoint *mountable;

	assert(vnode);
	assert(name);
	*vnode = NULL;
	if (!dir_node) {
		/* mountables */
		SLIST_FOREACH(mountable, &udf_mountables, all_next) {
			if (strcmp(mountable->mount_name, name) == 0) {
				/* found `root' of a mountable */
				*vnode = mountable->rootdir_node;
				return 0;
			};
		};
		return ENOENT;
	}

	/* intree nodes : pass on to udf_lookup */
	return udf_lookup(dir_node, vnode, name);
}


/* higher level of lookup; walk tree */
int udfclient_lookup_pathname(struct udf_node *cur_node, struct udf_node **res_node, char *restpath_given) {
	struct udf_node *sub_node;
	char     *restpath, *next_element, *slashpos, *pathpos;
	int       error;
	
	assert(restpath_given);
	restpath = strdup(restpath_given);

	/* start at root */
	*res_node = NULL;
	pathpos   = restpath;
	assert(*pathpos == '/');
	pathpos++;		/* strip leading '/' */

	next_element = pathpos;
	while (next_element && (strlen(next_element) > 0)) {
		/* determine next slash position */
		slashpos = strchr(next_element, '/');
		if (slashpos) *slashpos++ = 0;

		error = udfclient_lookup(cur_node, &sub_node, next_element);
		if (error) {
			free(restpath);
			return error;
		};

		/* advance */
		cur_node = sub_node;
		next_element = slashpos;
	} 
	/* we are at the end; return result */
	*res_node = cur_node;
	free(restpath);
	return 0;
};


char *udfclient_realpath(char *cur_path, char *relpath) {
	char *resultpath, *here, *pos;

	resultpath = calloc(1, sizeof(cur_path) + sizeof(relpath) +1024);
	assert(resultpath);

	strcpy(resultpath, "/");
	strcat(resultpath, cur_path);
	strcat(resultpath, "/");

	/* check if we are going back to `root' */
	if (relpath && *relpath == '/') {
		strcpy(resultpath, "");
	};
	strcat(resultpath, relpath);
	/* now clean up the resulting string by removing double slashes */
	here = resultpath;
	while (*here) {
		pos = here; while (strncmp(pos, "//", 2) == 0) pos++;
		if (pos != here) strcpy(here, pos);
		here++;
	};

	/* remove '.' and '..' sequences */
	here = resultpath;
	while (*here) {
		// printf("transformed to %s\n", resultpath);
		/* check for internal /./ and trailing /. */
		if (strncmp(here, "/./", 3) == 0) {
			strcpy(here+1, here + 3);
			continue;
		};
		if (strcmp(here, "/.")==0) {
			strcpy(here+1, here + 2);
			continue;
		};
		if (strncmp(here, "/../", 4) == 0) {
			strcpy(here+1, here + 4);
			/* go for the parent */
			pos = here-1; while (*pos && *pos != '/') pos--; pos++;
			strcpy(pos, here+1);
			here = pos;
			continue;
		};
		if (strcmp(here, "/..")==0) {
			strcpy(here+1, here + 3);
			/* go for the parent */
			pos = here-1; while (*pos && *pos != '/') pos--; pos++;
			strcpy(pos, here+1);
			here = pos;
			continue;
		};
		here++;
	};

	return resultpath;
}


static void print_dir_entry(struct udf_node *udf_node, struct dirent *dirent) {
	struct stat     stat;
	uint64_t        size;
	char *name;
	int   mode, this_mode, uid, gid;
	int   error;

	name = dirent->d_name;
	if (udf_node) {
		error = udf_getattr(udf_node, &stat);
		if (error) {
			fprintf(stderr, "Can't stat file %s\n", name);
			return;
		};
		dirent = &udf_node->dirent;

		mode = stat.st_mode;
		size = stat.st_size;
		uid  = stat.st_uid;
		gid  = stat.st_gid;
		if (dirent->d_type == DT_DIR) mode |= 512;	/* XXX */
	} else {
		/* dummy entry for `root' in VFS */
		mode = 0744 | 512;
		size = 0;
		uid  = 0;
		gid  = 0;
	};

	if (mode & 512) printf("d"); else printf("-");
	mode = mode & 511;

	this_mode = (mode >> 6) & 7;
	printf("%c%c%c", "----rrrr"[this_mode & 4], "--www"[this_mode & 2], "-x"[this_mode & 1]);
	this_mode = (mode >> 3) & 7;
	printf("%c%c%c", "----rrrr"[this_mode & 4], "--www"[this_mode & 2], "-x"[this_mode & 1]);
	this_mode = mode & 7;
	printf("%c%c%c", "----rrrr"[this_mode & 4], "--www"[this_mode & 2], "-x"[this_mode & 1]);

	printf("  %4d  %4d  %10lu  %s\n", gid, uid, (unsigned long) size, name);
}


void udfclient_ls(int args, char *arg1, char *arg2) {
	struct udf_node *udf_node, *entry_node;
	uint8_t        buffer[UDF_SECTOR_SIZE];
	struct uio     dir_uio;
	struct iovec   dir_uiovec;
	struct dirent *dirent;
	uint32_t       pos;
	int            eof;
	char          *node_name;
	int            error;

	if (args > 2) {
		printf("Syntax: ls [file | dir]\n");
		return;
	};

	node_name = udfclient_realpath(curdir.name, arg1);

	error = udfclient_lookup_pathname(NULL, &udf_node, node_name);
	if (error) {
		fprintf(stderr, "%s : %s\n", arg1, strerror(error));
		free(node_name); 
		return;
	};

	if (!udf_node || (udf_node->dirent.d_type == DT_DIR)) {
		printf("Directory listing of %s\n", udf_node ? udf_node->dirent.d_name : "/");

		/* start at the start of the directory */
		dir_uio.uio_offset = 0;
		dir_uio.uio_iov     = &dir_uiovec;
		dir_uio.uio_iovcnt  = 1;
		do {
			dir_uiovec.iov_base = buffer;
			dir_uiovec.iov_len  = UDF_SECTOR_SIZE;
			dir_uio.uio_resid   = UDF_SECTOR_SIZE;
			dir_uio.uio_rw      = UIO_WRITE;

			error = udfclient_readdir(udf_node, &dir_uio, &eof);
			pos = 0;
			while (pos < (UDF_SECTOR_SIZE - dir_uio.uio_resid)) {
				dirent = (struct dirent *) (buffer + pos);
				error = udfclient_lookup(udf_node, &entry_node, dirent->d_name);
				print_dir_entry(entry_node, dirent);

				pos += sizeof(struct dirent);
			};
		} while (!eof);
	} else {
		print_dir_entry(udf_node, &udf_node->dirent);
	};
	free(node_name);
}


void udfclient_pwd(int args) {
	char pwd[1024];

	if (args > 1) {
		printf("Syntax: pwd\n");
		return;
	};
	getcwd(pwd, 1024);
	printf("UDF working directory is     %s\n", curdir.name);
	printf("Current FS working directory %s\n", pwd);
}


void udfclient_free(int args) {
	if (args > 1) {
		printf("Syntax: pwd\n");
		return;
	};
	printf("Free space reporting not yet implemented\n");
}


void udfclient_cd(int args, char *arg1, char *arg2) {
	struct udf_node *udf_node;
	char *node_name, *new_curdir_name;
	int   error;

	if (args > 2) {
		printf("Syntax: cd [dir]\n");
		return;
	};
	new_curdir_name = udfclient_realpath(curdir.name, arg1);
	node_name = strdup(new_curdir_name);	/* working copy */

	error = udfclient_lookup_pathname(NULL, &udf_node, node_name);
	if (error) {
		fprintf(stderr, "%s : %s\n", arg1, strerror(error));
		free(new_curdir_name);
		free(node_name); 
		return;
	};

	if (!udf_node || udf_node->dirent.d_type == DT_DIR) {
		free(curdir.name);
		curdir.name     = new_curdir_name;
		curdir.udf_node = udf_node;
		free(node_name);

		udfclient_pwd(0);
	} else {
		fprintf(stderr, "%s is not a directory\n", arg1);
		free(new_curdir_name);
		free(node_name);
	};
}


void udfclient_lcd(int args, char *arg1) {
	char pwd[1024];

	if (args > 2) {
		printf("Syntax: lcd [dir]\n");
		return;
	};
	if (strcmp(arg1, "")==0)  arg1 = getenv("HOME");
	if (strcmp(arg1, "~")==0) arg1 = getenv("HOME");

	if (chdir(arg1)) {
		fprintf(stderr, "While trying to go to %s :", arg1);
		perror("");
	};
	getcwd(pwd, 1024);
	printf("Changing local directory to %s\n", pwd);
}


void udfclient_lls(int args) {
	if (args > 1) {
		printf("Syntax: lls\n");
		return;
	};
	if (system("/bin/ls")) {
		perror("While listing current directory\n");
	};
}


uint64_t getmtime(void) {
	struct timeval tp;

	gettimeofday(&tp, NULL);
	return 1000000*tp.tv_sec + tp.tv_usec;
}



void udfclient_get_file(struct udf_node *udf_node, char *node_name) {
	struct uio	 file_uio;
	struct iovec	 file_iov;
	uint64_t	 file_length;
	uint64_t 	 start, now, then, eta;
	uint64_t	 cur_speed, avg_speed, data_transfered;
	uint64_t	 file_block_size, file_transfer_size;
	uint8_t		*file_block;
	char		 cur_txt[32], avg_txt[32], eta_txt[32];
	int		 fileh, len;
	int		 error;

	assert(udf_node);
	assert(node_name);
	assert(strlen(node_name) >= 1);

	/* terminal directory node? */
	if (udf_node->dirent.d_type == DT_DIR) {
		len = strlen(node_name);
		if (strcmp(node_name + len -2, "/.")  == 0) node_name[len-2] = 0;
		if (strcmp(node_name + len -3, "/..") == 0) return;

		error = mkdir(node_name, (udf_node->stat.st_mode) & 07777);
		if (error) fprintf(stderr, "While creating directory : %s\n", strerror(errno));

		return;
	};

	/* terminal file node; setting mode correctly */
	fileh = open(node_name, O_WRONLY | O_CREAT, udf_node->stat.st_mode);
	if (fileh >= 0) {
		file_length = udf_node->stat.st_size;
		file_block_size = 256*1024;		/* block read in length */
		file_block = malloc(file_block_size);
		if (!file_block) {
			printf("Out of memory claiming file buffer\n");
			return;
		};

		/* move to uio_newuio(struct uio *uio) with fixed length uio_iovcnt? */
		bzero(&file_uio, sizeof(struct uio));
		file_uio.uio_rw     = UIO_WRITE;	/* WRITE into this space */
		file_uio.uio_iovcnt = 1;
		file_uio.uio_iov    = &file_iov;

		start = getmtime();
		then  = now = start;
		eta = data_transfered = 0;
		strcpy(avg_txt, "---"); strcpy(cur_txt, "---"); strcpy(eta_txt, "---");

		file_uio.uio_offset = 0;		/* begin at the start */
		do {
			/* fill in IO vector space; reuse blob file_block over and over */
			file_transfer_size = MIN(file_block_size, file_length - file_uio.uio_offset);
			file_uio.uio_resid  = file_transfer_size;
			file_uio.uio_iov->iov_base = file_block;
			file_uio.uio_iov->iov_len  = file_block_size;

			error = udf_read_file_part_uio(udf_node, node_name, 0, &file_uio);
			if (error) {
				fprintf(stderr, "While retrieving file block : %s\n", strerror(error));
				printf("\n\n\n");	/* XXX */
				break;
			};

			write(fileh, file_block, file_transfer_size);

			if ((getmtime() - now > 1000000) || (file_uio.uio_offset >= file_length)) {
				if (strlen(node_name) < 45) {
					printf("\r%-45s ", node_name);
				} else {
					printf("\r...%-42s ", node_name+strlen(node_name)-42);
				};
				printf("%10d / %10d bytes ", (uint32_t) file_uio.uio_offset, (uint32_t) file_length);
				if (file_length) printf("(%3d%%) ", (int) (100.0*(float) file_uio.uio_offset / file_length));

				then = now;
				now  = getmtime();
				cur_speed = 0;
				avg_speed = 0;
				if (now-start > 0) avg_speed = (1000000 * file_uio.uio_offset) / (now-start);
				if (now-then  > 0) cur_speed = (1000000 * (file_uio.uio_offset - data_transfered)) / (now-then);
				if (avg_speed > 0) eta = (file_length - file_uio.uio_offset) / avg_speed;
				data_transfered = file_uio.uio_offset;

				strcpy(avg_txt, "---"); strcpy(cur_txt, "---"); strcpy(eta_txt, "---");
				if (avg_speed > 0) sprintf(avg_txt, "%d", (int32_t) avg_speed/1000);
				if (cur_speed > 0) sprintf(cur_txt, "%d", (int32_t) cur_speed/1000);
				if (eta       > 0) sprintf(eta_txt, "%02d:%02d:%02d", (int) (eta/3600), (int) (eta/60) % 60, (int) eta % 60);

				printf("%6s KB/s (%6s KB/s) ETA %s", avg_txt, cur_txt, eta_txt);
				fflush(stdout);
			};
		} while ((file_uio.uio_offset < file_length));
		printf(" finished\n");
		free(file_block);
		close(fileh);
	} else {
		printf("Help! can't open file %s for output\n", node_name);
	};

	return;
}


#define GET_SUBTREE_DIR_BUFFER_SIZE (8*UDF_SECTOR_SIZE)
void udfclient_get_subtree(struct udf_node *udf_node, char *prefix, int recurse) {
	struct uio       dir_uio;
	struct iovec     dir_iovec;
	uint8_t         *buffer;
	uint32_t         pos;
	char             fullpath[1024];	/* XXX arbitrary length XXX */
	struct dirent   *dirent;
	struct udf_node *entry_node;
	int              new_recurse, eof;
	int		 error;

	if (!udf_node) return;

	if ((udf_node->dirent.d_type == DT_DIR) && recurse) {
		buffer = malloc(GET_SUBTREE_DIR_BUFFER_SIZE);
		if (!buffer) return;

		/* recurse into this directory */
		dir_uio.uio_offset = 0;			/* begin at start */
		do {
			dir_iovec.iov_base = buffer;
			dir_iovec.iov_len  = GET_SUBTREE_DIR_BUFFER_SIZE;
			dir_uio.uio_resid  = GET_SUBTREE_DIR_BUFFER_SIZE;
			dir_uio.uio_iovcnt = 1;
			dir_uio.uio_iov    = &dir_iovec;
			dir_uio.uio_rw     = UIO_WRITE;

			error = udf_readdir(udf_node, &dir_uio, &eof);
			pos = 0;
			while (pos < GET_SUBTREE_DIR_BUFFER_SIZE - dir_uio.uio_resid) {
				dirent = (struct dirent *) (buffer + pos);

				sprintf(fullpath, "%s/%s", prefix, dirent->d_name);

				error = udf_lookup(udf_node, &entry_node, dirent->d_name);
				new_recurse = (strcmp(dirent->d_name, ".") && strcmp(dirent->d_name, ".."));
				udfclient_get_subtree(entry_node, fullpath, new_recurse);

				pos += sizeof(struct dirent);	/* XXX variable size dirents possible XXX */
			};
		} while (!eof);

		free(buffer);
		return;
	};

	/* leaf node : prefix is complete name but with `/' prefix */
	if (*prefix == '/') prefix++;
	udfclient_get_file(udf_node, prefix);
}
#undef GET_SUBTREE_DIR_BUFFER_SIZE


void udfclient_get(int args, char *arg1, char *arg2) {
	struct udf_node *udf_node;
	char *node_name, *target_name;
	int   error;

	if (args != 2) {
		printf("Syntax: get (file | dir)\n");
		return;
	};

	node_name = udfclient_realpath(curdir.name, arg1);
	DEBUG(printf("Attempting to retrieve %s\n", node_name));

	error = udfclient_lookup_pathname(NULL, &udf_node, node_name);
	if (error) {
		fprintf(stderr, "%s : %s\n", arg1, strerror(error));
		free(node_name); 
		return;
	};

	/* make copy of the name the node is known by */
	target_name = strdup(udf_node->dirent.d_name);

	/* get the file/dir tree */
	udfclient_get_subtree(udf_node, target_name, 1);

	/* free copy */
	free(target_name);
}


int udfclient_put_file(struct udf_node *udf_node, char *node_name) {
	struct uio	 file_uio;
	struct iovec	 file_iov;
	uint64_t	 file_length;
	uint64_t 	 start, now, then, eta;
	uint64_t	 cur_speed, avg_speed, data_transfered;
	uint64_t	 file_block_size, file_transfer_size;
	uint8_t		*file_block;
	char		 cur_txt[32], avg_txt[32], eta_txt[32];
	int		 fileh;
	int   		 error;

	assert(udf_node);
	assert(node_name);

	DEBUG(printf("Attempting to write %s\n", node_name));

	fileh = open(node_name, O_RDONLY, 0666);
	if (fileh == -1) {
		fprintf(stderr, "Can't open local file %s for reading: %s\n", node_name, strerror(errno));
		return ENOENT;
	};

	file_length = lseek(fileh, 0, SEEK_END);
	lseek(fileh, 0, SEEK_SET);
	file_block_size = 128*1024;
	file_block = malloc(file_block_size);
	if (!file_block) {
		fprintf(stderr, "Out of memory claiming file buffer\n");
		free(node_name);
		return ENOMEM;
	};

	/* move to uio_newuio(struct uio *uio) with fixed length uio_iovcnt? */
	bzero(&file_uio, sizeof(struct uio));
	file_uio.uio_rw     = UIO_READ;			/* READ from this space */
	file_uio.uio_iovcnt = 1;
	file_uio.uio_iov    = &file_iov;

/* ------------ */
	start = getmtime();
	then  = now = start;
	eta = data_transfered = 0;
	strcpy(avg_txt, "---"); strcpy(cur_txt, "---"); strcpy(eta_txt, "---");
/* ------------ */

	error = 0;
	while (!error && (file_uio.uio_offset < file_length)) {
		file_transfer_size = MIN(file_block_size, file_length - file_uio.uio_offset);

		error = read(fileh, file_block, file_transfer_size);
		if (error<0) {
			fprintf(stderr, "While reading in file block for writing : %s\n", strerror(errno));
			error = errno;
			break;
		};

		file_uio.uio_resid  = file_transfer_size;
		file_uio.uio_iov->iov_base = file_block;
		file_uio.uio_iov->iov_len  = file_block_size;

		error = udf_write_file_part_uio(udf_node, node_name, UDF_CACHEFLAG_NOREADIN, &file_uio);
		if (error) break;

/* ------------ */
		if ((getmtime() - now > 1000000) || (file_uio.uio_offset >= file_length)) {
			if (strlen(node_name) < 45) {
				printf("\r%-45s ", node_name);
			} else {
				printf("\r...%-42s ", node_name+strlen(node_name)-42);
			};
			printf("%10d / %10d bytes ", (uint32_t) file_uio.uio_offset, (uint32_t) file_length);
			if (file_length) printf("(%3d%%) ", (int) (100.0*(float) file_uio.uio_offset / file_length));

			then = now;
			now  = getmtime();
			cur_speed = 0;
			avg_speed = 0;
			if (now-start > 0) avg_speed = (1000000 * file_uio.uio_offset) / (now-start);
			if (now-then  > 0) cur_speed = (1000000 * (file_uio.uio_offset - data_transfered)) / (now-then);
			if (avg_speed > 0) eta = (file_length - file_uio.uio_offset) / avg_speed;
			data_transfered = file_uio.uio_offset;

			strcpy(avg_txt, "---"); strcpy(cur_txt, "---"); strcpy(eta_txt, "---");
			if (avg_speed > 0) sprintf(avg_txt, "%d", (int32_t) avg_speed/1024);
			if (cur_speed > 0) sprintf(cur_txt, "%d", (int32_t) cur_speed/1024);
			if (eta       > 0) sprintf(eta_txt, "%02d:%02d:%02d", (int) (eta/3600), (int) (eta/60) % 60, (int) eta % 60);
			printf("%6s KB/s (%6s KB/s) ETA %s", avg_txt, cur_txt, eta_txt);
			fflush(stdout);
		};
/* ------------ */
	};
	if (!error) printf(" finished\n"); else printf("\n");
	close(fileh);
	free(file_block);

	return error;
}


int udfclient_put_subtree(struct udf_node *parent_node, char *prefix, char *fname, uint64_t *totalsize) {
	struct udf_node *file_node, *dir_node;
	struct dirent *dirent;
	struct stat    stat;
	DIR           *dir;
	char           fullpath[1024];	/* XXX arbitrary length XXX */
	int            error;

	sprintf(fullpath, "%s/%s", prefix, fname);

	bzero(&stat, sizeof(struct stat));
	error = lstat(fullpath, &stat);
	if (error) {
		fprintf(stderr, "Can't stat file/dir \"%s\"! : %s\n", fullpath, strerror(error));
		return error;
	};
#if 0
	stat.st_uid  = UINT_MAX;
	stat.st_gid  = UINT_MAX;
	stat.st_mode = 0755;
#endif

	/* test if `fname' refers to a directory */
	dir = opendir(fullpath);
	if (dir) {
		error = udfclient_lookup(parent_node, &dir_node, fname);
		if (error) {
			DEBUG(printf("Create dir %s on UDF\n", fullpath));
			error = udf_create_directory(parent_node, fname, &stat, &dir_node);
			if (error) {
				closedir(dir);
				fprintf(stderr, "UDF: couldn't create new directory %s : %s\n", fname, strerror(error));
				return error;
			};
		};

		dirent = readdir(dir);
		while (dirent) {
			if (strcmp(dirent->d_name, ".") && strcmp(dirent->d_name, "..")) {
				/* skip `.' and ',,' */
				error = udfclient_put_subtree(dir_node, fullpath, dirent->d_name, totalsize);
				if (error) break;
			};
			dirent = readdir(dir);
		};
		closedir(dir);
		return error;
	};

	/* leaf node : prefix is complete name but with `/' prefix */
	DEBUG(printf("Put leaf: %s\n", fullpath));

	error = udfclient_lookup(parent_node, &file_node, fname);
	if (!file_node) {
		error = udf_create_file(parent_node, fname, &stat, &file_node);
		if (error) {
			fprintf(stderr, "UDF: couldn't add new file entry in directory %s for %s : %s\n", prefix, fname, strerror(error));
			return error;
		};
	};
	error = udfclient_put_file(file_node, fullpath);
	if (error) fprintf(stderr, "UDF: Couldn't write file %s : %s\n", fullpath, strerror(error));
	if (error) udf_remove_file(parent_node, file_node, fname);

	if (!error) *totalsize += file_node->stat.st_size;

	return error;
}


void udfclient_put(int args, char *arg1, char *arg2) {
	struct udf_node *curdir_node;
	uint64_t	 start, now, totalsize, avg_speed;
	char		*target_name;
	int              error;

	if (args != 2) {
		printf("Syntax: put (file | dir)\n");
		return;
	};

	if (read_only) {
		printf("Modifying this filingsystem is prevented; use -W flag to enable writing on your own risk!\n");
		return;
	};

	error = udfclient_lookup_pathname(NULL, &curdir_node, curdir.name);
	if (error) {
		printf("Current directory not found?\n");
		return;
	};
	DEBUG(printf("Attempting to copy %s\n", arg1));

	/* make copy of name */
	target_name = strdup(arg1);

	/* writeout file/dir tree and measure the time/speed */
	totalsize = 0;
	start = getmtime();
		error = udfclient_put_subtree(curdir_node, ".", target_name, &totalsize);
	now  = getmtime();
	if (now-start > 0) {
		avg_speed = (1000000 * totalsize) / (now-start);
		printf("A total of %d kb transfered at an overal average of %d kb/sec\n", (uint32_t)totalsize/1024, (uint32_t)avg_speed/1024);
	} else {
		printf("Transfered %d kb\n", (uint32_t) totalsize/1024);
	};

	/* free copy */
	free(target_name);
}



void udfclient_trunc(int args, char *arg1, char *arg2) {
	struct udf_node *udf_node;
	char		*node_name;
	uint64_t	 length;
	int		 error;

	if (args != 3) {
		printf("Syntax: trunc file length\n");
		return;
	};
	length = strtoll(arg2, NULL, 10);

	node_name = udfclient_realpath(curdir.name, arg1);
	error = udfclient_lookup_pathname(NULL, &udf_node, node_name);
	if (error || !udf_node) {
		printf("Can only truncate existing file!\n");
		free(node_name);
		return;
	};

	udf_truncate_udf_node(udf_node, length);

	free(node_name);
}


void udfclient_sync(void) {
	struct udf_discinfo *udf_disc;

	SLIST_FOREACH(udf_disc, &udf_discs_list, next_disc) {
		udf_sync_disc(udf_disc);
	};
}


#define RM_SUBTREE_DIR_BUFFER_SIZE UDF_SECTOR_SIZE
void udfclient_rm_subtree(struct udf_node *parent_node, struct udf_node *udf_node, char *name, char *prefix, int recurse) {
	struct uio       dir_uio;
	struct iovec     dir_iovec;
	uint8_t         *buffer;
	uint32_t         pos;
	char             fullpath[1024];	/* XXX arbitrary length XXX */
	struct dirent   *dirent;
	struct udf_node *entry_node;
	int              new_recurse, eof;
	int		 error;

	if (!udf_node) return;

	if ((udf_node->dirent.d_type == DT_DIR) && recurse) {
		buffer = malloc(RM_SUBTREE_DIR_BUFFER_SIZE);
		if (!buffer) return;

		/* recurse into this directory */
		dir_uio.uio_offset = 0;			/* begin at start */
		do {
			dir_iovec.iov_base = buffer;
			dir_iovec.iov_len  = RM_SUBTREE_DIR_BUFFER_SIZE;
			dir_uio.uio_resid  = RM_SUBTREE_DIR_BUFFER_SIZE;
			dir_uio.uio_iovcnt = 1;
			dir_uio.uio_iov    = &dir_iovec;
			dir_uio.uio_rw     = UIO_WRITE;

			error = udf_readdir(udf_node, &dir_uio, &eof);
			pos = 0;
			while (pos < RM_SUBTREE_DIR_BUFFER_SIZE - dir_uio.uio_resid) {
				dirent = (struct dirent *) (buffer + pos);

				sprintf(fullpath, "%s/%s", prefix, dirent->d_name);

				error = udf_lookup(udf_node, &entry_node, dirent->d_name);
				new_recurse = (strcmp(dirent->d_name, ".") && strcmp(dirent->d_name, ".."));
				udfclient_rm_subtree(udf_node, entry_node, dirent->d_name, fullpath, new_recurse);

				pos += sizeof(struct dirent);	/* XXX variable size dirents possible XXX */
			};
		} while (!eof);

		free(buffer);

		/* leaving directory -> delete directory */
		printf("rmdir %s/%s\n", prefix, name);
		udf_remove_directory(parent_node, udf_node, name);
		return;
	};

	/* leaf node : prefix is complete name but with `/' prefix */
	if (*prefix == '/') prefix++;

	/* don't delete this dir or parent dir */
	if (strcmp(name, ".")  == 0) return;
	if (strcmp(name, "..") == 0) return;

	printf("rm %s\n", prefix);
	udf_remove_file(parent_node, udf_node, name);
}
#undef RM_SUBTREE_DIR_BUFFER_SIZE



void udfclient_rm(int args, char *arg1) {
	struct udf_node *udf_node, *parent_node;
	char		*node_name;
	int		 error;

	if (args != 2) {
		printf("Syntax: rm file\n");
		return;
	};

	node_name = udfclient_realpath(curdir.name, arg1);
	error = udfclient_lookup_pathname(NULL, &udf_node, node_name);
	if (error || !udf_node) {
		printf("Can only delete existing file/dir !\n");
		free(node_name);
		return;
	};
	error = udfclient_lookup_pathname(NULL, &parent_node, curdir.name);
	if (error || !parent_node) {
		printf("Current directory not found?\n");
		free(node_name);
		return;
	};
	udfclient_rm_subtree(parent_node, udf_node, arg1, "" /* arg1 */, 1);
	free(node_name);
}


void udfclient_mkdir(int args, char *arg1, char *arg2) {
	struct stat	 stat;
	struct udf_node	*udf_node, *parent_node;
	int		 error;

	if (args != 2) {
		printf("Syntax: mkdir dir\n");
		return;
	};

	error = udfclient_lookup_pathname(NULL, &parent_node, curdir.name);
	if (error || !parent_node) {
		printf("Current directory not found?\n");
		return;
	};

	bzero(&stat, sizeof(struct stat));
	stat.st_uid  = UINT_MAX;
	stat.st_gid  = UINT_MAX;
	stat.st_mode = 0755;

	error = udf_create_directory(parent_node, arg1, &stat, &udf_node);
	if (error) {
		printf("Can't create directory %s : %s\n", arg1, strerror(error));
		return;
	};
}


char *udfclient_get_one_arg(char *line, char *result) {
	char chr, limiter;

	*result = 0;

	/* get prepending whitespace */
	while (*line && (*line <= ' ')) line++;

	limiter = ' ';
	if (*line == '"') {
		line++;
		limiter = '"';
	};

	while (*line) {
		chr = *line++;
		if (chr && (chr < ' ')) chr = ' ';
		if (chr == 0 || chr == limiter) {
			break;
		} else {
			*result++ = chr;
		};
	};
	*result = 0;

	/* get appended whitespace */
	while (*line && (*line <= ' ')) line++;

	return line;
}


int udfclient_get_args(char *line, char *cmd, char *arg1, char *arg2, char *arg3) {
	int args;

	*cmd = *arg1 = *arg2 = *arg3 = 0;
	args = 0;

	/* get command */
	line = udfclient_get_one_arg(line, cmd);
	args++;
	if (!*line) return args;

	/* get arg1 */
	line = udfclient_get_one_arg(line, arg1);
	args++;
	if (!*line) return args;

	/* get arg2 */
	line = udfclient_get_one_arg(line, arg1);
	args++;
	if (!*line) return args;

	/* get arg3 */
	line = udfclient_get_one_arg(line, arg1);
	args++;
	if (!*line) return args;

	printf("UDFclient implementation limit: too many arguments\n");
	return 0;
}


void udfclient_interact(void) {
	int  args;
	char cmd[1024], arg1[1024], arg2[1024], arg3[1024];
	char line[4096];

	udfclient_pwd(0);
	while (1) {
		printf("UDF> ");
		clearerr(stdin);

		*line = 0;
		fgets(line, 4096, stdin);

		if (*line == 0 && feof(stdin))
			return;

		args = udfclient_get_args(line, cmd, arg1, arg2, arg3);

		if (args) {
			if (strcmp(cmd, "")==0) continue;

			if (strcmp(cmd, "ls")==0) {
				udfclient_ls(args, arg1, arg2);
				continue;
			};
			if (strcmp(cmd, "cd")==0) {
				udfclient_cd(args, arg1, arg2);
				continue;
			};
			if (strcmp(cmd, "lcd")==0) {
				udfclient_lcd(args, arg1);
				continue;
			};
			if (strcmp(cmd, "lls")==0) {
				udfclient_lls(args);
				continue;
			};
			if (strcmp(cmd, "pwd")==0) {
				udfclient_pwd(args);
				continue;
			};
			if (strcmp(cmd, "free")==0) {
				udfclient_free(args);
				continue;
			};
			if (strcmp(cmd, "get")==0) {
				udfclient_get(args, arg1, arg2);
				continue;
			};
			if (strcmp(cmd, "put")==0) {
				udfclient_put(args, arg1, arg2);
				continue;
			};
			if (strcmp(cmd, "trunc")==0) {
				udfclient_trunc(args, arg1, arg2);
				continue;
			};
			if (strcmp(cmd, "mkdir")==0) {
				udfclient_mkdir(args, arg1, arg2);
				continue;
			};
			if (strcmp(cmd, "rm")==0) {
				udfclient_rm(args, arg1);
				continue;
			};
			if (strcmp(cmd, "sync")==0) {
				udfclient_sync();
				continue;
			};
			if (strcmp(cmd, "help")==0) {
				printf("Selected commands available (use \" pair for filenames with spaces) :\n"
						"ls  [file | dir]\t\tlists the UDF directory\n"
						"cd  [dir]\t\tchange current UDF directory\n"
						"lcd [dir]\t\tchange current directory\n"
						"lls\t\t\tlists current directory\n"
						"pwd\t\t\tdisplay current directories\n"
						"free\t\t\tdisplay free space on disc\n"
						"get (file | dir)\tretrieve a file or a directory\n"
						"put file\t\twrite a file or a directory to disc\n"
						"trunc file length\ttrunc file to length\n"
						"mkdir dir\t\tcreate directory\n"
						"rm  (file | dir)\tdelete file or directory tree\n"
						"sync\t\t\tsync filingsystem\n"
						"quit\t\t\texits program\n"
						"exit\t\t\talias for quit\n"
				      );
				continue;
			};
			if (strcmp(cmd, "quit")==0 ||
			    strcmp(cmd, "exit")==0) {
				return;
			};
			printf("\nUnknown command %s\n", cmd);
		};
	};
}


int usage(char *program) {
	fprintf(stderr, "Usage: %s [options] devicename [devicename]*)\n", program);
	fprintf(stderr, "-u level	UDF system verbose level\n"
			"-s 		verbose SCSI command errors\n"
			"-r range	use only selected sessions like -3,5,7 or 6-\n"
			"-W		allow writing (temporary flag)\n"
	);
	return 1;
}


extern char	*optarg;
extern int	 optind;
extern int	 optreset;


int main(int argc, char *argv[]) {
	struct udf_discinfo *disc, *next_disc;
	int   flag;
	char *progname, *range;
	int   error;

	progname = argv[0];
	if (argc == 1) {
		return usage(progname);
	};

	udf_verbose	= UDF_VERBLEV_NONE;
	scsilib_verbose	= 0;
	read_only       = 1;
	range		= NULL;
	while ((flag = getopt(argc, argv, "u:sr:W")) != -1) {
		switch (flag) {
			case 'u' :
				udf_verbose = atoi(optarg);
				break;
			case 's' :
				scsilib_verbose = 1;
				break;
			case 'r' :
				range = strdup(optarg);
				if (udf_check_session_range(range)) {
					fprintf(stderr, "Invalid range %s\n", range);
					return usage(progname);
				};
				break;
			case 'W' :
				read_only = 0;
				break;
			default  :
				return usage(progname);
		};
	};
	argv += optind;
	argc -= optind;

	if (argc == 0) return usage(progname);

	if (!read_only) {
		printf("--------------------------------\n");
		printf("WARNING: writing enabled, use on own risc\n");
		printf("\t* DONT cancel program or data-loss might occure\n");
		printf("\t* set dataspace userlimits very high when writing thousands of files\n");
		printf("\nEnjoy wour writing!\n");
		printf("--------------------------------\n\n\n");
		printf("%c", 7); fflush(stdout); sleep(1); printf("%c", 7); fflush(stdout); sleep(1); printf("%c", 7); fflush(stdout);
	};

	/* all other arguments are devices */
	SLIST_INIT(&udf_discs_list);
	while (argc) {
		printf("Opening device %s\n", *argv);
		error = udf_mount_disc(*argv, range, &disc);
		if (error) {
			fprintf(stderr, "Can't open my device; bailing out : %s\n", strerror(error));
			exit(1);
		};
		if (read_only) disc->recordable = 0;
		if (read_only) disc->rewritable = 0;

		argv++; argc--;
		if (udf_verbose) printf("\n\n");
	};

	printf("\n");
	printf("Resulting list of alive sets :\n\n");
	udf_dump_alive_sets();

	/* interactive part */
	bzero(&curdir, sizeof(struct curdir));
	curdir.mountpoint = NULL;
	curdir.name = strdup("/");
	udfclient_ls(0, "", "");

	udfclient_interact();

	/* close part */
	printf("Closing discs\n");
	disc = SLIST_FIRST(&udf_discs_list);
	while (disc) {
		next_disc = SLIST_NEXT(disc, next_disc);

		udf_dismount_disc(disc);

		disc = next_disc;
	};

	return 0;
}

