
/* $NetBSD$ */

/*
 * File "udf.c" is part of the UDFclient toolkit.
 * File $Id: udf.c,v 1.194 2005/01/27 14:37:04 imago Exp $ $Name:  $
 *
 * Copyright (c) 2003, 2004 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udftoolkit/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */



#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>
#include <dirent.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <time.h>

#include "scsi_sense.h"
#include "scsilib.h"


/* for locals */
#include "udf.h"
#include "udf_bswap.h"
#include "udf_discop.h"
#include "uio.h"
#include <pthread.h>


/* for scsilib */
const char *dvname="UDF device";


#ifndef MAX
#	define MAX(a,b) ((a)>(b)?(a):(b))
#	define MIN(a,b) ((a)<(b)?(a):(b))
#endif


//#define DEBUG(a) { a; };
#define DEBUG(a) if (0) { a; };


#if 1
	extern void udf_dump_descriptor(union dscrptr *dscrpt);
	extern void udf_dump_vat_table(struct udf_part_mapping *udf_part_mapping);
	extern void udf_dump_disc_anchors(struct udf_discinfo *disc);
	extern void udf_dump_alive_sets(void);
	extern void udf_dump_root_dir(struct udf_mountpoint *mountpoint);
#else
	void udf_dump_descriptor(union dscrptr *dscrpt) {};
	void udf_dump_vat_table(struct udf_part_mapping *udf_part_mapping) {};
	void udf_dump_disc_anchors(struct udf_discinfo *disc) {};
	void udf_dump_alive_sets(void) {};
	void udf_dump_root_dir(struct udf_mountpoint *mountpoint) {};
#endif


/* globals */
int     udf_verbose = 0;
uint8_t zero_sector[UDF_SECTOR_SIZE];


#define DIRREAD_BUFFER_SIZE  (8*UDF_SECTOR_SIZE)
#define UDF_INODE_NUM_GUESS  512


/* predefines */
int udf_validate_tag_sums(union dscrptr *dscr);
void udf_node_mark_dirty(struct udf_node *udf_node);

static void udf_set_imp_id(struct regid *regid);
static void udf_set_app_id(struct regid *regid);
static void udf_set_timestamp_now(struct timestamp *timestamp);
static void udf_node_unmark_dirty(struct udf_node *udf_node);
static void udf_init_desc_tag(struct desc_tag *tag, uint16_t id, uint16_t dscr_ver, uint16_t serial_num);
static int udf_translate_icb_filetype_to_dirent_filetype(int udf_filetype);


/******************************************************************************************
 *
 * Filename space conversion
 *
 ******************************************************************************************/

void udf_to_unix_name(char *result, char *id, int len, struct charspec *chsp) {
	uint16_t  raw_name[1024], unix_name[1024];
	uint16_t *inchp, ch;
	uint8_t	 *outchp;
	int       ucode_chars, nice_uchars;

	outchp = result;
	if ((chsp->type == 0) && (strcmp(chsp->inf, "OSTA Compressed Unicode") == 0)) {
		*raw_name = *unix_name = 0;
		ucode_chars = udf_UncompressUnicode(len, id, raw_name);
		ucode_chars = MIN(ucode_chars, UnicodeLength((unicode_t *) raw_name));
		nice_uchars = UDFTransName(unix_name, raw_name, ucode_chars);
		for (inchp = unix_name; nice_uchars>0; inchp++, nice_uchars--) {
			ch = *inchp;
			/* sloppy unicode -> latin */
			*outchp++ = ch & 255;
			if (!ch) break;
		};
		*outchp++ = 0;
	} else {
		/* assume 8bit char length byte latin-1 */
		assert(*id == 8);
		strncpy(result, id+1, strlen(id+1));
	};
}


void unix_to_udf_name(char *result, char *name, uint8_t *result_len, struct charspec *chsp) {
	uint16_t  raw_name[1024];
	int       udf_chars, name_len;
	char     *inchp;
	uint16_t *outchp;

	/* convert latin-1 or whatever to unicode-16 */
	*raw_name = 0;
	name_len  = 0;
	inchp  = name;
	outchp = raw_name;
	while (*inchp) {
		*outchp++ = (uint16_t) (*inchp++);
		name_len++;
	};

	if ((chsp->type == 0) && (strcmp(chsp->inf, "OSTA Compressed Unicode") == 0)) {
		udf_chars = udf_CompressUnicode(name_len, 8, (unicode_t *) raw_name, (byte *) result);
	} else {
		/* assume 8bit char length byte latin-1 */
		*result++ = 8; udf_chars = 1;
		strncpy(result, name + 1, strlen(name+1));
		udf_chars += strlen(name);
	};
	*result_len = udf_chars;
}


static char *udf_get_compound_name(struct udf_mountpoint *mountpoint) {
	static char         compound[128+128+32+32+1];
	struct charspec    *charspec;
	struct udf_log_vol *udf_log_vol;
	struct udf_pri_vol *udf_pri_vol;
	char               *unix_name;

	udf_log_vol = mountpoint->udf_log_vol;
	udf_pri_vol = udf_log_vol->primary;

	charspec = &udf_pri_vol->pri_vol->desc_charset;
	assert(charspec->type == 0);
	assert(strcmp(charspec->inf, "OSTA Compressed Unicode")==0);

	unix_name = compound;

	udf_to_unix_name(unix_name, udf_pri_vol->pri_vol->volset_id, 128, charspec);
	strcat(unix_name, ":");
	unix_name += strlen(unix_name);

	udf_to_unix_name(unix_name, udf_pri_vol->pri_vol->vol_id, 32, charspec);
	strcat(unix_name, ":");
	unix_name += strlen(unix_name);

	udf_to_unix_name(unix_name, udf_log_vol->log_vol->logvol_id, 128, charspec);
	strcat(unix_name, ":");
	unix_name += strlen(unix_name);

	udf_to_unix_name(unix_name, mountpoint->fileset_desc->fileset_id, 32, charspec);

	return compound;
}


/******************************************************************************************
 *
 * Dump helpers for printing out information during parse
 *
 ******************************************************************************************/

void udf_dump_long_ad(char *prefix, struct long_ad *adr) {
	printf("%s at sector %d within partion space %d for %d bytes\n", prefix,
			udf_rw32(adr->loc.lb_num), udf_rw16(adr->loc.part_num),
			udf_rw32(adr->len)
			);
}


void udf_dump_id(char *prefix, int len, char *id, struct charspec *chsp) {
	uint16_t  raw_name[1024];
	uint16_t *pos, ch;
	int       ucode_chars;

	if (prefix) printf("%s ", prefix);
	if ((chsp->type == 0) && (strcmp(chsp->inf, "OSTA Compressed Unicode") == 0)) {
		/* print the identifier using the OSTA compressed unicode */
		printf("`");
		ucode_chars = udf_UncompressUnicode(len, id, raw_name);
		for (pos = raw_name; ucode_chars > 0; pos++, ucode_chars--) {
			ch = *pos;		/* OSTA code decompresses to machine endian */
			if (!ch) break;
			if ((ch < 32) || (ch > 255)) {
				printf("[%d]", ch);
			} else {
				printf("%c", ch & 255);
			};
		};
		printf("`");
	} else {
		printf("(roughly) `%s`", id+1);
	};
	if (prefix) printf("\n");
}


void udf_dump_volume_name(char *prefix, struct udf_log_vol *udf_log_vol) {
	if (prefix) printf("%s%s", prefix, udf_log_vol->primary->udf_session->session_offset?" (local) ":" ");
	udf_dump_id(NULL, 128, udf_log_vol->primary->pri_vol->volset_id, &udf_log_vol->primary->pri_vol->desc_charset);
	printf(":");
	udf_dump_id(NULL,  32, udf_log_vol->primary->pri_vol->vol_id, &udf_log_vol->primary->pri_vol->desc_charset);
	printf(":");
	udf_dump_id(NULL, 128, udf_log_vol->log_vol->logvol_id, &udf_log_vol->log_vol->desc_charset);
	if (prefix) printf("\n");
}


/******************************************************************************************
 *
 * UDF tag checkers and size calculator
 *
 ******************************************************************************************/


/* not used extensively enough yet */
int udf_check_tag(union dscrptr *dscr) {
	struct desc_tag *tag = &dscr->tag;
	uint8_t *pos, sum, cnt;

	/* check TAG header checksum */
	pos = (uint8_t *) tag;
	sum = 0;

	for(cnt = 0; cnt < 16; cnt++) {
		if (cnt != 4) sum += *pos;
		pos++;
	};
	if (sum != tag->cksum) {
		/* bad tag header checksum; this is not a valid tag */
		DEBUG(printf("Bad checksum\n"));
		return EINVAL;
	};
	return 0;
}


int udf_check_tag_payload(union dscrptr *dscr) {
	struct desc_tag *tag = &dscr->tag;
	uint16_t crc;

	/* check payload CRC if applicable */
	if (udf_rw16(tag->desc_crc_len) == 0) return 0;

	crc = udf_cksum(((uint8_t *) tag) + UDF_DESC_TAG_LENGTH, udf_rw16(tag->desc_crc_len));
	if (crc != udf_rw16(tag->desc_crc)) {
		DEBUG(printf("CRC bad\n"));
		/* bad payload CRC; this is a broken tag */
		return EINVAL;
	};

	return 0;
}


/* assumes sector number of descriptor to be saved allready present */
int udf_validate_tag_sums(union dscrptr *dscr) {
	struct desc_tag *tag = &dscr->tag;
	uint16_t crc;
	uint8_t *pos, sum, cnt;

	/* check payload CRC if applicable */
	if (udf_rw16(tag->desc_crc_len) > 0) {
		crc = udf_cksum(((uint8_t *) tag) + UDF_DESC_TAG_LENGTH, udf_rw16(tag->desc_crc_len));
		tag->desc_crc = udf_rw16(crc);
	};

	/* check TAG header checksum */
	pos = (uint8_t *) tag;
	sum = 0;

	for(cnt = 0; cnt < 16; cnt++) {
		if (cnt != 4) sum += *pos;
		pos++;
	};
	tag->cksum = sum;	/* 8 bit */

	return 0;
}


int udf_check_tag_presence(union dscrptr *dscr, int TAG) {
	struct desc_tag *tag = &dscr->tag;
	int error;

	error = udf_check_tag(dscr);
	if (error) return error;

	if (udf_rw16(tag->id) != TAG) {
		DEBUG(fprintf(stderr, "looking for tag %d but found %d\n", TAG, udf_rw16(tag->id)));
		return ENOENT;
	};

	return 0;
}



/*
 * for malloc() purposes ... rather have an upperlimit than an exact size
 */
int udf_calc_tag_malloc_size(union dscrptr *dscr) {
	uint32_t size, tag_id;

	tag_id = udf_rw16(dscr->tag.id);

	switch (tag_id) {
		case TAGID_LOGVOL :
			size  = sizeof(struct logvol_desc) - 1;					/* maps[1]		*/
			size += udf_rw32(dscr->lvd.mt_l);
			break;
		case TAGID_UNALLOC_SPACE :
			size  = sizeof(struct unalloc_sp_desc) - sizeof(struct extent_ad);	/* alloc_desc[1]	*/
			size += udf_rw32(dscr->usd.alloc_desc_num) * sizeof(struct extent_ad);
			break;
		case TAGID_FID :
			size = UDF_FID_SIZE + dscr->fid.l_fi + udf_rw16(dscr->fid.l_iu);
			size = (size + 3) & ~3;
			return size;		/* RETURN !! */
		case TAGID_LOGVOL_INTEGRITY :
			size  = sizeof(struct logvol_int_desc) - sizeof(uint32_t);		/* tables[1]		*/
			size += udf_rw32(dscr->lvid.l_iu);
			size += (2 * udf_rw32(dscr->lvid.num_part) * sizeof(uint32_t));
			break;
		case TAGID_SPACE_BITMAP :
			size  = sizeof(struct space_bitmap_desc) - 1;				/* data[1]		*/
			size += udf_rw32(dscr->sbd.num_bytes);
			break;
		case TAGID_SPARING_TABLE :
			size  = sizeof(struct udf_sparing_table) - sizeof(struct spare_map_entry);	/* entries[1]	*/
			size += udf_rw16(dscr->spt.rt_l) * sizeof(struct spare_map_entry);
			break;
		case TAGID_FENTRY :
			size  = sizeof(struct file_entry);
			size += udf_rw32(dscr->fe.l_ea) + udf_rw32(dscr->fe.l_ad)-1;		/* data[0] 		*/
			break;
		case TAGID_EXTFENTRY :
			size  = sizeof(struct extfile_entry);
			size += udf_rw32(dscr->efe.l_ea) + udf_rw32(dscr->efe.l_ad)-1;		/* data[0]		*/
			break;
		case TAGID_FSD :
			size  = sizeof(struct fileset_desc);
			break;
		default :
			size = sizeof(union dscrptr);
			break;
	};

	return (size / UDF_SECTOR_SIZE + 1) * UDF_SECTOR_SIZE;
}



/******************************************************************************************
 *
 * Logical to physical adres transformation
 *
 ******************************************************************************************/


/* convert (udf_log_vol, vpart_num) to a udf_partion structure */
int udf_logvol_vpart_to_partition(struct udf_log_vol *udf_log_vol, uint32_t vpart_num, struct udf_part_mapping **udf_part_mapping_ptr, struct udf_partition **udf_partition_ptr) {
	struct udf_volumeset	 *udf_volumeset;
	struct udf_partition	 *udf_partition;
	struct udf_part_mapping  *udf_part_mapping;
	uint32_t		  part_num;

	assert(udf_log_vol);
	assert(!SLIST_EMPTY(&udf_log_vol->part_mappings));

	/* clear result */
	if (udf_part_mapping_ptr) *udf_part_mapping_ptr = NULL;
	if (udf_partition_ptr)    *udf_partition_ptr    = NULL;

	/* bounds checking for virtual partition number */
	if (vpart_num >= udf_log_vol->num_part_mappings) {
		printf("\t\t\tPartition number %d outside bounds!\n", vpart_num);
		return EINVAL;
	};

	/* map the requested partition map to the physical udf partition */
	SLIST_FOREACH(udf_part_mapping, &udf_log_vol->part_mappings, next_mapping) {
		if (udf_part_mapping->udf_virt_part_num == vpart_num) break;
	};

	assert(udf_part_mapping);	/* call me paranoid */
	part_num = udf_part_mapping->udf_phys_part_num;

	/* search for the physical partition information */
	udf_volumeset = udf_log_vol->primary->volumeset;
	SLIST_FOREACH(udf_partition, &udf_volumeset->parts, next_partition) {
		if (udf_rw16(udf_partition->partition->part_num) == part_num) break;
	};

	if (!udf_partition) {
		printf("\t\t\tNo information known about partition %d yet!\n", part_num);
		printf("\t\t\t\tPlease insert volume %d of this volumeset and try again\n", udf_part_mapping->vol_seq_num);
		return ENOENT;
	};

	if (udf_part_mapping_ptr) *udf_part_mapping_ptr = udf_part_mapping;
	if (udf_partition_ptr)    *udf_partition_ptr    = udf_partition;
	return 0;
}


/* no recursive translations yet (UDF OK) */
/* All translation is done in 64 bits to prevent bitrot */
int udf_translate_addr(struct udf_part_mapping *udf_part_mapping, struct udf_partition *udf_partition, uint64_t offset, uint64_t *eff_addr, uint64_t *trans_valid_len) {
	struct spare_map_entry	 *sp_entry;

	uint64_t		  part_start, part_length;
	uint64_t		  eff_sector, eff_offset;
	uint64_t		  trans_sector;
	uint32_t		  entry, entries;

	uint64_t		  packet_sector, packet_offset;
	uint64_t		  packet_len;

	uint32_t		  vat_entries, *vat_pos;

	assert(udf_part_mapping);
	assert(udf_partition);
	assert(eff_addr);
	assert(trans_valid_len);

	*eff_addr        = 0;
	*trans_valid_len = 0;

	part_start  = udf_rw32(udf_partition->partition->start_loc);
	part_length = udf_rw32(udf_partition->partition->part_len);

	/* get the offset (in bytes) in the partition and check its validity */
	if (offset >= (part_start + part_length) * UDF_SECTOR_SIZE) {
		printf("\t\toffset %d is outside partition %d!\n", (uint32_t) offset, udf_rw16(udf_partition->partition->part_num));
		return EFAULT;
	};

	/* do the address translations based on the partition mapping type */
	switch (udf_part_mapping->udf_part_mapping_type) {
		case UDF_PART_MAPPING_PHYSICAL :
			/* nothing to be done; physical is logical */
			*eff_addr	 = part_start*UDF_SECTOR_SIZE  + offset;
			*trans_valid_len = part_length*UDF_SECTOR_SIZE - offset;	/* rest of partition */
			return 0;

		case UDF_PART_MAPPING_VIRTUAL :
			vat_entries = udf_part_mapping->vat_entries;
			vat_pos = (uint32_t *) udf_part_mapping->vat_translation;

			/* this translation is dependent on partition relative sector numbers */
			eff_sector = offset / UDF_SECTOR_SIZE;
			eff_offset = offset - eff_sector * UDF_SECTOR_SIZE;

			trans_sector     = vat_pos[eff_sector];
			*eff_addr        = (part_start + trans_sector)*UDF_SECTOR_SIZE + eff_offset;
			*trans_valid_len = UDF_SECTOR_SIZE - eff_offset;			/* maximum one sector */
			return 0;

		case UDF_PART_MAPPING_SPARABLE :
			/* transform on packet-length base */
			packet_len = udf_rw16(udf_part_mapping->udf_pmap->pms.packet_len);
			entries    = udf_rw16(udf_part_mapping->sparing_table->rt_l);

			*eff_addr     = part_start*UDF_SECTOR_SIZE + offset;
			packet_sector = packet_len * (*eff_addr / (packet_len * UDF_SECTOR_SIZE));
			packet_offset = *eff_addr - (packet_sector * UDF_SECTOR_SIZE);		/* within packet */

			/* translate this packet */
			trans_sector = packet_sector;
			sp_entry = &udf_part_mapping->sparing_table->entries[0];
			for(entry = 0; entry < entries; entry++) {
				if (udf_rw32(sp_entry->org) - packet_sector == 0) {
					trans_sector = udf_rw32(sp_entry->map);
					break;
				};
				sp_entry++;
			};
			*eff_addr        = trans_sector*UDF_SECTOR_SIZE + packet_offset;
			*trans_valid_len = packet_len * UDF_SECTOR_SIZE - packet_offset;	/* maximum one packet */
			return 0;

		case UDF_PART_MAPPING_META :
			/* 
			 * XXX meta partition mapping unsupported still due to
			 * change of read_file_part structure ... FIX ME XXX
			 */
			printf("ERR!!! unsupported META partion mapping while translating address\n");
			/* how do we do this?? we need to get a file offset??  how ?? */
			/* idea : just follow the file-descriptor :-/ */
			return EFAULT;
		case UDF_PART_MAPPING_ERROR :
		default :
			break;
	};
	printf("Unsupported or bad mapping %d; can't translate\n", udf_part_mapping->udf_part_mapping_type);

	return EFAULT;
}


/******************************************************************************************
 *
 * Descriptor readers and writers 
 *
 ******************************************************************************************/


/*
 * Read in an descriptor in either logvol space or in session space determined
 * by the specification of log_vol.
 *
 * In logvol space, lb_num specifies the logical block number in the logical
 * volume wich can be bigger than a sector.
 *
 * In session space, lb_num specifies a distinct sector.
 *
 * The function returns the read in descriptor blob and its length; it deals
 * with both short and long descriptors.
 */

int udf_read_descriptor(struct udf_log_vol *log_vol, uint32_t vpart_num, struct udf_session *session, uint32_t lb_num, char *what, uint32_t cache_flags, union dscrptr **dscr, uint32_t *length) {
	union dscrptr	*cur_dscr, *new_dscr;
	struct uio	 uio;
	struct iovec	 iovec;
	void		*sector0;
	uint32_t	 cur_length, new_length;
	int		 error;

	assert(dscr);
	if (length) *length = 0;
	*dscr   = NULL;

	cur_length = UDF_SECTOR_SIZE;		/* All discriptors have minimum size of UDF_SECTOR_LENGTH for easier updating; could become 512 */
	sector0    = malloc(cur_length);
	cur_dscr   = sector0;
	if (!sector0) {
		printf("\t\t\tOut of memory claiming memory for %s\n", what);
		return ENOMEM;
	};

	/* move to uio_newuio(struct uio *uio) with fixed length uio_iovcnt? */
	bzero(&uio, sizeof(struct uio));
	uio.uio_rw     = UIO_WRITE;		/* WRITE into this space */
	uio.uio_iovcnt = 1;
	uio.uio_iov    = &iovec;

	/* fill in our sector to be read into the uio structure */
	uio.uio_offset = 0;			/* begin at the start	*/
	uio.uio_resid  = cur_length;		/* we expect one sector	*/
	uio.uio_iov->iov_base = sector0;	/* fill in IO vector	*/
	uio.uio_iov->iov_len  = cur_length;

	/* start reading in sector; read at offset 0 into the logic block */

	if (log_vol) {
		error = udf_read_logvol_bytes_uio(log_vol, vpart_num, lb_num, 0, cur_length, what, &uio, cache_flags);
	} else {
		error = udf_read_session_bytes_uio(session, lb_num, 0, cur_length, what, &uio, cache_flags);
	};

	if (!error) error = udf_check_tag(cur_dscr);
	if (!error) {
		new_length = udf_calc_tag_malloc_size(cur_dscr);

		DEBUG(
			if (new_length < udf_rw16(cur_dscr->tag.desc_crc_len) + UDF_DESC_TAG_LENGTH) {
				printf("UDF warning: reading in %s for %d bytes but descriptor crc len is %d bytes\n", what, new_length,
						udf_rw16(cur_dscr->tag.desc_crc_len) + UDF_DESC_TAG_LENGTH);
				udf_dump_descriptor(cur_dscr);
			};
		);

		if (new_length > cur_length) {
			/* extent the current descriptor ; read in more */
			new_dscr   = malloc(new_length);
			if (new_dscr) {
				/* copy read-in stuff into the new allocated space */
				memcpy(new_dscr, sector0, cur_length);
				free(sector0);

				uio.uio_iovcnt = 1;
				uio.uio_offset = 0;						/* begin at start (of blob)	*/
				uio.uio_resid  = new_length - cur_length;			/* we expect the rest	 	*/
				uio.uio_iov->iov_base = ((void *) new_dscr) + cur_length;	/* fill in IO vector 		*/
				uio.uio_iov->iov_len  = new_length - cur_length;

				/* read at offset cur_length in current logic block */
				if (log_vol) {
					error = udf_read_logvol_bytes_uio(log_vol, vpart_num, lb_num, cur_length, new_length - cur_length, what, &uio, cache_flags);
				} else {
					error = udf_read_session_bytes_uio(session, lb_num, cur_length, new_length - cur_length, what, &uio, cache_flags);
				};
				cur_dscr   = new_dscr;
				cur_length = new_length;
			} else {
				free(sector0);
			};
		};
	};
	if (!error) {
		*dscr   = cur_dscr;
		if (length) *length = cur_length;	/* if requested return length */
		error = udf_check_tag(*dscr);
		if (!error) error = udf_check_tag_payload(*dscr);
	};
	return error;
}


/* Reads descriptor as in currenly recorded on disc or as is in the cache */
int udf_read_session_descriptor(struct udf_session *udf_session, uint32_t lb_num, char *what, union dscrptr **dscr, uint32_t *length) {
	uint32_t cache_flags;

	cache_flags = 0;			/* read in complete cache line */
	return udf_read_descriptor(NULL, 0, udf_session, lb_num, what, cache_flags, dscr, length);
}


/* Reads descriptor as in currenly recorded on disc or as is in the cache */
int udf_read_logvol_descriptor(struct udf_log_vol *udf_log_vol, uint32_t vpart_num, uint32_t lb_num, char *what, union dscrptr **dscr, uint32_t *length) {
	uint32_t cache_flags;

	cache_flags = UDF_CACHEFLAG_METADATA;	/* read in complete cache line */
	return udf_read_descriptor(udf_log_vol, vpart_num, NULL, lb_num, what, cache_flags, dscr, length);
}


static int udf_write_descriptor(struct udf_log_vol *log_vol, uint32_t vpart_num, struct udf_session *session, uint32_t lb_num_write, uint32_t lb_num, char *what, union dscrptr *dscr, uint32_t length, uint32_t cache_flags, struct udf_wrcallback *wrcallback) {
	struct uio	 uio;
	struct iovec	 iovec;
	uint64_t	 offset;
	uint32_t         dscr_length;
	uint32_t	 lb_size, eff_sector;
	int		 error;

	assert(dscr);
	dscr_length = udf_calc_tag_malloc_size(dscr);
	if (length && dscr_length > length) {
		printf("UDF: write descriptor called with length < descriptor length\n");
		return EIO;
	};
	length = dscr_length;

	/* move to uio_newuio(struct uio *uio) with fixed length uio_iovcnt? */
	bzero(&uio, sizeof(struct uio));
	uio.uio_rw     = UIO_READ;		/* READ from this space */
	uio.uio_iovcnt = 1;
	uio.uio_iov    = &iovec;

	/* fill in our sector to be read into the uio structure */
	uio.uio_offset = 0;			/* begin at the start              */
	uio.uio_resid  = length;		/* we writeout the complete length */
	uio.uio_iov->iov_base = dscr;		/* fill in IO vector               */
	uio.uio_iov->iov_len  = length;

	/* write out descriptor */
	wrcallback->flags |= UDF_WRCALLBACK_FLAG_DESCRIPTOR;
	cache_flags       |= UDF_CACHEFLAG_NOREADIN;

	if (log_vol) {
		/* prepare descriptor for writing */
		/* not nice that we need to calculate effective sector here */
		lb_size     = udf_rw32(log_vol->log_vol->lb_size);
		offset      = (uint64_t) lb_num * lb_size;
		eff_sector  = offset / UDF_SECTOR_SIZE;
		dscr->tag.tag_loc = udf_rw32(eff_sector);
		udf_validate_tag_sums(dscr);

		/* wrcallback->function is given */
		wrcallback->lb_num    = lb_num;
		wrcallback->length    = length;
		wrcallback->vpart_num = vpart_num;
		error = udf_write_logvol_bytes_uio(log_vol, vpart_num, lb_num_write, 0, length, what, &uio, cache_flags, wrcallback);
	} else {
		/* prepare descriptor for writing */
		dscr->tag.tag_loc = udf_rw32(lb_num);
		udf_validate_tag_sums(dscr);

		/* wrcallback->function is given */
		wrcallback->lb_num    = lb_num;
		wrcallback->length    = length;
		error = udf_write_session_bytes_uio(session, lb_num_write, 0, length, what, &uio, cache_flags, wrcallback);
	};

	return error;
}


/* Write descriptor trough cache if present */
int udf_write_session_descriptor(struct udf_session *udf_session, uint32_t lb_num, char *what, union dscrptr *dscr, uint32_t length, struct udf_wrcallback *wrcallback) {
	uint32_t cache_flags;
	uint32_t lb_num_write;

	lb_num_write = lb_num;
	cache_flags  = UDF_CACHEFLAG_NOREADIN;

	return udf_write_descriptor(NULL, 0, udf_session, lb_num_write, lb_num, what, dscr, length, cache_flags, wrcallback);
}


int udf_write_partition_descriptor(struct udf_partition *udf_partition, uint32_t lb_num, char *what, union dscrptr *dscr, uint32_t length, struct udf_wrcallback *wrcallback) {
	uint32_t cache_flags;
	uint32_t lb_num_write;

	lb_num_write = udf_rw32(lb_num + udf_partition->partition->start_loc) + lb_num;
	cache_flags  = UDF_CACHEFLAG_NOREADIN;

	return udf_write_descriptor(NULL, 0, udf_partition->udf_session, lb_num_write, lb_num, what, dscr, length, cache_flags, wrcallback);
}


/* Write descriptor trough cache if present */
int udf_write_logvol_descriptor(struct udf_log_vol *udf_log_vol, uint32_t vpart_num, uint32_t lb_num, char *what, union dscrptr *dscr, uint32_t length, struct udf_wrcallback *wrcallback) {
	uint32_t cache_flags;
	uint32_t lb_num_write;

	lb_num_write = lb_num;
	cache_flags  = UDF_CACHEFLAG_NOREADIN | UDF_CACHEFLAG_METADATA;

	return udf_write_descriptor(udf_log_vol, vpart_num, NULL, lb_num_write, lb_num, what, dscr, length, cache_flags, wrcallback);
}


/******************************************************************************************
 *
 * udf_node creator, destructor and syncer
 *
 ******************************************************************************************/


static mode_t udf_perm_to_unix_mode(uint32_t perm) {
	mode_t mode;

	mode  = ((perm & UDF_FENTRY_PERM_USER_MASK)      );
	mode |= ((perm & UDF_FENTRY_PERM_GRP_MASK  ) >> 2);
	mode |= ((perm & UDF_FENTRY_PERM_OWNER_MASK) >> 4);

	return mode;
}


static uint32_t unix_mode_to_udf_perm(mode_t mode) {
	uint32_t perm;
	
	perm  = ((mode & S_IRWXO)     );
	perm |= ((mode & S_IRWXG) << 2);
	perm |= ((mode & S_IRWXU) << 4);
	perm |= ((mode & S_IWOTH) << 3);
	perm |= ((mode & S_IWGRP) << 5);
	perm |= ((mode & S_IWUSR) << 7);

	return perm;
}
 

static void udf_node_get_fileinfo(struct udf_node *udf_node, union dscrptr *dscrptr) {
	struct stat          *stat;
	struct file_entry    *file_entry;
	struct extfile_entry *extfile_entry;
	struct timestamp     *atime, *mtime, *attrtime;
	uint64_t inf_len, unique_id;
	uint32_t uid, gid, udf_perm;
	uint16_t fe_tag;
	uint16_t udf_icbtag_flags, serial_num, link_cnt;
	uint8_t  filetype;

	assert(udf_node);
	assert(dscrptr);
	stat   = &udf_node->stat;

	/* check if its an normal file entry or a extended file entry ICB */
	fe_tag = udf_rw16(dscrptr->tag.id);
	if (fe_tag == TAGID_FENTRY) {
		file_entry = &dscrptr->fe;
		//prev_direct_entries = udf_rw32(file_entry->icbtag.prev_num_dirs);
		//strat_param16  = udf_rw16(* (uint16_t *) (file_entry->icbtag.strat_param));
		//entries        = udf_rw16(file_entry->icbtag.max_num_entries);
		//strategy       = udf_rw16(file_entry->icbtag.strat_type);
		//data_length    = udf_rw32(file_entry->l_ad);
		//addr_type      = udf_rw16(file_entry->icbtag.flags) & UDF_ICB_TAG_FLAGS_ALLOC_MASK;
		//pos            = &file_entry->data[0] + udf_rw32(file_entry->l_ea);
		filetype         = file_entry->icbtag.file_type;
		inf_len          = udf_rw64(file_entry->inf_len);
		uid              = udf_rw32(file_entry->uid);
		gid              = udf_rw32(file_entry->gid);
		udf_perm         = udf_rw32(file_entry->perm);
		serial_num       = udf_rw16(file_entry->tag.serial_num);
		udf_icbtag_flags = udf_rw16(file_entry->icbtag.flags);
		link_cnt         = udf_rw16(file_entry->link_cnt);
		unique_id        = udf_rw64(file_entry->unique_id);
		atime            = &file_entry->atime;
		mtime            = &file_entry->mtime;
		attrtime         = &file_entry->attrtime;
	} else if (fe_tag == TAGID_EXTFENTRY) {
		extfile_entry = &dscrptr->efe;
		//prev_direct_entries = udf_rw32(extfile_entry->icbtag.prev_num_dirs);
		//strat_param16  = udf_rw16(* (uint16_t *) (extfile_entry->icbtag.strat_param));
		//entries        = udf_rw16(extfile_entry->icbtag.max_num_entries);
		//strategy       = udf_rw16(extfile_entry->icbtag.strat_type);
		//data_length    = udf_rw32(extfile_entry->l_ad);
		//addr_type      = udf_rw16(extfile_entry->icbtag.flags) & UDF_ICB_TAG_FLAGS_ALLOC_MASK;
		//pos            = &extfile_entry->data[0] + udf_rw32(extfile_entry->l_ea);
		filetype         = extfile_entry->icbtag.file_type;
		inf_len          = udf_rw64(extfile_entry->inf_len);
		uid              = udf_rw32(extfile_entry->uid);
		gid              = udf_rw32(extfile_entry->gid);
		udf_perm         = udf_rw32(extfile_entry->perm);
		serial_num       = udf_rw16(extfile_entry->tag.serial_num);
		udf_icbtag_flags = udf_rw16(extfile_entry->icbtag.flags);
		link_cnt         = udf_rw16(extfile_entry->link_cnt);	/* how many FID's are linked to this (ext)fentry	*/
		unique_id        = udf_rw64(extfile_entry->unique_id);	/* unique file ID					*/
		atime            = &extfile_entry->atime;
		mtime            = &extfile_entry->mtime;
		attrtime         = &extfile_entry->attrtime;
	} else {
		printf("udf_node_set_file_infot :  help! i can't be here!!! i got a %d tag\n", fe_tag);
		udf_dump_descriptor(dscrptr);
		return;
	};

	/* fill in (parts of) the stat structure */
	/* XXX important info missing still like access mode, times etc. XXX */
	bzero(stat, sizeof(struct stat));
	stat->st_uid   = uid;
	stat->st_gid   = gid;
	stat->st_size  = inf_len;
	stat->st_mode  = udf_perm_to_unix_mode(udf_perm);	/* CONVERT from udf_perm */
	stat->st_mode |= (filetype == UDF_ICB_FILETYPE_DIRECTORY) ? S_IFDIR : 0;

	udf_node->udf_filetype = filetype;

	udf_node->serial_num       = serial_num;
	udf_node->udf_icbtag_flags = udf_icbtag_flags;
	udf_node->link_cnt         = link_cnt;		/* how many FID's are linked to this (ext)fentry	*/
	udf_node->unique_id        = unique_id;		/* unique file ID					*/

	return;
}


static void udf_node_set_fileinfo(struct udf_node *udf_node, union dscrptr *dscrptr) {
	struct stat          *stat;
	struct file_entry    *file_entry;
	struct extfile_entry *extfile_entry;
	struct timestamp     *atime, *mtime, *attrtime;
	uint64_t inf_len, unique_id;
	uint32_t uid, gid, udf_perm;
	uint16_t fe_tag, serial_num, link_cnt;
	uint8_t  filetype;

	assert(udf_node);
	assert(dscrptr);
	stat   = &udf_node->stat;

	/* set (parts of) the stat structure */
	/* XXX important info missing still like times etc. XXX */
	uid      = stat->st_uid;
	gid      = stat->st_gid;
	inf_len  = stat->st_size;
	udf_perm = unix_mode_to_udf_perm(stat->st_mode);	/* conversion to UDF perm. */

	filetype   = udf_node->udf_filetype;
	unique_id  = udf_node->unique_id;
	serial_num = udf_node->serial_num;
	link_cnt   = udf_node->link_cnt;

	/* check if its to be written in an normal file entry or a extended file entry ICB */
	fe_tag = udf_rw16(dscrptr->tag.id);
	if (fe_tag == TAGID_FENTRY) {
		file_entry = &dscrptr->fe;
		//prev_direct_entries = udf_rw32(file_entry->icbtag.prev_num_dirs);
		//strat_param16  = udf_rw16(* (uint16_t *) (file_entry->icbtag.strat_param));
		//entries        = udf_rw16(file_entry->icbtag.max_num_entries);
		//strategy       = udf_rw16(file_entry->icbtag.strat_type);
		//data_length    = udf_rw32(file_entry->l_ad);
		//addr_type      = udf_rw16(file_entry->icbtag.flags) & UDF_ICB_TAG_FLAGS_ALLOC_MASK;
		//pos            = &file_entry->data[0] + udf_rw32(file_entry->l_ea);
		file_entry->icbtag.file_type = filetype;
		file_entry->inf_len         = udf_rw64(inf_len);
		file_entry->uid             = udf_rw32(uid);
		file_entry->gid             = udf_rw32(gid);
		file_entry->perm            = udf_rw32(udf_perm);
		file_entry->tag.serial_num  = udf_rw16(serial_num);
		file_entry->link_cnt        = udf_rw16(link_cnt);
		file_entry->unique_id       = udf_rw64(unique_id);
		atime                       = &file_entry->atime;
		mtime                       = &file_entry->mtime;
		attrtime                    = &file_entry->attrtime;
	} else if (fe_tag == TAGID_EXTFENTRY) {
		extfile_entry = &dscrptr->efe;
		//prev_direct_entries = udf_rw32(extfile_entry->icbtag.prev_num_dirs);
		//strat_param16  = udf_rw16(* (uint16_t *) (extfile_entry->icbtag.strat_param));
		//entries        = udf_rw16(extfile_entry->icbtag.max_num_entries);
		//strategy       = udf_rw16(extfile_entry->icbtag.strat_type);
		//data_length    = udf_rw32(extfile_entry->l_ad);
		//addr_type      = udf_rw16(extfile_entry->icbtag.flags) & UDF_ICB_TAG_FLAGS_ALLOC_MASK;
		//pos            = &extfile_entry->data[0] + udf_rw32(extfile_entry->l_ea);
		extfile_entry->icbtag.file_type = filetype;
		extfile_entry->inf_len        = udf_rw64(inf_len);
		extfile_entry->uid            = udf_rw32(uid);
		extfile_entry->gid            = udf_rw32(gid);
		extfile_entry->perm           = udf_rw32(udf_perm);
		extfile_entry->tag.serial_num = udf_rw16(serial_num);
		extfile_entry->link_cnt       = udf_rw16(link_cnt);	/* how many FID's are linked to this (ext)fentry	*/
		extfile_entry->unique_id      = udf_rw64(unique_id);	/* unique file ID					*/
		atime          = &extfile_entry->atime;
		mtime          = &extfile_entry->mtime;
		attrtime       = &extfile_entry->attrtime;
	} else {
		printf("udf_node_set_file_info :  help! i can't be here!!! i got a %d tag\n", fe_tag);
		udf_dump_descriptor(dscrptr);
		return;
	};
	/* FILL in {atime, mtime, attrtime} TIMES! */

	return;
}


/* with 32 bits ino_t, a maximum of about 8 TB discs are supported (1<<32) * 2KB*/
inline ino_t udf_calc_hash(struct long_ad *icbptr) {
	return (ino_t) udf_rw32(icbptr->loc.lb_num);
}


void udf_insert_node_in_hash(struct udf_node *udf_node) {
	struct udf_mountpoint *mountable;
	uint32_t     bucket;
	ino_t        hashkey;
	struct long_ad icb;

	icb.loc.lb_num = udf_rw32(TAILQ_FIRST(&udf_node->dscr_allocs)->lb_num);

	mountable         = udf_node->mountpoint;
	hashkey           = udf_calc_hash(&icb);
	udf_node->hashkey = hashkey;
	bucket            = hashkey & UDF_INODE_HASHMASK;
	LIST_INSERT_HEAD(&mountable->udf_nodes[bucket], udf_node, next_node);
}


void udf_dispose_udf_node(struct udf_node *udf_node) {
	struct udf_node *lookup;
	uint32_t bucket;
	ino_t    hashkey;

	if (!udf_node) return;
	/* if internal data, free it */
	if (udf_node->file_data) free(udf_node->file_data);

	/* delete dscr_allocs queue */
	while (TAILQ_FIRST(&udf_node->dscr_allocs) != NULL) {
		TAILQ_REMOVE(&udf_node->dscr_allocs, TAILQ_FIRST(&udf_node->dscr_allocs), next_alloc);
	};

	/* if it has an allocation queue, delete them all */
	while (TAILQ_FIRST(&udf_node->alloc_entries) != NULL) {
		TAILQ_REMOVE(&udf_node->alloc_entries, TAILQ_FIRST(&udf_node->alloc_entries), next_alloc);
	};

	/* if its part of a mountpoint, delete it in its hash table */
	if (udf_node->mountpoint) {
		hashkey = udf_node->hashkey;
		bucket  = hashkey & UDF_INODE_HASHMASK;
		LIST_FOREACH(lookup, &udf_node->mountpoint->udf_nodes[bucket], next_node) {
			if (lookup->hashkey == hashkey) {
				DEBUG(printf("removal of udf_node from the hash table\n"));
				LIST_REMOVE(lookup, next_node);
				break;
			};
		};
	};

	/* free the node */
	free(udf_node);
}


uint64_t udf_increment_unique_id(struct udf_log_vol *udf_log_vol) {
	uint64_t unique_id, next_unique_id;

	/* lock? */
	unique_id = udf_log_vol->next_unique_id;

	/* increment according to UDF 3/3.2.1.1 */
	next_unique_id = unique_id + 1;
	if (((next_unique_id << 32) >> 32) < 16) next_unique_id |= 16;

	udf_log_vol->next_unique_id = next_unique_id;
	DEBUG(printf("next unique_id <-- %"PRIu64"\n", udf_log_vol->next_unique_id));

	return unique_id;
}


int udf_init_udf_node(struct udf_mountpoint *mountpoint, struct udf_log_vol *udf_log_vol, char *what, struct udf_node **udf_nodeptr) {
	struct udf_node *udf_node;

	udf_node = calloc(1, sizeof(struct udf_node));
	if (!udf_node) return ENOMEM;

	udf_node->addr_type = UDF_ICB_LONG_ALLOC;
//	udf_node->addr_type = UDF_ICB_INTERN_ALLOC; 				/* how much to allocate ? */
	udf_node->icb_len   = sizeof(struct long_ad);

	udf_node->serial_num       = 1;
	udf_node->udf_icbtag_flags = udf_rw16(UDF_ICB_LONG_ALLOC);
	udf_node->link_cnt         = 1;						/* how many FID's are linked to this node */
	udf_node->unique_id        = 0;						/* unique file ID unknown		  */

	udf_node->mountpoint  = mountpoint;
	udf_node->udf_log_vol = udf_log_vol;
	TAILQ_INIT(&udf_node->dscr_allocs);
	TAILQ_INIT(&udf_node->alloc_entries);
	UDF_MUTEX_INIT(&udf_node->alloc_mutex);
	pthread_rwlock_init(&udf_node->udf_node_lock, NULL);

	/* give it a name just in case for later identifying */
	sprintf(udf_node->dirent.d_name, "Anonymous udf_node for %s", what);

	*udf_nodeptr = udf_node;
	return 0;
}


int udf_allocate_udf_node_on_disc(struct udf_node *udf_node, int filetype) {
	struct udf_allocentry *alloc_entry;
	uint32_t  lb_num, lb_size;
	uint16_t  vpart_num;
	int       error;

	assert(udf_node);
	assert(udf_node->udf_log_vol);
	assert(udf_node->udf_log_vol->log_vol);

	lb_size   = udf_rw32(udf_node->udf_log_vol->log_vol->lb_size);
	assert(lb_size);

	/* pre-allocated node; in strat 4 we need it for its our hash for now!*/
	error = udf_request_free_logblks(udf_node->udf_log_vol, 1, 1, "New direntry", &vpart_num, &lb_num, NULL);
	alloc_entry = calloc(1, sizeof(struct udf_allocentry));
	if (error || !alloc_entry) {
		if (alloc_entry) free(alloc_entry);
		return ENOMEM;
	};

	alloc_entry->len        = lb_size;
	alloc_entry->vpart_num  = vpart_num;
	alloc_entry->lb_num     = lb_num;
	alloc_entry->flags      = 0;
	TAILQ_INSERT_TAIL(&udf_node->dscr_allocs, alloc_entry, next_alloc);

	udf_node->udf_filetype  = filetype;
	udf_node->dirent.d_type = udf_translate_icb_filetype_to_dirent_filetype(filetype);
	udf_node->unique_id     = udf_increment_unique_id(udf_node->udf_log_vol);

	assert(error == 0);
	return error;
}


/* note: the udf_node (inode) is not stored in a hashtable! hash value is still invalid */
int udf_readin_anon_udf_node(struct udf_log_vol *udf_log_vol, union dscrptr *given_dscrptr, struct long_ad *udf_icbptr, char *what, struct udf_node **udf_nodeptr) {
	union  dscrptr  	*dscrptr;
	struct udf_node 	*udf_node;
	struct udf_allocentry	*alloc_entry;
	struct udf_allocentry	*cur_alloc, *next_alloc;
	struct file_entry	*file_entry;
	struct extfile_entry	*extfile_entry;
	struct alloc_ext_entry	*alloc_ext_entry;
	struct long_ad		*l_ad;
	struct short_ad		*s_ad;
	uint64_t		 inf_len, calculated_len;
	uint32_t		 entries, data_length, data_space_avail;
	uint32_t		 fe_tag;
	uint32_t		 len, lb_num, vpart_num;
	uint32_t		 icb_len;
	int16_t			 addr_type;
	uint8_t			*file_data_pos;
	uint8_t			*pos;
	uint8_t			 flags;
	int			 error, advance_sector;

	if ((udf_icbptr->loc.lb_num == 0) && (udf_icbptr->loc.part_num == 0) && (udf_icbptr->len == 0)) return ENOENT;

	DEBUG(printf("udf_readin_anon_udf_node for %s\n", what));

	error = udf_init_udf_node(/*mountpoiunt*/ NULL, udf_log_vol, what, &udf_node);
	if (error) return error;

	*udf_nodeptr = udf_node;

	assert(udf_log_vol);
	/* read in descriptor if not provided to us */
	len       = UDF_SECTOR_SIZE;
	lb_num    = udf_rw32(udf_icbptr->loc.lb_num);
	vpart_num = udf_rw16(udf_icbptr->loc.part_num);
	if (!given_dscrptr) {
		error = udf_read_logvol_descriptor(udf_log_vol, vpart_num, lb_num, what, &dscrptr, NULL);
		if (error) {
			*udf_nodeptr = NULL;
			return error;
		};
	} else {
		dscrptr = given_dscrptr;
	};

	fe_tag = udf_rw16(dscrptr->tag.id);
	if (fe_tag != TAGID_FENTRY && fe_tag != TAGID_EXTFENTRY) {
		/* something wrong with this address; it doesn't start with a file entry */
		printf("UDF: bad udf_node for %s; got a %d tag\n", what, fe_tag);
		if (dscrptr != given_dscrptr) free(dscrptr);
		udf_dispose_udf_node(udf_node);
		*udf_nodeptr = NULL;
		return EFAULT;
	};

	/* get as much info as possible */
	udf_node_get_fileinfo(udf_node, dscrptr);

	/* reset pending write count */
	udf_node->pending_writes = 0;

	/* initialise various variables for extracting allocation information */
	inf_len = 0; data_length = 0; lb_num = 0; len = 0; vpart_num = 0; pos = NULL;
	data_space_avail = 0;

	next_alloc = calloc(1, sizeof(struct udf_allocentry));
	if (!next_alloc) {
		if (dscrptr != given_dscrptr) free(dscrptr);
		udf_dispose_udf_node(udf_node);
		*udf_nodeptr = NULL;
		return ENOMEM;
	};
	next_alloc->len       = UDF_SECTOR_SIZE;
	next_alloc->lb_num    = udf_rw32(udf_icbptr->loc.lb_num);
	next_alloc->vpart_num = udf_rw16(udf_icbptr->loc.part_num);

	entries = 1;
	file_data_pos  = NULL;
	calculated_len = 0;

	error = 0;
	addr_type = -1;
	do {
		/* keep a trace of the descriptor locations in this udf_node */
		UDF_VERBOSE_MAX(udf_dump_descriptor(dscrptr));

		cur_alloc = next_alloc;
		next_alloc = calloc(1, sizeof(struct udf_allocentry));
		if (!next_alloc) {
			if (dscrptr != given_dscrptr) free(dscrptr);
			udf_dispose_udf_node(udf_node);
			*udf_nodeptr = NULL;
			return ENOMEM;
		};
		memcpy(next_alloc, cur_alloc, sizeof(struct udf_allocentry));
		TAILQ_INSERT_TAIL(&udf_node->dscr_allocs, cur_alloc, next_alloc);

		/* process this allocation descriptor */
		/* note that we don't store the file descriptors */
		fe_tag = udf_rw16(dscrptr->tag.id);
		switch (fe_tag) {
			case TAGID_FENTRY :
				/* allocation descriptors follow this tag */
				file_entry       = &dscrptr->fe;
				entries          = udf_rw16(file_entry->icbtag.max_num_entries);
				data_length      = udf_rw32(file_entry->l_ad);
				pos              = &file_entry->data[0] + udf_rw32(file_entry->l_ea);
				inf_len          = udf_rw64(file_entry->inf_len);
				addr_type        = udf_rw16(file_entry->icbtag.flags) & UDF_ICB_TAG_FLAGS_ALLOC_MASK;
				data_space_avail = UDF_SECTOR_SIZE - sizeof(struct file_entry) - udf_rw32(file_entry->l_ea);
				/* process extended attributes */
				break;
			case TAGID_EXTFENTRY :
				/* allocation descriptors follow this tag */
				extfile_entry    = &dscrptr->efe;
				entries          = udf_rw16(extfile_entry->icbtag.max_num_entries);
				data_length      = udf_rw32(extfile_entry->l_ad);
				pos              = &extfile_entry->data[0] + udf_rw32(extfile_entry->l_ea);
				inf_len          = udf_rw64(extfile_entry->inf_len);
				addr_type        = udf_rw16(extfile_entry->icbtag.flags) & UDF_ICB_TAG_FLAGS_ALLOC_MASK;
				data_space_avail = UDF_SECTOR_SIZE - sizeof(struct extfile_entry) - udf_rw32(extfile_entry->l_ea);
				/* process extended attributes */
				break;
			case TAGID_ALLOCEXTENT :
				/* allocation descriptors follow this tag; treat as if continuation of a (ext)file entry*/
				alloc_ext_entry  = &dscrptr->aee;
				data_length      = udf_rw32(alloc_ext_entry->l_ad);
				pos		 = &alloc_ext_entry->data[0];
				assert(addr_type >= 0);
				break;
			case TAGID_INDIRECT_ENTRY :
				printf("create_anon_udf_node called with indirect entry; following chain\n");
				l_ad = &dscrptr->inde.indirect_icb;
				next_alloc->len       = udf_rw32(l_ad->len);
				next_alloc->lb_num    = udf_rw32(l_ad->loc.lb_num);
				next_alloc->vpart_num = udf_rw16(l_ad->loc.part_num);
				entries   = 1;		/* at least one more entry	*/
				advance_sector = 0;	/* don't advance to next sector */
				break;
			default:
				printf("read_file_part_extents: i can't be here! (tag = %d) in ICB hiargy of %s\n", fe_tag, what);
				udf_dump_descriptor(dscrptr);
				return EFAULT;
		};
		if (fe_tag != TAGID_INDIRECT_ENTRY) {
			udf_node->addr_type = addr_type;
			if (addr_type == UDF_ICB_INTERN_ALLOC) {
				udf_node->file_data = calloc(1, UDF_SECTOR_SIZE);
				error = (udf_node->file_data) ? 0 : ENOMEM;
				if (!error) {
					udf_node->file_data_length       = data_length;
					udf_node->file_data_alloc_length = data_space_avail;
					memcpy(udf_node->file_data, pos, inf_len);
				};
				if (dscrptr != given_dscrptr) free(dscrptr);
				return error;
			};

			icb_len = 0;
			switch (udf_node->addr_type) {
				case UDF_ICB_SHORT_ALLOC :
					icb_len   = sizeof(struct short_ad);
					break;
				case UDF_ICB_LONG_ALLOC  :
					icb_len   = sizeof(struct long_ad);
					break;
				default :
					printf("UDF encountered an unknown allocation type %d\n", udf_node->addr_type);
					break;
			};
			udf_node->icb_len = icb_len;
			advance_sector = 1;
			while (icb_len && data_length) {
				switch (udf_node->addr_type) {
					case UDF_ICB_SHORT_ALLOC  :
						s_ad = (struct short_ad *) pos;
						len       = udf_rw32(s_ad->len);
						lb_num    = udf_rw32(s_ad->lb_num);
						vpart_num = cur_alloc->vpart_num;
						break;
					case UDF_ICB_LONG_ALLOC   :
						l_ad = (struct long_ad *) pos;
						len       = udf_rw32(l_ad->len);
						lb_num    = udf_rw32(l_ad->loc.lb_num);
						vpart_num = udf_rw16(l_ad->loc.part_num);
						if (l_ad->impl.im_used.flags & UDF_ADIMP_FLAGS_EXTENT_ERASED) {
							printf("UDF: got a `extent erased' flag in a file's long_ad; ignoring\n");
						};
						break;
					default :
						printf("UDF encountered an unknown allocation type %d\n", udf_node->addr_type);
						break;
				};
				/* ecma-167 48.14.1.1 */
				flags = (uint8_t) ((uint32_t) (len >> 30) & 3);
				len   = len & ((1<<30)-1);
				if (flags == 3) {
					/* fill in next extent */
					next_alloc->len       = len;
					next_alloc->lb_num    = lb_num;
					next_alloc->vpart_num = vpart_num;
					advance_sector = 0;	/* don't advance to next sector */
					icb_len = data_length;	/* must be last */
					printf("Continuing extent flagged at vpart = %d, lb_num = %d, len = %d\n", (int) vpart_num, (int) lb_num, (int) len);
				} else {
					if (len) {
						alloc_entry = calloc(1, sizeof(struct udf_allocentry));
						if (!alloc_entry) {
							if (dscrptr != given_dscrptr) free(dscrptr);
							return ENOMEM;
						};
						alloc_entry->len       = len;
						alloc_entry->lb_num    = lb_num;
						alloc_entry->vpart_num = vpart_num;
						alloc_entry->flags     = flags;
						TAILQ_INSERT_TAIL(&udf_node->alloc_entries, alloc_entry, next_alloc);
					};
				};
				data_length    -= icb_len;
				calculated_len += len;
				pos += icb_len;
			}; /* while */
			if (advance_sector) {
				next_alloc->lb_num++;	/* advance one sector */
				entries--;
			};
		}; /* indirect ICB check */
		if (entries) {
			/* load in new dscrptr */
			if (dscrptr != given_dscrptr) free(dscrptr);
			error = udf_read_logvol_descriptor(udf_log_vol, next_alloc->vpart_num, next_alloc->lb_num, what, &dscrptr, NULL);
		};
	} while (entries && !error);

	/* error from reading next sector in extent is not considered an error */
	if (dscrptr != given_dscrptr && dscrptr) free(dscrptr);

	if (calculated_len != udf_node->stat.st_size) {
		printf("UDF: create node length mismatch; stated as %g but calculated as %g bytes. fixing\n", (double) udf_node->stat.st_size, (double) calculated_len);
		udf_node->stat.st_size = calculated_len;
	};

	return 0;
}


void udf_syncnode_callback(int reason, struct udf_wrcallback *wrcallback, int error, uint8_t *sectordata) {
	struct udf_node *udf_node = (struct udf_node *) wrcallback->structure;

	if (reason == UDF_WRCALLBACK_REASON_PENDING) {
		/* what to do? */
		return;
	};
	if (reason == UDF_WRCALLBACK_REASON_ANULATE) {
		/* what to do? */
		return;
	};
	assert(reason == UDF_WRCALLBACK_REASON_WRITTEN);

	DEBUG(
		printf("syncnode write callback called with sector data %p\n", sectordata);
		printf("\toffset    %08d\n", (int) wrcallback->offset);
		printf("\tlb_num    %08d\n", (int) wrcallback->lb_num);
		printf("\tlength    %08d\n", (int) wrcallback->length);
		printf("\terror     %d\n",   (int) error);
	);

	if (error) {
		printf("UDF error: write errors in syncnode not fixed! marking dirty again \n");
		udf_node_mark_dirty(udf_node);
		return;
	};
}


void udf_node_mark_dirty(struct udf_node *udf_node) {
	if (udf_node->dirty) return;

	if (udf_node->mountpoint) {
		UDF_MUTEX_LOCK(&udf_node->mountpoint->dirty_nodes_mutex);
			STAILQ_INSERT_TAIL(&udf_node->mountpoint->dirty_nodes, udf_node, next_dirty);
		UDF_MUTEX_UNLOCK(&udf_node->mountpoint->dirty_nodes_mutex);
	};
	udf_node->dirty = 1;
}


static void udf_node_unmark_dirty(struct udf_node *udf_node) {
	if (!udf_node->dirty) return;

	if (udf_node->mountpoint) {
		/* remove me just in case; why were we called otherwise? */
		UDF_MUTEX_LOCK(&udf_node->mountpoint->dirty_nodes_mutex);
			STAILQ_REMOVE(&udf_node->mountpoint->dirty_nodes, udf_node, udf_node, next_dirty);
		UDF_MUTEX_UNLOCK(&udf_node->mountpoint->dirty_nodes_mutex);
	};
	udf_node->dirty = 0;
}


/* VOP_FSYNC : data FDATASYNC */
int udf_sync_udf_node(struct udf_node *udf_node, char *why) {
	int cnt = 0;

	if (!udf_node->dirty) {
		/* Not dirty??!!! */
		udf_node_unmark_dirty(udf_node);
		return 0;
	};

	if (udf_node->pending_writes) {
		DEBUG(printf("Syncing udf_node `%s` because of %s\n", udf_node->dirent.d_name, why));

		udf_writequeue_kick(udf_node->udf_log_vol->primary->udf_session, "Sync node");
		printf("[Wait for pending writes  ");
		while (udf_node->pending_writes) {
			printf("\b\b\b\b\b\b\b\b%06d %c", udf_node->pending_writes, "|/-\\||//-\\"[cnt++]);
			cnt = cnt & 7;
			fflush(stdout);
			usleep(20);
		};
		printf("\b]\n");
	};

	return 0;
}


extern void udf_merge_allocentry_queue(struct udf_alloc_entries *queue, uint32_t lb_size);

/* VOP_FSYNC : metadata FFILESYNC */
/* writeout the udf_node back to disc */
int udf_writeout_udf_node(struct udf_node *udf_node, char *why) {
	struct udf_wrcallback   wr_callback;
	struct udf_allocentry  *dscr_entry, *next_dscr_entry, *alloc_entry;
	union dscrptr          *dscrptr;
	struct file_entry      *fe;
	struct extfile_entry   *efe;
	struct alloc_ext_entry *aee;
	struct lb_addr          parent_icb;
	struct icb_tag         *icbtag;
	struct long_ad         *l_ad;
	struct short_ad        *s_ad;
	uint32_t               *l_adptr;	/* points to length of alloc. descr. */
	uint8_t                *pos;
	char                   *what;
	uint32_t                alloc_entries;
	uint32_t                rest;		/* in bytes */
	uint32_t                len;
	uint32_t                logblks_rec, lb_size, descr_ver;

	lb_size     = udf_rw32(udf_node->udf_log_vol->log_vol->lb_size);

	/* assure all file data is written out! and clean up */
/* XXX is it OK to not sync here? XXX */
/*	udf_sync_udf_node(udf_node, why); */
	udf_merge_allocentry_queue(&udf_node->alloc_entries, lb_size);

	/* choose between using a normal fentry or an extfentry -> extfentry for now */
	dscrptr = calloc(1, UDF_SECTOR_SIZE);
	if (!dscrptr) return ENOMEM;

	/* calculate logical blocks recorded */
	logblks_rec = 0;
	TAILQ_FOREACH(alloc_entry, &udf_node->alloc_entries, next_alloc) {
		if (alloc_entry->flags == UDF_SPACE_ALLOCATED) {
			logblks_rec += (alloc_entry->len + lb_size-1) / lb_size;
		};
	};

	/* fill in (ext)fentry descriptor */
	/* copy descriptor version from the logvol's */

	bzero(&parent_icb, sizeof(struct lb_addr));
	descr_ver = udf_rw16(udf_node->udf_log_vol->log_vol->tag.descriptor_ver);
	if (descr_ver == 3) {
		efe = &dscrptr->efe;
		efe->tag.id             = udf_rw16(TAGID_EXTFENTRY);
		efe->tag.descriptor_ver = udf_rw16(descr_ver);
		udf_set_imp_id(&efe->imp_id);
		udf_set_timestamp_now(&efe->atime);			/* XXX atime, ctime, mtime, attrtime XXX */
		udf_set_timestamp_now(&efe->ctime);			/* XXX atime, ctime, mtime, attrtime XXX */
		udf_set_timestamp_now(&efe->mtime);			
		udf_set_timestamp_now(&efe->attrtime);
		efe->logblks_rec = udf_rw64(logblks_rec);

		/* set additional fileinfo (access, uid/gid) etc. */
		udf_node_set_fileinfo(udf_node, dscrptr);
		efe->obj_size           = efe->inf_len;			/* doesn't have to be */

		icbtag = &efe->icbtag;
	} else if (descr_ver == 2) {
		fe = &dscrptr->fe;
		fe->tag.id             = udf_rw16(TAGID_FENTRY);
		fe->tag.descriptor_ver = udf_rw16(descr_ver);
		udf_set_imp_id(&fe->imp_id);
		udf_set_timestamp_now(&fe->atime);			/* XXX atime, ctime, mtime, attrtime XXX */
		//udf_set_timestamp_now(&fe->ctime);			/* XXX atime, ctime, mtime, attrtime XXX */
		udf_set_timestamp_now(&fe->mtime);			
		udf_set_timestamp_now(&fe->attrtime);
		fe->logblks_rec = udf_rw64(logblks_rec);

		/* set additional fileinfo (access, uid/gid) etc. */
		udf_node_set_fileinfo(udf_node, dscrptr);
		//fe->obj_size           = fe->inf_len;			/* doesn't have to be */

		icbtag = &fe->icbtag;
	} else {
		printf("UDF: i dont know this descriptor version %d\n", descr_ver);
		return EBADF;
	};

	/* strategy 4 has no parent nodes */

	/* XXX NO extended attributes recorded YET (!!) XXX (like device nodes !!! ) */

	dscr_entry  = TAILQ_FIRST(&udf_node->dscr_allocs);

	/* ensure allocation of (ext) file descriptor */
	if (!dscr_entry) {
		/* have to allocate one and add to queue ! */
		printf("UDF: XXX no allocation of file descriptor entry yet in sync_udf_node\n");
		return ENOENT;
	};

	what = "UDF sync: File descriptor";
	alloc_entry = TAILQ_FIRST(&udf_node->alloc_entries);
	do {
		assert(dscr_entry->len <= UDF_SECTOR_SIZE);
		if (icbtag) {
			/* fill in ICB fields */			/* XXX we're only writing out strategy type 4 !!! XXX */
			icbtag->prev_num_dirs   = udf_rw32(0);		/* prev_alloc_entries = 0 due to strategy type 4 XXX */
			icbtag->strat_type      = udf_rw16(4);		/* default UDF strategy type 4 XXX what about 4096? */
			icbtag->parent_icb      = parent_icb;
			icbtag->file_type       = udf_node->udf_filetype;
			icbtag->flags           = udf_rw16(udf_node->udf_icbtag_flags);
			icbtag->max_num_entries = udf_rw16(1);		/* due to strategy type 4 ! XXX */
		};

		pos  = 0;
		rest = 0;
		l_adptr = &rest;	/* invalid */
		switch (udf_rw16(dscrptr->tag.id)) {
			case TAGID_FENTRY         :
				/* not used now */
				fe   = &dscrptr->fe;
				pos  = &fe->data[0] + udf_rw32(fe->l_ea);
				rest = dscr_entry->len - sizeof(struct file_entry) - udf_rw32(fe->l_ea);
				l_adptr = &fe->l_ad;
				break;
			case TAGID_EXTFENTRY      :
				efe  = &dscrptr->efe;
				pos  = &efe->data[0] + udf_rw32(efe->l_ea);
				rest = dscr_entry->len - sizeof(struct extfile_entry) - udf_rw32(efe->l_ea);
				l_adptr = &efe->l_ad;
				break;
			case TAGID_ALLOCEXTENT    :
				aee  = &dscrptr->aee;
				pos  = &aee->data[0];
				rest = dscr_entry->len - sizeof(struct alloc_ext_entry);
				l_adptr = &aee->l_ad;
				break;
			case TAGID_INDIRECT_ENTRY :
				/* do we even do these in strat 4 ? */
				printf("UDF: sanity check; request for writeout of indirect entry\n");
				free(dscrptr);
				return ENOENT;	/* panic really */
		};

		/* fill in remaining allocation entries 	*/
		/* remember to keep one SPARE for extending	*/

		alloc_entries = 0;
		DEBUG(printf("Allocated at "));

		if (udf_node->addr_type == UDF_ICB_INTERN_ALLOC) {
			/* internal allocation */
			DEBUG(printf("[internal in this ICB for %d bytes]   ", (uint32_t) udf_node->file_data_length));
			if (udf_node->file_data_length > rest) {
				printf("INTERNAL allocation overflow!!! help!!! this shouldn't happen!\n");
				/* XXX emergency repair!!! XXX */
				return ENOENT;	/* panic really */
			};
			memcpy(pos, udf_node->file_data, udf_node->file_data_length);

			/* alloc_entry is zero, so it'll fall trough */
			assert(alloc_entry == NULL);
		};

		while ((rest > 2*udf_node->icb_len) && alloc_entry) {
			assert(udf_node->icb_len);
			DEBUG(printf("[%p, lb_num = %d, len = %d]   ", alloc_entry, (uint32_t) alloc_entry->lb_num, (uint32_t) alloc_entry->len));

			/* XXX assumption: UDF_SPACE_* is equal to UDF flags XXX */
			len = alloc_entry->len | (((uint32_t) alloc_entry->flags) << 30);
			switch (udf_node->addr_type) {
				case UDF_ICB_SHORT_ALLOC :
					s_ad = (struct short_ad *) pos;
					s_ad->len          = udf_rw32(len);
					s_ad->lb_num       = udf_rw32(alloc_entry->lb_num);
					assert(alloc_entry->vpart_num == 0);
					if (alloc_entry->len   == 0)              s_ad->lb_num = udf_rw32(0);
					if (alloc_entry->flags == UDF_SPACE_FREE) s_ad->lb_num = udf_rw32(0);
					break;
				case UDF_ICB_LONG_ALLOC  :
					l_ad = (struct long_ad *) pos;
					l_ad->len          = udf_rw32(len);
					l_ad->loc.lb_num   = udf_rw32(alloc_entry->lb_num);
					l_ad->loc.part_num = udf_rw16(alloc_entry->vpart_num);
					l_ad->impl.im_used.unique_id = udf_rw64(udf_node->unique_id);
					if (alloc_entry->len   == 0)              l_ad->loc.lb_num = udf_rw32(0);
					if (alloc_entry->flags == UDF_SPACE_FREE) l_ad->loc.lb_num = udf_rw32(0);
					break;
			};
			alloc_entries++;
			alloc_entry = TAILQ_NEXT(alloc_entry, next_alloc);
			pos  += udf_node->icb_len;
			rest -= udf_node->icb_len;
		};
		DEBUG(printf("\nend alloc\n\n"));

		next_dscr_entry = TAILQ_NEXT(dscr_entry, next_alloc);
		if (alloc_entry) {
			/* overflow */

			/* ensure allocation of (ext) file descriptor */
			if (!next_dscr_entry) {
				/* have to allocate one and add to queue ! */
				printf("UDF: XXX no allocation of allocation extent descriptor yet in sync_udf_node\n");
				return ENOENT;
			};
			/* flag next extent being specified */
			len = next_dscr_entry->len | ((uint32_t) UDF_SPACE_REDIRECT << 30);
			switch (udf_node->addr_type) {
				case UDF_ICB_SHORT_ALLOC :
					s_ad = (struct short_ad *) pos;
					s_ad->len          = udf_rw32(len);
					s_ad->lb_num       = udf_rw32(next_dscr_entry->lb_num);
					assert(next_dscr_entry->vpart_num == 0);
					break;
				case UDF_ICB_LONG_ALLOC  :
					l_ad = (struct long_ad *) pos;
					l_ad->len          = udf_rw32(len);
					l_ad->loc.lb_num   = udf_rw32(next_dscr_entry->lb_num);
					l_ad->loc.part_num = udf_rw16(next_dscr_entry->vpart_num);
					l_ad->impl.im_used.unique_id = udf_rw64(udf_node->unique_id);
					break;
			};
			alloc_entries++;
			pos  += udf_node->icb_len;
			rest -= udf_node->icb_len;
		};
		/* manage lengths */
		*l_adptr = udf_rw32(alloc_entries * udf_node->icb_len);
		dscrptr->tag.desc_crc_len = udf_rw16((pos - (uint8_t *) dscrptr) - UDF_DESC_TAG_LENGTH);

		/* writeout */
		wr_callback.function  = udf_syncnode_callback;
		wr_callback.structure = (void *) udf_node;
		wr_callback.vpart_num = dscr_entry->vpart_num;
		wr_callback.lb_num    = dscr_entry->lb_num;
		wr_callback.offset    = 0; /* ? */
		wr_callback.length    = dscr_entry->len;

		UDF_VERBOSE_MAX(
			udf_validate_tag_sums(dscrptr);	/* for dumping */
			udf_dump_descriptor(dscrptr);
		);
		errno = udf_write_logvol_descriptor(udf_node->udf_log_vol, dscr_entry->vpart_num, dscr_entry->lb_num, what, dscrptr, dscr_entry->len, &wr_callback);

		/* advance */
		if (alloc_entry) {
			dscr_entry = next_dscr_entry;
			assert(dscr_entry);

			/* allocate and initialise new allocation extent descriptor	*/
			/* also flag no icbtag follows					*/
			bzero(dscrptr, UDF_SECTOR_SIZE);
			aee = &dscrptr->aee;
			aee->tag.id             = udf_rw16(TAGID_ALLOCEXTENT);
			aee->tag.descriptor_ver = udf_node->udf_log_vol->log_vol->tag.descriptor_ver;
			icbtag = NULL;

			what = "UDF sync: allocation extent";
		};

	} while (alloc_entry);

	/* XXX Free extra non used allocation extent descriptors XXX */

	/* Mark me clean */
	if (udf_node->mountpoint) {
		/* remove me just in case; why were we called otherwise? */
		udf_node_unmark_dirty(udf_node);
	};

	return 0;
}


/******************************************************************************************
 *
 * Space bitmap reader and writer 
 *
 ******************************************************************************************/

/* tested OK */
int udf_read_in_space_bitmap(struct udf_alloc_entries *queue, struct space_bitmap_desc *sbd, uint32_t lb_size, uint64_t *freespace) {
	struct udf_allocentry *alloc_entry;
	uint64_t bits, from, now, start, end;
	uint8_t byte, bit, bitpos, state, *pos;
	int cnt;

	assert(udf_rw16(sbd->tag.id) == TAGID_SPACE_BITMAP);

	DEBUG(printf("processing space bitmap : \n"););
	bits = udf_rw32(sbd->num_bits);

	/*
	 * Mark the disc as completely full;
	 * Bugallert: use `udf_mark_allocentry_queue()' for the extent might
	 * not fit in just one alloc entry for bigger discs
	 */
	assert(TAILQ_EMPTY(queue));
	udf_mark_allocentry_queue(queue, lb_size, 0, bits * lb_size, UDF_SPACE_ALLOCATED, NULL, NULL);

	pos = sbd->data;
	from = 0; now = 0; bitpos = 0; byte = *pos; state = byte & 1;
	*freespace = 0;
	while (now < bits) {
		if (bitpos == 0) {
			byte = *pos++;
		};
		bit = byte & 1;
		if (bit != state) {
			if (state) {
				start = from;
				end   = now-1;
				//printf("[%08d - %08d]", start, end);
				udf_mark_allocentry_queue(queue, lb_size, start*lb_size, (end-start+1)*lb_size, UDF_SPACE_FREE, NULL, NULL);
				*freespace += (end-start+1)*lb_size;
			};
			from = now;
			state = bit;
		};
		byte >>= 1;
		bitpos = (bitpos+1) & 7;
		now++;
	};
	if (state) {
		start = from;
		end   = now;
		//printf("[%08d - %08d]", start, end);
		udf_mark_allocentry_queue(queue, lb_size, start*lb_size, (end-start)*lb_size, UDF_SPACE_FREE, NULL, NULL);
		*freespace += (end-start)*lb_size;
	};

	UDF_VERBOSE_TABLES(
		printf("\t\tFree space found on this partition");
		cnt = 0;
		start = 0;
		TAILQ_FOREACH(alloc_entry, queue, next_alloc) {
			if (alloc_entry->flags == 0) {
				//printf("... ");
			} else {
				if (cnt == 0) printf("\n\t\t\t");
				printf("[%08"PRIu64" - %08"PRIu64"]   ", start / lb_size, ((start + alloc_entry->len) / lb_size)-1);
				cnt++; if (cnt > 4) cnt = 0;
			};
			start += alloc_entry->len;
		};
		printf("\n");
	);

	/* merge is not nessisary */
	return 0;
}


/* inverse of readin space bitmap; it synchronises the bitmap with the queue */
/* tested OK */
int udf_sync_space_bitmap(struct udf_alloc_entries *queue, struct space_bitmap_desc *sbd, uint32_t lb_size) {
	struct udf_allocentry *alloc_entry;
	uint32_t start, bits, total_bits;
	uint32_t cnt, byte;
	uint8_t  bit, bitmask, setting;
	uint8_t *pos;

	/* merge it just in case */
	udf_merge_allocentry_queue(queue, lb_size);

	total_bits = udf_rw32(sbd->num_bits);
	DEBUG(printf("SYNC SPACE BITMAP DEBUG: total bits = %d\n", total_bits));

	alloc_entry = TAILQ_FIRST(queue);
	start = alloc_entry->lb_num;
	assert(start == 0);

	TAILQ_FOREACH(alloc_entry, queue, next_alloc) {
		DEBUG(printf(" [%d : %d + %d]", alloc_entry->flags, alloc_entry->lb_num, alloc_entry->len));
		bits  = alloc_entry->len / lb_size;
		assert(bits*lb_size == alloc_entry->len);

		byte = start / 8;
		bit  = start - byte*8;

		pos = sbd->data + byte;

		if (byte*8 + bit + bits > total_bits) {
			printf("UDF: not enough space writing back space bitmap! HELP!\n");
			return EBADF;
		};

		cnt = 0;
		setting = (alloc_entry->flags != UDF_SPACE_FREE) ? 0 : 255;
		while (cnt < bits) {
			bitmask = (1 << bit);

			/* simple sanity check */
			if (byte*8 + bit >= total_bits) {
				printf("IEEEE!!!! too big; %d instead of %d\n", (byte*8 + bit), total_bits);
			};
			*pos = (*pos & (~bitmask)) | (setting ? bitmask: 0);

			cnt++;
			bit++;
			if (bit == 8) {
				/* byte transition */
				byte++; bit = 0;
				pos++;
#if 0
				/* speedup by doing bytes at a time */
				while (bits-cnt > 8) {
					*pos = setting;
					cnt += 8;
					pos++; byte++;
				};
#endif
			};
		};
		start += bits;
	};
	DEBUG(printf("\n\n"));
	return 0;
}



/******************************************************************************************
 *
 * Filepart readers and writers 
 *
 ******************************************************************************************/


/* VOP_READ */
int udf_read_file_part_uio(struct udf_node *udf_node, char *what, int cache_flags, struct uio *data_uio) {
	struct udf_allocentry	*alloc_entry;
	uint64_t         cur_offset;
	uint32_t         offset, data_length, overlap_length;
	uint32_t         read_offset;
	uint32_t         len, lb_num, vpart_num, piece_length;
	int32_t	         error;
	uint8_t	        *base;
	uint8_t	         flags;

	if (!udf_node) return EINVAL;

	if (udf_node->addr_type == UDF_ICB_INTERN_ALLOC) {
		/* internal */
		offset = data_uio->uio_offset;
		base   = udf_node->file_data;
		data_length = udf_node->file_data_length - offset;	/* max length of transfer data */
		data_length = MIN(data_length, data_uio->uio_resid);	/* never more than requested   */
		uiomove(base + offset, data_length, data_uio);
		return 0;
	};

	/* process all allocation descriptors */
	cur_offset = 0;
	TAILQ_FOREACH(alloc_entry, &udf_node->alloc_entries, next_alloc) {
		/* still data requested? */
		if (data_uio->uio_resid == 0) return 0;

		len       = alloc_entry->len;
		lb_num    = alloc_entry->lb_num;
		vpart_num = alloc_entry->vpart_num;
		flags     = alloc_entry->flags;

		/* `normal` space */
		if (cur_offset + len >= data_uio->uio_offset) {
			/* found an extent that fits */
			overlap_length = MIN(data_uio->uio_resid, cur_offset + len - data_uio->uio_offset);	/* shouldn't matter */
			read_offset    = MAX(0, data_uio->uio_offset - cur_offset);				/* ought to be zero */

			if (flags != UDF_SPACE_FREE) {
				/* written to disc or in the write queue/cache to be written out */
				error = udf_read_logvol_bytes_uio(udf_node->udf_log_vol, vpart_num, lb_num, read_offset, overlap_length, what, data_uio, cache_flags);
				if (error) {
					fprintf(stderr, "\tgot error from read_logvol_bytes_uio : %s\n", strerror(error));
					return error;
				};
			} else {
				/* non allocated and not written out */
				bzero(zero_sector, UDF_SECTOR_SIZE);	/* XXX all the time?? XXX */
				while (overlap_length) {
					piece_length = MIN(overlap_length, UDF_SECTOR_SIZE);
					uiomove(zero_sector, piece_length, data_uio);
					overlap_length -= piece_length;
				};
			};

			/* are we finished now? */
			if (data_uio->uio_resid == 0) return 0;
		}; /* overlap */

		cur_offset += len;
	}; /* FOREACH */

	if (data_uio->uio_resid) {
		printf("UDF: WARNING file is truncated; missing %d bytes while reading %s\n", (int) data_uio->uio_resid, what);
		return EIO;
	};

	return 0;
}


void udf_filepart_write_callback(int reason, struct udf_wrcallback *wrcallback, int error, uint8_t *sectordata) {
	struct udf_node *udf_node = (struct udf_node *) wrcallback->structure;

	if (reason == UDF_WRCALLBACK_REASON_PENDING) {
		/* XXX spinlock */
		udf_node->pending_writes++;
		/* XXX end spinlock */
		return;
	};

	if (reason == UDF_WRCALLBACK_REASON_ANULATE) {
		/* XXX spinlock */
		udf_node->pending_writes--;
		/* XXX end spinlock */
		DEBUG(printf("FILEPART ANULATE! lb_num = %d, offset = %d, length = %d\n", (int) wrcallback->lb_num, (int) wrcallback->offset, (int) wrcallback->length));
		return;
	};

	assert(reason == UDF_WRCALLBACK_REASON_WRITTEN);
	if (error) {
		DEBUG(
			printf("filepart write callback called with sector data %p\n", sectordata);
			printf("\toffset    %08d\n", (int) wrcallback->offset);
			printf("\tlb_num    %08d\n", (int) wrcallback->lb_num);
			printf("\tlength    %08d\n", (int) wrcallback->length);
			printf("\terror     %d\n",   (int) error);
		);
		printf("UDF error: write errors not yet fixed!\n");
		return;
		/* ## we have to relocate this sector ##			*/
		/* depending on partition mapping type :( 			*/
	} else {
		/*
		 * DONT mark the extent as `allocated' anymore since its
		 * allready allocated. No need to mess with alloc_entries
		 * type/markings */

		/* udf_filepart_mark_extent(udf_node, wrcallback->offset, wrcallback->length, UDF_SPACE_ALLOCATED); */

		/* XXX spinlock */
		udf_node->pending_writes--;
		/* XXX end spinlock */
	};
}


/* VOP_TRUNCATE */
/* doesn't lock udf_node */
int udf_truncate_udf_node(struct udf_node *udf_node, uint64_t length /* ,ioflags */) {
	struct udf_wrcallback  wr_callback;
	struct udf_log_vol    *udf_log_vol;
	struct udf_allocentry *alloc_entry, *last_alloc_entry, *next_alloc_entry;
	struct uio	 conv_uio, *uio;
	struct iovec	 conv_iovec;
	uint64_t	 overlap_length, lb_size;
	uint64_t	 overlap_offset;
	uint8_t		 flags;
	uint64_t	 len, lb_num, vpart_num;
	uint64_t	 new_possible_extent, cur_offset;
	int32_t		 error;
	char 		*what;

	if (!udf_node) return EINVAL;

	/* lock node directly to avoid multiple writers entering */
	UDF_MUTEX_LOCK(&udf_node->alloc_mutex);

	udf_log_vol = udf_node->udf_log_vol;

	/* we might change this node AND access times ... */
	if (!udf_node->dirty) {
		/* mark node as dirty */
		udf_node_mark_dirty(udf_node);
	};

	/* Convert internal node when nessisary */
	bzero(&conv_uio, sizeof(struct uio));
	if (udf_node->addr_type == UDF_ICB_INTERN_ALLOC) {
		if (length < udf_node->file_data_alloc_length) {
			if (length > udf_node->file_data_length) {
				bzero(udf_node->file_data + udf_node->file_data_length, length - udf_node->file_data_length);
			};
			udf_node->file_data_length = length;
			udf_node->stat.st_size     = length;

			UDF_MUTEX_UNLOCK(&udf_node->alloc_mutex);
			return 0;
		};

		/* Won't fit, convert to alloc entries */
		printf("\nUDF: Warning; yet untested morph from internal allocation entry to long allocation entry\n");

		/* create an extra uio to transfer first */
		conv_uio.uio_rw = UIO_READ;			/* read from uio data structure */
		conv_iovec.iov_base = udf_node->file_data;
		conv_iovec.iov_len  = UDF_SECTOR_SIZE;
		conv_uio.uio_offset = 0;
		conv_uio.uio_resid  = MAX(UDF_SECTOR_SIZE, udf_node->file_data_alloc_length);
		conv_uio.uio_iovcnt = 1;
		conv_uio.uio_iov    = &conv_iovec;

		/* treat as an empty node */
		udf_node->addr_type    = UDF_ICB_LONG_ALLOC;
		udf_node->stat.st_size = 0;
	};

	/* File extention */
	lb_size    = udf_rw32(udf_log_vol->log_vol->lb_size);
	udf_merge_allocentry_queue(&udf_node->alloc_entries, lb_size);

	/* extend the file when nessisary */
	new_possible_extent = length;
	if (new_possible_extent > udf_node->stat.st_size) {
		udf_cut_allocentry_queue(&udf_node->alloc_entries, lb_size, new_possible_extent);
		udf_node->stat.st_size = new_possible_extent;
	};

	/* allow access to allocations */
	UDF_MUTEX_UNLOCK(&udf_node->alloc_mutex);

	if (conv_uio.uio_resid) {
		/* writeout the preserved old internal alloc data; Its size is at most one sector */
		error = udf_filepart_mark_extent(udf_node, 0, UDF_SECTOR_SIZE, UDF_SPACE_ALLOCATED);
		if (error) {
			UDF_MUTEX_UNLOCK(&udf_node->alloc_mutex);
			return error;
		};

		alloc_entry = TAILQ_FIRST(&udf_node->alloc_entries);

		DEBUG(printf("EXPIRIMENTAL write old internal content (%d, %d), flag %d\n", (int) alloc_entry->lb_num, alloc_entry->len, alloc_entry->flags));
		printf("EXPIRIMENTAL write old internal content (%d, %d), flag %d\n", (int) alloc_entry->lb_num, alloc_entry->len, alloc_entry->flags);
		/* read alloc_entry information */
		len       = alloc_entry->len;
		lb_num    = alloc_entry->lb_num;
		vpart_num = alloc_entry->vpart_num;
		flags     = alloc_entry->flags;

		uio = &conv_uio;

		/* calculate overlap size */
		overlap_length = conv_uio.uio_resid;
		overlap_offset = 0;

		wr_callback.function  = udf_filepart_write_callback;
		wr_callback.structure = (void *) udf_node;
		wr_callback.vpart_num = vpart_num;
		wr_callback.lb_num    = lb_num;
		wr_callback.offset    = 0;
		wr_callback.length    = overlap_length;

		flags = UDF_CACHEFLAG_NOREADIN;
		what  = udf_node->dirent.d_name;
		error = udf_write_logvol_bytes_uio(udf_log_vol, vpart_num, lb_num, overlap_offset, overlap_length, what, uio, flags, &wr_callback);
		if (error) {
			/* IEEE unlocking? */
			UDF_MUTEX_UNLOCK(&udf_node->alloc_mutex);
			fprintf(stderr, "\tgot error from write_logvol_bytes_uio : %s\n", strerror(error));
			return error;
		};
	};

	/* check if we need to trim alloc entries */
	if (new_possible_extent < udf_node->stat.st_size) {
		/* sync node; otherwise late callbacks give trouble when shrinking */
		//printf("TRIMMING %s from %d to %d\n", udf_node->dirent.d_name, (int32_t) udf_node->stat.st_size, (uint32_t) length);
		//sleep(2);
		udf_sync_udf_node(udf_node, "truncating file");

		/* keep others out */
		UDF_MUTEX_LOCK(&udf_node->alloc_mutex);
		error = udf_splitup_allocentry_queue(&udf_node->alloc_entries, lb_size, /*offset*/ length, /*size*/ lb_size, NULL, NULL);
		if (!error) {
			alloc_entry = TAILQ_FIRST(&udf_node->alloc_entries);
			cur_offset  = 0;
			while (alloc_entry && cur_offset + alloc_entry->len < length) {
				cur_offset += alloc_entry->len;
				alloc_entry = TAILQ_NEXT(alloc_entry, next_alloc);
			};
			last_alloc_entry = alloc_entry;
			if (last_alloc_entry) {
				/* delete trailing allocation entries */
				alloc_entry = TAILQ_NEXT(last_alloc_entry, next_alloc);
				while (alloc_entry) {
					next_alloc_entry = TAILQ_NEXT(alloc_entry, next_alloc);
					TAILQ_REMOVE(&udf_node->alloc_entries, alloc_entry, next_alloc);
					if (alloc_entry->flags != UDF_SPACE_FREE) {
						/* bring back free space */
						udf_partmap_free_lbs(udf_log_vol, alloc_entry->vpart_num, alloc_entry->lb_num, alloc_entry->len);
					};
					free(alloc_entry);
					alloc_entry = next_alloc_entry;
				};
			};
			alloc_entry = last_alloc_entry;

			/* trim last alloc_entry part; it can be two blocks */
			assert(length - cur_offset <= lb_size);
			if (alloc_entry->len > lb_size) {
				udf_partmap_free_lbs(udf_log_vol, alloc_entry->vpart_num, alloc_entry->lb_num + 1, alloc_entry->len - lb_size);
			};
			alloc_entry->len = length - cur_offset;

			if (alloc_entry->len == 0 || length == 0) {
				/* also remove this last alloc_entry */
				udf_partmap_free_lbs(udf_log_vol, alloc_entry->vpart_num, alloc_entry->lb_num, lb_size);
				TAILQ_REMOVE(&udf_node->alloc_entries, alloc_entry, next_alloc);
			};
			udf_node->stat.st_size = length;
		};
		UDF_MUTEX_UNLOCK(&udf_node->alloc_mutex);
	};

	/* XXX check if we can convert an alloc entry back to internal? */

	return 0;
}


/* VOP_WRITE */
int udf_write_file_part_uio(struct udf_node *udf_node, char *what, int cache_flags, struct uio *data_uio) {
	struct udf_wrcallback  wr_callback;
	struct udf_allocentry *alloc_entry;
	uint8_t		*base;
	uint32_t	 offset, copy_length, overlap_length, lb_size;
	uint64_t	 cur_offset;
	uint32_t	 overlap_offset;
	uint8_t		 flags;
	uint32_t	 len, lb_num, vpart_num;
	uint64_t	 new_possible_extent;
	int		 error, last;

	if (!udf_node) return EINVAL;

	/* Zero length write -> finished */
	if (data_uio->uio_resid == 0) return 0;

	/* lock node directly to avoid multiple writers entering */
//	pthread_rwlock_wrlock(&udf_node->udf_node_lock);

	/* we might change this node AND access times ... */
	if (!udf_node->dirty) {
		udf_node_mark_dirty(udf_node);
	};

	new_possible_extent = data_uio->uio_offset + data_uio->uio_resid;
	if (new_possible_extent > udf_node->stat.st_size) {
		udf_truncate_udf_node(udf_node, new_possible_extent);
	};

	/* Internal nodes */
	if (udf_node->addr_type == UDF_ICB_INTERN_ALLOC) {
		/* still enough storage */
		offset              = data_uio->uio_offset;
		base                = udf_node->file_data;
		copy_length         = udf_node->file_data_length - offset;	/* max length of transfer data */
		copy_length         = MIN(copy_length, data_uio->uio_resid);	/* never more than requested   */
		uiomove(base + offset, copy_length, data_uio);

		/* finished */
//		pthread_rwlock_unlock(&udf_node->udf_node_lock);
		return 0;
	};

	UDF_MUTEX_LOCK(&udf_node->alloc_mutex);

	/* File extention */
	lb_size    = udf_rw32(udf_node->udf_log_vol->log_vol->lb_size);
	udf_merge_allocentry_queue(&udf_node->alloc_entries, lb_size);

	/* mark the new places for allocation */
	error = udf_filepart_mark_extent(udf_node, data_uio->uio_offset, data_uio->uio_resid, UDF_SPACE_FREED);
	if (error) {
		UDF_MUTEX_UNLOCK(&udf_node->alloc_mutex);
		return error;
	};

	/* allow access to allocations */
	UDF_MUTEX_UNLOCK(&udf_node->alloc_mutex);

	/* process all allocation descriptors */
	cur_offset = 0;
	TAILQ_FOREACH(alloc_entry, &udf_node->alloc_entries, next_alloc) {
		/* still data requested? */
		if (data_uio->uio_resid == 0) break;
		DEBUG(printf("writefilepart extent (%d, %d), flag %d\n", (int) alloc_entry->lb_num, alloc_entry->len, alloc_entry->flags));

		/* overlapping? */
		if (cur_offset + alloc_entry->len > data_uio->uio_offset) {
			/* read alloc_entry information */
			len       = alloc_entry->len;
			lb_num    = alloc_entry->lb_num;
			vpart_num = alloc_entry->vpart_num;
			flags     = alloc_entry->flags;

			/* calculate overlap size */
			overlap_length = MIN(data_uio->uio_resid, cur_offset + len - data_uio->uio_offset);
			overlap_offset = MAX(0, data_uio->uio_offset - cur_offset);

			wr_callback.function  = udf_filepart_write_callback;
			wr_callback.structure = (void *) udf_node;
			wr_callback.vpart_num = vpart_num;
			wr_callback.lb_num    = lb_num;
			wr_callback.offset    = cur_offset + overlap_offset;
			wr_callback.length    = overlap_length;

			last = (cur_offset + overlap_offset + overlap_length == udf_node->stat.st_size);
			if (last) cache_flags |= UDF_CACHEFLAG_LASTSECTOR;

DEBUG(
	printf("write_logvol_bytes_uio called with lb_num %d, overlap_offset %d (%d lb_num), overlap_length %d (%d lb_num)\n",
		(int) lb_num, (int) overlap_offset, (int) overlap_offset/lb_size, (int) overlap_length, (int) overlap_length/lb_size);
);
			error = udf_write_logvol_bytes_uio(udf_node->udf_log_vol, vpart_num, lb_num, overlap_offset, overlap_length, what, data_uio, cache_flags, &wr_callback);
			if (error) {
//				pthread_rwlock_unlock(&udf_node->udf_node_lock);
				fprintf(stderr, "\tgot error from write_logvol_bytes_uio : %s\n", strerror(error));
				return error;
			};
			/* ignore returncode of mark extent? */
			udf_filepart_mark_extent(udf_node, cur_offset + overlap_offset, overlap_length, UDF_SPACE_ALLOCATED);
		}; /* overlap */

		cur_offset += alloc_entry->len;
	}; /* while */
//	pthread_rwlock_unlock(&udf_node->udf_node_lock);

	return 0;
}


/******************************************************************************************
 *
 * UDF volume and descriptor logic
 *
 ******************************************************************************************/


struct udf_volumeset *udf_search_volumeset(char *volset_id) {
	struct udf_volumeset *volumeset;
	struct udf_pri_vol   *primary;

	/* XXX this is a bit ugly XXX */
	SLIST_FOREACH(volumeset, &udf_volumeset_list, next_volumeset) {
		primary = STAILQ_FIRST(&volumeset->primaries);
		assert(primary->pri_vol);
		if (memcmp(primary->pri_vol->volset_id, volset_id, 128) == 0) return volumeset;
	};
	return NULL;
}


struct udf_pri_vol *udf_search_primary(struct udf_volumeset *set, char *id) {
	struct udf_pri_vol *primary;

	STAILQ_FOREACH(primary, &set->primaries, next_primary) {
		assert(primary->pri_vol);
		if (memcmp(primary->pri_vol->vol_id, id, 32) == 0) return primary;
	};
	return NULL;
}


struct udf_log_vol *udf_search_logical_volume_in_primary(struct udf_pri_vol *primary, char *logvol_id) {
	struct udf_log_vol *here;

	SLIST_FOREACH(here, &primary->log_vols, next_logvol) {
		if (memcmp(here->log_vol->logvol_id, logvol_id, 128) == 0) return here;
	};
	return NULL;
}



/* called not very often ... ; free incomming when not needed */
int udf_proc_pri_vol(struct udf_session *udf_session, struct udf_pri_vol **current, struct pri_vol_desc *incomming) {
	struct udf_volumeset *volset;
	struct udf_pri_vol   *primary;

	volset = udf_search_volumeset(incomming->volset_id);
	if (!volset) {
		/* create a volume set */
		volset = calloc(1, sizeof(struct udf_volumeset));
		if (!volset) {
			free(incomming);
			return ENOMEM;
		};

		/* populate and link in */
		volset->max_partnum = 0;
		STAILQ_INIT(&volset->primaries);
		SLIST_INSERT_HEAD(&udf_volumeset_list, volset, next_volumeset);
	};
	assert(volset);

	primary = udf_search_primary(volset, incomming->vol_id);
	*current = primary;
	if (!primary) {
		/* create a primary volume */
		primary = calloc(1, sizeof(struct udf_pri_vol));
		if (!primary) {
			free(incomming);
			return ENOMEM;
		};

		/* add to volset's primaries list */
		STAILQ_INSERT_TAIL(&volset->primaries, primary, next_primary);

		*current = primary;
	} else {
		/* mark as current */

		/* ok ... we now need to check if this new descriptor is a newer version */
		if (udf_rw32(incomming->seq_num) <= udf_rw32(primary->pri_vol->seq_num)) {
			if (udf_session->session_num <= (*current)->udf_session->session_num) {
				DEBUG(printf("UDF: DISCARDING primary descriptor for its the same but higher session number\n"));
				/* its an older one; ignore */
				free(incomming);
				return 0;
			}
		};
		DEBUG(printf("UPDATING primary descriptor for it has a higher session number\n"));
	};

	/* update the primary volume descriptor */
	if (primary->pri_vol) free(primary->pri_vol);
	primary->volumeset   = volset;
	primary->pri_vol     = incomming;
	primary->udf_session = udf_session;

	return 0;
}


/* not called often ; free lvid when not used */
int udf_proc_logvol_integrity(struct udf_log_vol *udf_log_vol, struct logvol_int_desc *new_lvid) {
	struct udf_logvol_info *impl;
	int error, tagid;

	error = udf_check_tag((union dscrptr *) new_lvid);
	if (error) {
		free(new_lvid);
		return error;	/* return error on faulty tag */
	};

	tagid = udf_rw16(new_lvid->tag.id);
	/* getting a terminator tag or zero is an OK condition */
	if ((tagid == TAGID_TERM) || (tagid == 0)) {
		free(new_lvid);
		return 0;
	};

	/* not getting an logical volume itegrity volume descriptor is an error now */
	if (tagid != TAGID_LOGVOL_INTEGRITY) {
		printf("IEE! got a %d tag while searching for a logical volume integrity descriptor\n", tagid);
		free(new_lvid);
		return EBADF;	/* XXX error code? XXX */
	};

	/* check CRC on the contents of the logvol integrity */
	error = udf_check_tag_payload((union dscrptr *) new_lvid);
	if (error) {
		free(new_lvid);
		return error;
	};

	/* process information contained in logical volume integrity descriptor */
	udf_log_vol->logvol_state     = udf_rw32(new_lvid->integrity_type);
	udf_log_vol->integrity_serial = udf_rw16(new_lvid->tag.serial_num);
	impl = (struct udf_logvol_info *) (new_lvid->tables + 2*udf_rw16(new_lvid->num_part));

	udf_log_vol->min_udf_readver  = udf_rw16(impl->min_udf_readver);
	udf_log_vol->min_udf_writever = udf_rw16(impl->min_udf_writever);
	udf_log_vol->max_udf_writever = udf_rw16(impl->max_udf_writever);

	udf_log_vol->num_files        = udf_rw32(impl->num_files);
	udf_log_vol->num_directories  = udf_rw32(impl->num_directories);
	udf_log_vol->next_unique_id   = udf_rw64(new_lvid->logvol_content_use.logvolhdr.next_unique_id);

	free(new_lvid);

	return 0;
}


void udf_derive_new_logvol_integrity(struct udf_log_vol *udf_log_vol) {
	udf_log_vol->logvol_state     = UDF_INTEGRITY_OPEN;
	udf_log_vol->integrity_serial = 1;

	/* analyse the log vol to check out minimum and maximum read/write versions */
	if (udf_rw16(udf_log_vol->log_vol->tag.descriptor_ver) == 2) {
		udf_log_vol->min_udf_readver  = 0x0102;
		udf_log_vol->min_udf_writever = 0x0150;
		udf_log_vol->max_udf_writever = 0x0150;
	} else {
		udf_log_vol->min_udf_readver  = 0x0201;
		udf_log_vol->min_udf_writever = 0x0201;
		udf_log_vol->max_udf_writever = 0x0201;		/* 2.50? */
	};
	udf_log_vol->num_files       = 0;
	udf_log_vol->num_directories = 0;
	udf_log_vol->next_unique_id  = 0;	/* zero first, rest 15/16+ minimum */
}


int udf_proc_logvol_integrity_sequence(struct udf_log_vol *udf_log_vol) {
	union dscrptr	*dscr;
	uint32_t	 sector, length, lvid_len, num_sectors;
	int		 error;

	sector = udf_rw32(udf_log_vol->log_vol->integrity_seq_id.loc);
	length = udf_rw32(udf_log_vol->log_vol->integrity_seq_id.len);

	if (!length) {
		fprintf(stderr, "UDF: no volume integrity descriptor sequence space defined... badly formatted media\n");
		return EBADF;
	};

	error = 0;
	while (length) {
		error = udf_read_session_descriptor(udf_log_vol->primary->udf_session, sector, "Logical volume integrity descriptor (LVID)", &dscr, &lvid_len);
		if (error) {
			if (dscr) free(dscr);
			break;
		};

		UDF_VERBOSE_MAX(udf_dump_descriptor(dscr));

		error = udf_proc_logvol_integrity(udf_log_vol, &dscr->lvid);
		if (error) break;
		if (udf_rw16(dscr->tag.id) == TAGID_TERM) break;

		num_sectors = (lvid_len + UDF_SECTOR_SIZE-1) / UDF_SECTOR_SIZE;
		length -= num_sectors * UDF_SECTOR_SIZE;
		sector += num_sectors;

		if (udf_rw32(dscr->lvid.next_extent.len)) {
			sector = udf_rw32(dscr->lvid.next_extent.loc);
			length = udf_rw32(dscr->lvid.next_extent.len);
		};
	};

	if (error) {
		fprintf(stderr, "UDF: Warning! no logical volume integrity descriptor found\n");
		printf("Creating logical volume integrity descriptor for no logical volume integrity is found\n");
		udf_derive_new_logvol_integrity(udf_log_vol);
	};

	return 0;
}


/* Add partition mapping to specified logvol descriptor */
void udf_add_physical_to_logvol(struct logvol_desc *logvol, uint16_t vol_seq_num, uint16_t phys_part_num) {
	union  udf_pmap *pmap;
	uint8_t         *pmap_pos;

	pmap_pos = logvol->maps + udf_rw32(logvol->mt_l);

	pmap = (union udf_pmap *) pmap_pos;
	pmap->pm1.type        = 1;
	pmap->pm1.len         = sizeof(struct part_map_1);
	pmap->pm1.vol_seq_num = udf_rw16(vol_seq_num);
	pmap->pm1.part_num    = udf_rw16(phys_part_num);

	/* increment partition mapping count */
	logvol->n_pm = udf_rw32(udf_rw32(logvol->n_pm) + 1);
	logvol->mt_l = udf_rw32(udf_rw32(logvol->mt_l) + sizeof(struct part_map_1));

	logvol->tag.desc_crc_len = udf_rw16(udf_rw16(logvol->tag.desc_crc_len) + sizeof(struct part_map_1));
}


#if 0
/* not yet */
void udf_add_sparable_to_logvol(struct logvol_desc *logvol, uint16_t vol_seq_num, uint16_t phys_part_num, uint16_t packet_len, ) {
	union  udf_pmap *pmap;
	uint8_t         *pmap_pos;

	pmap_pos = logvol->maps + udf_rw32(logvol->mt_l);

	pmap = (union udf_pmap *) pmap_pos;
	pmap->pm1.type        = 1;
	pmap->pm1.len         = sizeof(struct part_map_1);
	pmap->pm1.vol_seq_num = vol_seq_num;
	pmap->pm1.part_num    = phys_part_num;

	/* increment partition mapping count */
	logvol->n_pm = udf_rw32(udf_rw32(logvol->n_pm) + 1);
	logvol->mt_l = udf_rw32(udf_rw32(logvol->mt_l) + sizeof(struct part_map_1));

	logvol->tag.desc_crc_len = udf_rw16(udf_rw16(logvol->tag.desc_crc_len) + sizeof(struct part_map_1));
}
#endif


/* not called often; free incomming when not needed */
int udf_proc_log_vol(struct udf_pri_vol *primary, struct udf_log_vol **current, struct logvol_desc *incomming) {
	struct udf_log_vol	 *logical;
	struct udf_part_mapping	 *part_mapping, *next_part_mapping, *data_part_mapping;
	union  udf_pmap		 *pmap;
	uint32_t		  part_cnt, pmap_type, pmap_size;
	uint32_t		  data_part_num;
	uint8_t 		 *pmap_pos;

	logical = udf_search_logical_volume_in_primary(primary, incomming->logvol_id);
	if (!logical) {
		/* create a logical volume */
		logical = calloc(1, sizeof(struct udf_log_vol));
		if (!logical) {
			free(incomming);
			return ENOMEM;
		};

		/* link in */
		SLIST_INSERT_HEAD(&primary->log_vols, logical, next_logvol);
	} else {
		/* ok ... we now need to check if this new descriptor is a newer version */
		if (udf_rw32(incomming->seq_num) < udf_rw32(logical->log_vol->seq_num)) {
			/* its an older one; ignore */
			free(incomming);
			return 0;
		};
	};

	/* update the logical volume descriptor and its mappings; first delete old partition mappings allocated before */
	logical->primary  = primary;
	if (current) *current = logical;

	part_mapping = SLIST_FIRST(&logical->part_mappings);
	while (part_mapping) {
		next_part_mapping = SLIST_NEXT(part_mapping, next_mapping);
		SLIST_REMOVE_HEAD(&logical->part_mappings, next_mapping);
		/* TODO cleanup old cruft */
		/*
			free(part_mapping->sparing_table);
			free(part_mapping->vat_file_entry);
			free(part_mapping->vat);
			free(part_mapping->meta_file);
			free(part_mapping->meta_mirror_file);
			free(part_mapping->meta_bitmap_file);
		*/
		free(part_mapping);
		part_mapping = next_part_mapping;
	};
	SLIST_INIT(&logical->part_mappings);

	/* use the new logical volume and preprocess it */
	if (logical->log_vol) free(logical->log_vol);
	logical->log_vol = incomming;

	/* process the logical volume integrity sequence */
	udf_proc_logvol_integrity_sequence(logical);

	/* build up the partion mappings */
	logical->num_part_mappings = udf_rw32(incomming->n_pm);

	/* process partition mappings */
	pmap_pos = &logical->log_vol->maps[0];
	for(part_cnt = 0; part_cnt < logical->num_part_mappings; part_cnt++) {
		/* get a new part_mapping structure */
		part_mapping = calloc(1, sizeof(struct udf_part_mapping));
		assert(part_mapping);		/* XXX check with partition mapping destructor etc XXX */

		/* add to list */
		SLIST_INSERT_HEAD(&logical->part_mappings, part_mapping, next_mapping);

		/* process */
		pmap = (union udf_pmap *) pmap_pos;
		pmap_type = pmap->data[0];
		pmap_size = pmap->data[1];

		part_mapping->udf_virt_part_num = part_cnt;
		part_mapping->udf_pmap = pmap;
		switch (pmap_type) {
			case 1:
				part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_PHYSICAL;
				part_mapping->vol_seq_num = udf_rw16(pmap->pm1.vol_seq_num);
				part_mapping->udf_phys_part_num = udf_rw16(pmap->pm1.part_num);
				break;
			case 2:
				part_mapping->vol_seq_num = udf_rw16(pmap->pm2.vol_seq_num);
				part_mapping->udf_phys_part_num = udf_rw16(pmap->pm2.part_num);
				if (strncmp(pmap->pm2.part_id.id, "*UDF Virtual Partition", UDF_REGID_ID_SIZE) == 0) {
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_VIRTUAL;
					break;
				};
				if (strncmp(pmap->pm2.part_id.id, "*UDF Sparable Partition", UDF_REGID_ID_SIZE) == 0) {
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_SPARABLE;
					break;
				};
				if (strncmp(pmap->pm2.part_id.id, "*UDF Metadata Partition", UDF_REGID_ID_SIZE) == 0) {
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_META;
					fprintf(stderr, "WARNING: untested `metadata' partition type encountered\n");
					break;
				};
				printf("HELP ... found unsupported type 2 partition mapping id `%s`; marking broken\n", pmap->pm2.part_id.id);
			default:
				part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
		};
		switch (part_mapping->udf_part_mapping_type) {
			case UDF_PART_MAPPING_PHYSICAL :
			case UDF_PART_MAPPING_ERROR :
			case UDF_PART_MAPPING_VIRTUAL :
				/* logical volume integrity warning is not usefull for these types */
				break;
			default :
				if (logical->logvol_state != UDF_INTEGRITY_CLOSED) {
					printf("UDF: Warning, logical volume is not closed!\n");
				};
				break;
		};

		pmap_pos += pmap_size;		/* variable length array :( */
	};

	/* flag all partion mappings data and metadata writable */
	SLIST_FOREACH(part_mapping, &logical->part_mappings, next_mapping) {
		part_mapping->data_writable     = 1;
		part_mapping->metadata_writable = 1;
	};

	/* update writable flags depending on mapping type */
	SLIST_FOREACH(part_mapping, &logical->part_mappings, next_mapping) {
		switch (part_mapping->udf_part_mapping_type) {
			case UDF_PART_MAPPING_ERROR :
				part_mapping->data_writable     = 0;
				part_mapping->metadata_writable = 0;
				break;
			case UDF_PART_MAPPING_PHYSICAL :
				break;
			case UDF_PART_MAPPING_VIRTUAL :
			case UDF_PART_MAPPING_META :
				/*
				 * These are special in that there is a special metadata partition where no data
				 * is meant to be written on and vice versa
				 */

				/* find the associated data partition */
				data_part_num     = part_mapping->udf_phys_part_num;
				SLIST_FOREACH(data_part_mapping, &logical->part_mappings, next_mapping) {
					if (data_part_mapping->udf_phys_part_num == data_part_num) {
						if (data_part_mapping != part_mapping) {
							data_part_mapping->metadata_writable = 0;
							break;
						};
					};
				};
				part_mapping->data_writable = 0;
				break;
			case UDF_PART_MAPPING_SPARABLE :
				break;
		};
	};

	return 0;
}


/* not called often; free incomming when not needed */
int udf_proc_part(struct udf_pri_vol *primary, struct udf_partition **current, struct part_desc *incomming) {
	struct udf_partition *udf_partition;
	struct udf_volumeset *udf_volset;
	uint32_t new_part_num;

	assert(primary);
	assert(primary->pri_vol);

	udf_volset = udf_search_volumeset(primary->pri_vol->volset_id);
	assert(udf_volset);

	new_part_num = udf_rw16(incomming->part_num);
	/* check if its a partition type we recognize */
	if (strncmp(incomming->contents.id, "+NSR0", 5) != 0) {
		fprintf(stderr, "Unrecognized partition content type %s encountered; ignoring\n", incomming->contents.id);
		free(incomming);
		return 0;
	};

	/* look if we allready got it */

	SLIST_FOREACH(udf_partition, &udf_volset->parts, next_partition) {
		if (udf_rw16(udf_partition->partition->part_num) == new_part_num) break;
	};

	/* we have space... now check if this is a newer one than the one known */
	if (udf_partition) {
		if (udf_rw32(incomming->seq_num) < udf_rw32(udf_partition->partition->seq_num)) {
			/* its an older version */
			free(incomming);
			return 0;
		};
	} else {
		/* get us a new udf_partition */
		udf_partition = calloc(1, sizeof(struct udf_partition));
		if (!udf_partition) {
			free(incomming);
			return ENOMEM;
		};

		/* link it in */
		SLIST_INSERT_HEAD(&udf_volset->parts, udf_partition, next_partition);
	};
	assert(udf_partition);

	/* copy this new partition descriptor in the list */
	if (udf_partition->partition) free(udf_partition->partition);
	udf_partition->partition   = incomming;
	udf_partition->udf_session = primary->udf_session;
	udf_volset->max_partnum = MAX(udf_volset->max_partnum, new_part_num+1);

	/* just in case */
	TAILQ_INIT(&udf_partition->unalloc_space_queue);
	TAILQ_INIT(&udf_partition->freed_space_queue);

	if (current) *current = udf_partition;

	return 0;
}


/* not called often; free incomming when not needed */
int udf_proc_filesetdesc(struct udf_log_vol *udf_log_vol, struct fileset_desc *incomming) {
	struct udf_mountpoint *mp;

	if (udf_rw16(incomming->tag.id) != TAGID_FSD) {
		printf("IEEE! Encountered a non TAGID_FSD in this fileset descriptor sequence!!!\n");
		free(incomming);
		return EFAULT;
	};

	/* lookup fileset descriptor in this logical volume; interestingly fileset_num is KEY! */
	SLIST_FOREACH(mp, &udf_log_vol->mountpoints, logvol_next) {
		if (mp->fileset_desc->fileset_num == incomming->fileset_num) break;
	};

	if (!mp) {
		/* add a new mountpoint! */
		mp = calloc(1, sizeof(struct udf_mountpoint));
		if (!mp) {
			free(incomming);
			return ENOMEM;
		};
		mp->fileset_desc = incomming;

		/* insert into udf_log_vol and into mountables list */
		SLIST_INSERT_HEAD(&udf_log_vol->mountpoints, mp, logvol_next);
		SLIST_INSERT_HEAD(&udf_mountables, mp, all_next);
	} else {
		/* should we update mountpoint? */
		if (udf_rw32(incomming->fileset_desc_num) <= udf_rw32(mp->fileset_desc->fileset_desc_num)) {
			/* we allready got a newer one */
			free(incomming);
			return 0;
		};

		fprintf(stderr, "UDF DEBUG: would be updating mountpoint... HELP!\n");
		/* FIXME delete all inode hash entries */
		/* XXX how to do that? inodes OK but associated vnodes? XXX */
#if 0
		if (!SLIST_EMPTY(&mp->inodes)) {
			printf("UDF: asked to delete mountpoint with inodes in hashtable!\n");
			printf("Can't cope with that... aborting\n");
			exit(1);
		};
#endif

		/* free old information (allready in lists though!) */
		free(mp->fileset_desc);
		free(mp->mount_name);
	};

	mp->udf_log_vol  = udf_log_vol;
	mp->fileset_desc = incomming;
	mp->mount_name   = strdup(udf_get_compound_name(mp));

	STAILQ_INIT(&mp->dirty_nodes);
	UDF_MUTEX_INIT(&mp->dirty_nodes_mutex);

	return 0;
}


int udf_retrieve_volume_space(struct udf_discinfo *disc, struct udf_session *udf_session, struct extent_ad *extent) {
	struct udf_pri_vol *udf_pri_vol;
	union dscrptr	*dscr;
	uint32_t	 sector, length, dscr_len, num_sectors;
	int		 tag_id;
	int		 error;

	udf_pri_vol = NULL;

	sector = udf_rw32(extent->loc);
	length = udf_rw32(extent->len);

	error = 0;	/* XXX zero length area's possible? XXX */
	while (length) {
		error = udf_read_session_descriptor(udf_session, sector, "volume descriptor", &dscr, &dscr_len);
		if (error) {
			if (dscr) free(dscr);
			break;
		};

		tag_id = udf_rw16(dscr->tag.id);
		num_sectors = (dscr_len + UDF_SECTOR_SIZE-1) / UDF_SECTOR_SIZE;

		/* proc volume descriptor starting at sector `volume_sector' */
		UDF_VERBOSE_MAX(udf_dump_descriptor(dscr));
		switch (tag_id) {
			case TAGID_PRI_VOL       :
				error = udf_proc_pri_vol(udf_session, &udf_pri_vol, &dscr->pvd);
				break;
			case TAGID_PARTITION     :
				error = udf_proc_part(udf_pri_vol, NULL, &dscr->pd);
				break;
			case TAGID_LOGVOL        :
				error = udf_proc_log_vol(udf_pri_vol, NULL, &dscr->lvd);
				break;
			case TAGID_TERM          :
				free(dscr);
				return 0;	/* terminator */
			case TAGID_UNALLOC_SPACE :
				/* unallocated space descriptor */
				/* Specifies space that is not claimed yet in partitions (!) */
				UDF_VERBOSE(fprintf(stderr, "UDF: Unhandled `unallocated space descriptor'; ignored\n"));
				break;
			case TAGID_IMP_VOL       :
				/* implemenation use volume descriptor */
				/* Specifies information relevant for the implementator */
				UDF_VERBOSE_MAX(fprintf(stderr, "UDF: `implementation use volume descriptor' ignored\n"));
				break;
			case TAGID_VOL           :
				fprintf(stderr, "UDF : untested volume space extender encountered\n");
				break;
			default :
				printf("XXX Unhandled volume sequence %d; freeing\n", tag_id);
				free(dscr);
				break;
		};

		length -= num_sectors * UDF_SECTOR_SIZE;
		sector += num_sectors;

		if (tag_id == TAGID_VOL) {
			sector = udf_rw32(dscr->vdp.next_vds_ex.loc);
			length = udf_rw32(dscr->vdp.next_vds_ex.len);
			free(dscr);
		};
	};

	return error;
}


int udf_get_filelength(union dscrptr *dscr, uint64_t *length) {
	int32_t	fe_tag;

	fe_tag = udf_rw16(dscr->tag.id);
	if (fe_tag == TAGID_FENTRY) {
		*length = udf_rw64(dscr->fe.inf_len);
		return 0;
	} else if (fe_tag == TAGID_EXTFENTRY) {
		*length = udf_rw64(dscr->efe.inf_len);
		return 0;
	};
	return ENOENT;
}


int udf_check_for_vat(struct udf_log_vol *udf_log_vol, struct udf_part_mapping *part_mapping, struct file_entry *file_entry) {
	struct udf_node		*vat_udf_node;
	struct long_ad		 udf_icbptr;
	struct regid 		*regid;
	struct uio		 vat_uio;
	struct iovec		 vat_iovec;
	uint64_t		 vat_length;
	uint32_t		*vat_pos;
	uint8_t			*vat, *pos;
	int			 error;

	/* prepare a `uio' structure for reading in complete VAT file */
	error = udf_get_filelength((union dscrptr *) file_entry, &vat_length);
	if (error) return error;

	vat = malloc(vat_length);
	if (!vat) return ENOMEM;

	/* move to uio_newuio(struct uio *uio) with fixed length uio_iovcnt? */
	bzero(&vat_uio, sizeof(struct uio));
	vat_uio.uio_rw     = UIO_WRITE;	/* WRITE into this space */
	vat_uio.uio_iovcnt = 1;
	vat_uio.uio_iov    = &vat_iovec;
	vat_uio.uio_offset = 0;		/* begin at the start */
	vat_uio.uio_resid  = vat_length;
	/* fill in IO vector */
	vat_uio.uio_iov->iov_base = vat;
	vat_uio.uio_iov->iov_len  = vat_length;

	/* prepare udf_icbptr file node for easy file reading */
	udf_icbptr.loc.part_num	= udf_rw16(part_mapping->udf_pmap->pmv.part_num);
	udf_icbptr.loc.lb_num	= 1;	/* not used, but may not be zero */
	udf_icbptr.len		= 2048;	/* not used, but may not be zero */

	/*
	 * this udf_node creation and disposing may look a bit inefficient but
	 * its beneficiary for normal file access.  its only used once for
	 * reading in the VAT.
	 */

	/* create the udf_vat_node; anonymous since it can't be in a mountpoint */
	error = udf_readin_anon_udf_node(udf_log_vol, (union dscrptr *) file_entry, &udf_icbptr, "VAT", &vat_udf_node);
	if (!error) {
		error = udf_read_file_part_uio(vat_udf_node, "VAT contents", 0, &vat_uio);
	};

	/* XXX allow for SHORT VAT's ? XXX */
	if (!error) {
		if (vat_uio.uio_resid) {
			fprintf(stderr, "Warning: VAT file can't be read in completely\n");
		};

		part_mapping->vat_udf_node = vat_udf_node;
		part_mapping->vat          = (struct udf_vat *) vat;
		part_mapping->vat_length   = vat_length;

		/* extract next unique file ID from the VAT file entry's unique ID incremented by one */
		udf_log_vol->next_unique_id = vat_udf_node->unique_id;
		udf_increment_unique_id(udf_log_vol);

		/*XXX move to be a part_mapping_quirk ? XXX */
		if (file_entry->icbtag.file_type == UDF_ICB_FILETYPE_VAT) {
			/* we are in UDF 2.00+ userland */
			part_mapping->vat_translation = &(part_mapping->vat->data[0]) + udf_rw16(part_mapping->vat->impl_use_len);
			part_mapping->vat_entries     = (udf_rw16(part_mapping->vat->header_len) - udf_rw16(part_mapping->vat->impl_use_len))/4;
			udf_log_vol->num_files        = udf_rw32(part_mapping->vat->num_files);
			udf_log_vol->num_directories  = udf_rw32(part_mapping->vat->num_directories);
			udf_log_vol->min_udf_readver  = udf_rw16(part_mapping->vat->min_udf_readver);
			udf_log_vol->min_udf_writever = udf_rw16(part_mapping->vat->min_udf_writever);
			udf_log_vol->max_udf_writever = udf_rw16(part_mapping->vat->max_udf_writever);
		} else {
			/* still in the old UDF 1.50 userland; update? its notoriously broken */
			/* check this VAT for validity */

			vat_length = udf_rw64(file_entry->inf_len);	/* hack hack  */
			part_mapping->vat_entries = (vat_length-36)/4;	/* definition */
			part_mapping->vat_length = vat_length;
			part_mapping->vat_translation = vat;
			part_mapping->vat = NULL;

			/* check the old UDF 1.50 VAT */
			DEBUG(printf("CHECK UDF 1.50 VAT\n"));
			pos = part_mapping->vat_translation;
			vat_pos = (uint32_t *) pos;
			regid = (struct regid *) (vat_pos+part_mapping->vat_entries);

#if 0
udf_dump_long_ad("XXX possible VAT", &vat_udf_node->file_icb);
{
	static int c = 0;
	char buf[1000];
	int f;

	c++;
	sprintf(buf, "VAT_%02d", c);
	f = open(buf, O_CREAT | O_WRONLY, 0644);
	write(f, vat, vat_length);
	close(f);
};
#endif

			error = (strncmp(regid->id, "*UDF Virtual Alloc Tbl", 22) == 0) ? 0 : ENOENT;
			if (error) part_mapping->vat_translation = NULL;	/* no reference for freed memory please */
		};
	};

	/* clean up uio structure */
	if (error) {
		if (vat) free(vat);
		if (vat_udf_node) udf_dispose_udf_node(vat_udf_node);
		part_mapping->vat_udf_node = NULL;
	};

	return error;
}


int udf_retrieve_supporting_tables(struct udf_log_vol *udf_log_vol) {
	struct udf_part_mapping  *part_mapping;
	struct udf_session	 *udf_session;
	union  dscrptr		 *possible_vat_fe;
	union  dscrptr		 *sparing_table_dscr;
	uint8_t			 *vat;
	uint32_t		  spar_loc;
	uint64_t		  vat_loc, last_vat_loc;
	uint32_t		  vat_part;
	int			  part_num, spar_num;
	int			  session_num;
	int			  error;

	/*
	 * if there are any virtual or sparable partition in this logical
	 * volume, try to find their supporting tables so we can find the rest
	 */
	SLIST_FOREACH(part_mapping, &udf_log_vol->part_mappings, next_mapping) {
		part_num = part_mapping->udf_virt_part_num;
		UDF_VERBOSE_TABLES(printf("\tFor partition mapping %d->%d\n", part_num, part_mapping->udf_phys_part_num));
		switch (part_mapping->udf_part_mapping_type) {
			case UDF_PART_MAPPING_ERROR    :
				/* nothing to be done for these */
				break;
			case UDF_PART_MAPPING_PHYSICAL :
				/* nothing to be done for these; no supporting tables */
				break;
			case UDF_PART_MAPPING_VIRTUAL  :
				/*
				 * we have to find a good VAT at the END of the session. Since VAT's are
				 * only to be used on WORM's and need to written as last, the strategy is
				 * to go for the predicted end of this session and walk UP
				 */
				udf_session = udf_log_vol->primary->udf_session;
				session_num = udf_session->session_num;

				UDF_VERBOSE_TABLES(printf("\t\tSearching for the VAT :"));
				if (udf_session->session_length == 0) {
					UDF_VERBOSE(
						printf("\t\tThis virtual partition is inaccessible since its its size is not known;\n");
						printf("\t\tTry to insert the disc in a CD or DVD recordable device to access it.\n");
					);
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
					continue;
				};

				vat = NULL; vat_part = 0;
				if (udf_session->disc->next_writable[session_num]) {
					vat_loc = udf_session->disc->next_writable[session_num];
				} else {
					vat_loc = udf_session->disc->session_end[session_num];
				}
				last_vat_loc = vat_loc + 1024;
				vat_loc -= 20; /* give some extra slack since sizes are not allways given up correctly */

				/* try to find the fileid for the VAT */
				do {
					DEBUG(
						printf("Trying VAT at sector %d in session\n", (int) vat_loc)
					);
					error = udf_read_session_descriptor(udf_session, vat_loc, "VAT file entry", &possible_vat_fe, NULL);
					if (!error) error = udf_check_tag_presence(possible_vat_fe, TAGID_FENTRY);
					if (!error) error = udf_check_tag_payload( possible_vat_fe);
					if (!error) error = udf_check_for_vat(udf_log_vol, part_mapping, &possible_vat_fe->fe);
					if (!error) {
						break;
					} else {
						if (possible_vat_fe) free(possible_vat_fe);
						vat_loc++;
						if (vat_loc > last_vat_loc) error = EIO;
					};
				} while (error != EIO);
				if (error) {
					printf("IEE!! I was looking for a VAT but didnt find it\n");
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
					udf_log_vol->broken = 1;
					continue;
				};
				UDF_VERBOSE_TABLES(printf("found VAT file-entry at device logical sector %d\n", (uint32_t) vat_loc));
				UDF_VERBOSE_TABLES(printf("\t\tFound %d byte VAT descriptor+table\n", (uint32_t) part_mapping->vat_length));

				UDF_VERBOSE_TABLES(udf_dump_descriptor(possible_vat_fe));
				UDF_VERBOSE_MAX(udf_dump_vat_table(part_mapping));

				if (part_mapping->vat) {
					/* update UDF versions */
					udf_log_vol->min_udf_readver  = udf_rw16(part_mapping->vat->min_udf_readver);
					udf_log_vol->min_udf_writever = udf_rw16(part_mapping->vat->min_udf_writever);
					udf_log_vol->max_udf_writever = udf_rw16(part_mapping->vat->max_udf_writever);
					/* XXX update logvol name XXX */
				};
				break;
			case UDF_PART_MAPPING_SPARABLE :
				/* we have to find a good sparing table; address are in device logical blocks */
				udf_session = udf_log_vol->primary->udf_session;

				for(spar_num = 0; spar_num < part_mapping->udf_pmap->pms.n_st; spar_num++) {
					spar_loc = udf_rw32(part_mapping->udf_pmap->pms.st_loc[spar_num]);

					/* fetch spar_loc's table ; on THIS session. */
					error = udf_read_session_descriptor(udf_session, spar_loc, "Sparing table", &sparing_table_dscr, NULL);
					if (!error) error = udf_check_tag_presence(sparing_table_dscr, TAGID_SPARING_TABLE);
					if (!error) {
						UDF_VERBOSE_TABLES(printf("\t\tFound the sparing table\n"));
						part_mapping->sparing_table = &sparing_table_dscr->spt;
						break;
					} else {
						if (sparing_table_dscr) free(sparing_table_dscr);
					};
				};
				if (!part_mapping->sparing_table) {
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
				};
				UDF_VERBOSE_TABLES(udf_dump_descriptor((union dscrptr *) part_mapping->sparing_table));
				break;
			case UDF_PART_MAPPING_META :
				/*
				 * set up common locator parts; the files are located inside the `part_num' partion where
				 * this partition is a added layer on.
				 */
				printf("XXX WARNING : reading `supporting tables' for a metadata partition; code not tested yet\n");
				printf("XXX metadata partition not supported yet !!!! XXX\n");
#if 0
				location.loc.part_num = udf_rw16(part_mapping->udf_pmap->pmm.part_num);
				location.len          = UDF_SECTOR_SIZE;		/* defined as maximum size */

				UDF_VERBOSE_TABLES(printf("Reading metadata partition filedescriptor\n"));
				location.loc.lb_num   = udf_rw32(part_mapping->udf_pmap->pmm.meta_file_lbn);
				udf_create_icb_node(udf_log_vol, &location, &part_mapping->meta_file);
				UDF_VERBOSE_TABLES(udf_dump_descriptor((union dscrptr *) part_mapping->meta_file));

				UDF_VERBOSE_TABLES(printf("Reading metadata partition mirror filedescriptor\n"));
				location.loc.lb_num   = udf_rw32(part_mapping->udf_pmap->pmm.meta_mirror_file_lbn);
				udf_create_icb_node(udf_log_vol, &location, &part_mapping->meta_mirror_file);
				UDF_VERBOSE_TABLES(udf_dump_descriptor((union dscrptr *) part_mapping->meta_mirror_file));

				UDF_VERBOSE_TABLES(printf("Reading metadata partition allocation bitmap filedescriptor\n"));
				location.loc.lb_num   = udf_rw32(part_mapping->udf_pmap->pmm.meta_bitmap_file_lbn);
				udf_create_icb_node(udf_log_vol, &location, &part_mapping->meta_bitmap_file);
				UDF_VERBOSE_TABLES(udf_dump_descriptor((union dscrptr *) part_mapping->meta_bitmap_file));

				/* if something is wrong, then mark it as a broken partition */
				if (!(part_mapping->meta_file || part_mapping->meta_mirror_file) || !part_mapping->meta_bitmap_file) {
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
				};
				break;
#endif
		};
	};
	UDF_VERBOSE_TABLES(printf("\n"));
	if (udf_log_vol->broken) return EIO;

	return 0;
}


int udf_retrieve_space_tables(struct udf_log_vol *udf_log_vol) {
	struct udf_partition	 *udf_partition;
	struct udf_part_mapping  *part_mapping;
	struct part_hdr_desc	 *part_hdr_desc;
	union  dscrptr		 *dscrptr;
	uint32_t		  sector;
	uint32_t		  lb_size;
	uint64_t		  length;
	int			  part_num;
	int			  error;

	lb_size = udf_rw32(udf_log_vol->log_vol->lb_size);
	SLIST_FOREACH(part_mapping, &udf_log_vol->part_mappings, next_mapping) {
		part_num = part_mapping->udf_virt_part_num;
		UDF_VERBOSE_TABLES(printf("\tFor partition mapping %d->%d\n", part_num, part_mapping->udf_phys_part_num));

		/* retrieve and process unallocated- and freed-space information for all used partitions of the logvol */
		error = udf_logvol_vpart_to_partition(udf_log_vol, part_num, NULL, &udf_partition);
		assert(udf_partition);
		part_hdr_desc = (struct part_hdr_desc *) udf_partition->partition->contents_use;

		sector   = udf_rw32(part_hdr_desc->unalloc_space_table.lb_num);
		length   = udf_rw32(part_hdr_desc->unalloc_space_table.len);		/* needed? */
		if (length) {
			error = udf_read_logvol_descriptor(udf_log_vol, part_num, sector, "Unallocated space table", &dscrptr, NULL);
			UDF_VERBOSE_MAX(printf("\tUnalloced space table\n"));
			UDF_VERBOSE_MAX(udf_dump_descriptor(dscrptr));
			//udf_process_space_table(&udf_partition->unalloc_space, dscrptr);
			free(dscrptr);
		};

		sector   = udf_rw32(part_hdr_desc->unalloc_space_bitmap.lb_num);
		length   = udf_rw32(part_hdr_desc->unalloc_space_bitmap.len);
		if (length) {
			error = udf_read_logvol_descriptor(udf_log_vol, part_num, sector, "Unallocated space bitmap", &dscrptr, NULL);
			UDF_VERBOSE_MAX(printf("\tUnalloced space bitmap\n"));
			UDF_VERBOSE_MAX(udf_dump_descriptor(dscrptr));
			udf_read_in_space_bitmap(&udf_partition->unalloc_space_queue, &dscrptr->sbd, lb_size, &udf_partition->free_unalloc_space);
			printf("Physical partition's %d freespace = %"PRIu64"\n", part_num, udf_partition->free_unalloc_space);
			free(dscrptr);
		};

		sector    = udf_rw32(part_hdr_desc->freed_space_table.lb_num);
		length    = udf_rw32(part_hdr_desc->freed_space_table.len);
		if (length) {
			error = udf_read_logvol_descriptor(udf_log_vol, part_num, sector, "Freed space table", &dscrptr, NULL);
			UDF_VERBOSE_MAX(printf("\tFreed space table\n"));
			UDF_VERBOSE_MAX(udf_dump_descriptor(dscrptr));
			//udf_process_space_table(&udf_partition->freed_space, dscrptr);
			free(dscrptr);
		};

		sector    = udf_rw32(part_hdr_desc->freed_space_bitmap.lb_num);
		length    = udf_rw32(part_hdr_desc->freed_space_bitmap.len);
		if (length) {
			error = udf_read_logvol_descriptor(udf_log_vol, part_num, sector, "Freed space bitmap", &dscrptr, NULL);
			UDF_VERBOSE_MAX(printf("\tFreed space bitmap\n"));
			UDF_VERBOSE_MAX(udf_dump_descriptor(dscrptr));
			udf_read_in_space_bitmap(&udf_partition->freed_space_queue, &dscrptr->sbd, lb_size, &udf_partition->free_freed_space);
			free(dscrptr);
		};
	};
	UDF_VERBOSE_TABLES(printf("\n"));

	return 0;
}


/*
 * Fileset descriptors are on a logical volume's partitions; since virtual
 * partitions are then also possible its OK to use the VAT for redefining the
 * fileset descriptors.
 */
int udf_retrieve_fileset_descriptor(struct udf_log_vol *udf_log_vol) {
	struct udf_mountpoint	*mountable;
	struct long_ad		*fsd_loc;
	struct fileset_desc	*new_fsd;
	struct udf_node		*vnode;
	uint32_t		 part_num, lb_num, length;
	int32_t			 error, bucket;
	ino_t			 hashkey;

	error = 0; /* flag OK */

	fsd_loc   = &udf_log_vol->log_vol->_lvd_use.fsd_loc;
	part_num  = udf_rw16(fsd_loc->loc.part_num);
	lb_num    = udf_rw32(fsd_loc->loc.lb_num);
	length    = udf_rw32(fsd_loc->len);

	while (length && !error) {
		UDF_VERBOSE_TABLES(
			printf("\tFileset descriptor extent remaining at sector %d within partion %d for %d bytes\n", lb_num, part_num, length)
		);

		/* only go for ONE fsb at a time */

		error = udf_read_logvol_descriptor(udf_log_vol, part_num, lb_num, "Fileset descriptor", (union dscrptr **) &new_fsd, NULL);
		if (!error) error = udf_check_tag((union dscrptr *) new_fsd);
		if (error || !new_fsd || (new_fsd && (udf_rw16(new_fsd->tag.id) == TAGID_TERM))) {
			/* end of sequence */
			UDF_VERBOSE_TABLES(
				printf("\t\t(Terminator) ");
				if (!new_fsd || error) printf("; unrecorded"); else printf("; explicit");
				printf("\n");
			);
			if (new_fsd) free(new_fsd);
			break;
		};

		UDF_VERBOSE_MAX(udf_dump_descriptor((union dscrptr *) new_fsd));
		udf_proc_filesetdesc(udf_log_vol, new_fsd);

		if (udf_rw32(new_fsd->next_ex.len) == 0) {
			/* next entry */
			lb_num += 1;
			length -= UDF_SECTOR_SIZE;
		} else {
			/* follow the next extent */
			fsd_loc  = &new_fsd->next_ex;
			part_num = udf_rw16(fsd_loc->loc.part_num);
			lb_num   = udf_rw32(fsd_loc->loc.lb_num);
			length   = udf_rw32(fsd_loc->len);
		};
	};
	UDF_VERBOSE_TABLES(printf("\n"));

	if (error) return error;

	/* if no error occured, create rootdir udf_nodes */
	SLIST_FOREACH(mountable, &udf_log_vol->mountpoints, logvol_next) {
		/* errors are OK */
		udf_readin_anon_udf_node(udf_log_vol, NULL, &mountable->fileset_desc->rootdir_icb,   "Rootdir",   &mountable->rootdir_node);
		udf_readin_anon_udf_node(udf_log_vol, NULL, &mountable->fileset_desc->streamdir_icb, "Streamdir", &mountable->streamdir_node);

		/* keep names the same ? (duplicate code ahead ... ) */
		if (mountable->rootdir_node) {
			vnode = mountable->rootdir_node;

			strcpy(vnode->dirent.d_name,   mountable->mount_name);
			vnode->dirent.d_type = DT_DIR;
			vnode->mountpoint = mountable;
			vnode->stat.st_uid = vnode->stat.st_gid = UINT_MAX;
			vnode->stat.st_mode = 0777;

			/* XXX split off inserting? XXX */
			hashkey = udf_calc_hash(&mountable->fileset_desc->rootdir_icb);
			vnode->hashkey = hashkey;
			bucket  = hashkey & UDF_INODE_HASHMASK;
			LIST_INSERT_HEAD(&mountable->udf_nodes[bucket], vnode, next_node);
		};
		if (mountable->streamdir_node) {
			vnode = mountable->streamdir_node;

			strcpy(vnode->dirent.d_name,   mountable->mount_name);
			vnode->dirent.d_type = DT_DIR;
			vnode->mountpoint = mountable;
			vnode->stat.st_uid = vnode->stat.st_gid = UINT_MAX;
			vnode->stat.st_mode = 0777;

			/* XXX split off inserting? XXX */
			hashkey = udf_calc_hash(&mountable->fileset_desc->streamdir_icb);
			vnode->hashkey = hashkey;
			bucket  = hashkey & UDF_INODE_HASHMASK;
			LIST_INSERT_HEAD(&mountable->udf_nodes[bucket], vnode, next_node);
		};
	};

	return 0;
}


int udf_check_writable_filesets(struct udf_log_vol *udf_log_vol) {
	//struct udf_part_mapping *part_mapping;

	udf_log_vol->primary->udf_session->writable = 1;
	/* follow all mountpoints of this logical volume and check if they are writable */
	
	//part_mapping = part_mapping->next;
	return 0;
};


/*
 * udf_eliminate_predescessor_volumesets()
 *
 * We are faced with a curious problem : we are to examine the partitions and
 * determine which are successors of eachother.  This is propably most
 * relevant only on WORM media though. We could consider the following rules :
 * 	1) `glue' according to strict UDF level 1 rules ?
 * 	2) use heuristics info i.e. NERO, DirectCD and mkisofs quirks ?
 * 	3) use overlapping partitions to detect relationships ?
 *
 * Using 1 would imply no multi-volume discs and thus glue everything but that
 * could easily be wrong. Selecting by volumeset names is not possible for
 * Nero f.e. just creates random volumeset names every session and uses no
 * volume version information and thus also violates the UDF rules.
 *
 * Using 2 would be tricky; we know a few programs but what if more are
 * developped? We then would be at loss.
 *
 * Using 3 would imply some calculation but is fail-safe in both supporting
 * multiple volumes on one disc (they seperate) and in supporting
 * multi-session WORM media for these will refer to eachother. Offcource NERO
 * could faul this by just extending the zero partion to the whole disc in its
 * ignorance and thus create false overlapping over other independent
 * sessions. This is to be investigated. I dont know how NERO will react on
 * this situation.
 *
 * Propably method 3 would be good to try :
 *
 * Follow the disc and check for all sessions in order to mark the ones with
 * overlapping partitions as `inactive' and keep the latest one active.
 * Sessions with the `local' quirk are seperate allmost by default; should
 * be change the offsets? would not be too difficult but possible.
 */

void udf_eliminate_predescessor_volumesets(struct udf_discinfo *disc) {
	struct udf_volumeset	*anc_vol_set;
	struct udf_volumeset	*sib_vol_set;
	struct pri_vol_desc	*anc_pri_vol;
	struct pri_vol_desc	*sib_pri_vol;
	struct udf_partition	*anc_part;
	struct udf_partition	*sib_part;
	int			 anc_partnum;
	int			 sib_partnum;
	uint32_t		 anc_start, anc_end;
	uint32_t		 sib_start, sib_end;
	uint32_t		 overlap_start, overlap_end;
	uint32_t		 anc_session;
	uint32_t		 sib_session;

	SLIST_FOREACH(anc_vol_set, &udf_volumeset_list, next_volumeset) {
		anc_pri_vol = STAILQ_FIRST(&anc_vol_set->primaries)->pri_vol;
		sib_vol_set = SLIST_NEXT(anc_vol_set, next_volumeset);
		while (sib_vol_set) {
			sib_pri_vol = STAILQ_FIRST(&sib_vol_set->primaries)->pri_vol;
			DEBUG(
				printf("checking volset %s with volset %s\n", anc_pri_vol->volset_id+1, sib_pri_vol->volset_id+1)
			);
			/* compare these two volume sets but only process partitions on _this_ disc */
			SLIST_FOREACH(anc_part, &anc_vol_set->parts, next_partition) {
				if (anc_part->udf_session->disc != disc)  continue;

				anc_session = anc_part->udf_session->session_num;
				anc_start = 0;
				if (disc->session_quirks[anc_session] & CD_SESS_QUIRK_SESSION_LOCAL)
					anc_start += disc->session_start[anc_session];
				anc_start += udf_rw32(anc_part->partition->start_loc);
				anc_end    = anc_start + udf_rw32(anc_part->partition->part_len);

				SLIST_FOREACH(sib_part, &sib_vol_set->parts, next_partition) {
					if (sib_part->udf_session->disc != disc)  continue;

					sib_session = sib_part->udf_session->session_num;
#if 1
					/* can `session local' volumes even be considered part/successor ? */
					if (disc->session_quirks[sib_session] & CD_SESS_QUIRK_SESSION_LOCAL) continue;
#endif
					sib_start = 0;
					if (disc->session_quirks[sib_session] & CD_SESS_QUIRK_SESSION_LOCAL)
						sib_start += disc->session_start[sib_session];
					sib_start += udf_rw32(sib_part->partition->start_loc);
					sib_end    = sib_start + udf_rw32(sib_part->partition->part_len);
DEBUG(
anc_partnum = udf_rw16(anc_part->partition->part_num);
sib_partnum = udf_rw16(sib_part->partition->part_num);
printf("\t\tchecking partition %d with partition %d ([%d-%d] x [%d-%d])\n", anc_partnum, sib_partnum, anc_start, anc_end, sib_start, sib_end)
);
					overlap_start = MAX(sib_start, anc_start);
					overlap_end   = MIN(sib_end,   sib_end);
					if (overlap_start < overlap_end)  {
DEBUG(
printf("\t\t\tOVERLAP!\n")
);
						if (sib_session < anc_session) {
							/* the sibbling is older */
			UDF_VERBOSE_TABLES(
				printf("\tVolume set ");
				udf_dump_id(NULL, 128, anc_pri_vol->vol_id, &anc_pri_vol->desc_charset);
				printf(" is a newer version of volume set ");
				udf_dump_id(NULL, 128, sib_pri_vol->vol_id, &sib_pri_vol->desc_charset);
				printf("\n");
			);
							sib_vol_set->obsolete = 1;
							break;
						};
					}; /* overlap */
					if (sib_vol_set->obsolete) break;
				}; /* sibling partition */
				if (sib_vol_set->obsolete) break;
			}; /* ancestor partition */
			sib_vol_set = SLIST_NEXT(sib_vol_set, next_volumeset);
		}; /* sibling volume set */
	}; /* ancestor volume set */
}


int udf_add_session_to_discinfo(struct udf_discinfo *disc, int session, struct anchor_vdp *avdp, int error) {
	struct udf_session	*udf_session;

	udf_session = calloc(1, sizeof(struct udf_session));
	assert(udf_session);

	if (!error) {
		memcpy(&udf_session->anchor, avdp, sizeof(struct anchor_vdp));
	};

	udf_session->disc              = disc;
	udf_session->session_num       = session;
	udf_session->session_offset    = 0;
	udf_session->session_length    = disc->session_end[session] - disc->session_start[session];
	disc->session_quirks[session]  = 0;

	/* writable session administration */
	udf_session->writable = 0;		/* default off */
	error = udf_init_session_caches(udf_session);

	if (!error) {
		/* detect quirks */
		/* XXX session local disabled due to wrong heuristic XXX */
#if 0
		if (disc->session_start[session] > 0) {
			if ((udf_session->anchor.main_vds_ex.loc < disc->session_start[session])) {
				disc->session_quirks[session] |= CD_SESS_QUIRK_SESSION_LOCAL;
				udf_session->session_offset = disc->session_start[session];
			};
		};
#endif
	};

	/* add to tail of session list */
	STAILQ_INSERT_TAIL(&disc->sessions, udf_session, next_session);

	disc->num_udf_sessions++;

	/* record status of this volume */
	disc->session_is_UDF[session] = error ? 0 : 1;

	return error;
}


int udf_get_anchors(struct udf_discinfo *disc) {
	uint8_t			 sector[UDF_SECTOR_SIZE];
	union dscrptr		*dscr = (union dscrptr *) sector;
	uint32_t		 session_start, session_end;
	int			 session, error;

	/* Get all anchors */
	STAILQ_INIT(&disc->sessions);

	for(session = 0; session < disc->num_sessions; session++) {
		/* check for anchors ; no volume recognition data ? */
		session_start = disc->session_start[session];
		session_end   = disc->session_end  [session];
		error = udf_read_physical_sectors(disc, session_end-1, 1, "Anchor", sector);	/* -1? */
		if (!error) error = udf_check_tag_presence(dscr, TAGID_ANCHOR);
		if (error) {
			error = udf_read_physical_sectors(disc, session_end - 256, 1, "Anchor", sector);
			if (!error) error = udf_check_tag_presence(dscr, TAGID_ANCHOR);
			if (error) {
				error = udf_read_physical_sectors(disc, session_start + 256, 1, "Anchor", sector);
				if (!error) error = udf_check_tag_presence(dscr, TAGID_ANCHOR);
				if (error) {
					/* unclosed CD recordable case due to track reservation for iso9660 filesystems*/
					error = udf_read_physical_sectors(disc, session_start + 512, 1, "Anchor", sector);
					if (!error) error = udf_check_tag_presence(dscr, TAGID_ANCHOR);
				};
			}
		};

		udf_add_session_to_discinfo(disc, session, (struct anchor_vdp *) sector, error);
	};
	return 0;
}


int udf_get_volumeset_space(struct udf_discinfo *disc) {
	struct udf_session *udf_session;
	int error;

	/* Rip all volume spaces */
	UDF_VERBOSE(printf("Retrieving Volume information\n"));
	STAILQ_FOREACH(udf_session, &disc->sessions, next_session) {
		UDF_VERBOSE_MAX(printf("Session %d volumes : \n", udf_session->session_num));

		error = udf_retrieve_volume_space(disc, udf_session, &udf_session->anchor.main_vds_ex);
		if (error) {
			printf("\nRetrieving session %d's reserve volume information due to error\n", udf_session->session_num);
			error = udf_retrieve_volume_space(disc, udf_session, &udf_session->anchor.reserve_vds_ex);
		};
	};
	UDF_VERBOSE_MAX(printf("\n\n"));
	return 0;
}


int udf_get_logical_volumes_supporting_tables(struct udf_discinfo *disc) {
	struct udf_volumeset	*udf_volumeset;
	struct udf_pri_vol	*udf_pri_vol;
	struct udf_log_vol	*udf_log_vol;
	int			 error;

	SLIST_FOREACH(udf_volumeset, &udf_volumeset_list, next_volumeset) {
		if (!udf_volumeset->obsolete) {
			STAILQ_FOREACH(udf_pri_vol, &udf_volumeset->primaries, next_primary) {
				if (udf_pri_vol->udf_session->disc == disc) {
					SLIST_FOREACH(udf_log_vol, &udf_pri_vol->log_vols, next_logvol) {
						/* load in supporting tables */
						UDF_VERBOSE(udf_dump_volume_name("Retrieving support tables for logical volume", udf_log_vol));
						error = udf_retrieve_supporting_tables(udf_log_vol);
	
						/* get fileset descriptors */
						UDF_VERBOSE(udf_dump_volume_name("Retrieving fileset(s) for logical volume", udf_log_vol));
						if (!error) error = udf_retrieve_fileset_descriptor(udf_log_vol);
	
						/* check if the logical volume is writable */
						UDF_VERBOSE(udf_dump_volume_name("Check writability for logical volume", udf_log_vol));
						if (!error) error = udf_check_writable_filesets(udf_log_vol);

						/* load in free/used space tables for writable volsets */
						UDF_VERBOSE(udf_dump_volume_name("Retrieve used/freed space tables for logical volume", udf_log_vol));
						if (!error) error = udf_retrieve_space_tables(udf_log_vol);

						if (error) udf_log_vol->broken = 1;
					}; /* logical */
				}; /* disc */
			}; /* primary */
		}; /* if */
	}; /* volumeset */

	return 0;
};


/******************************************************************************************
 *
 * Disc sync
 *
 ******************************************************************************************/


void udf_sync_tables_callback(int reason, struct udf_wrcallback *wrcallback, int error, uint8_t *sectordata) {
//	struct udf_node *udf_node = (struct udf_node *) wrcallback->structure;

	if (reason == UDF_WRCALLBACK_REASON_PENDING) {
		/* what to do? */
		return;
	};
	if (reason == UDF_WRCALLBACK_REASON_ANULATE) {
		/* what to do? */
		return;
	};
	assert(reason == UDF_WRCALLBACK_REASON_WRITTEN);

#if 0
	DEBUG(
		printf("sync tables write callback called with sector data %p\n", sectordata);
		printf("\toffset    %08d\n", (int) wrcallback->offset);
		printf("\tlb_num    %08d\n", (int) wrcallback->lb_num);
		printf("\tlength    %08d\n", (int) wrcallback->length);
		printf("\terror     %d\n",   (int) error);
	);
#endif

	if (error) {
		printf("UDF error: sync tables write errors in syncnode not fixed!\n");
		return;
	};
}


/* TODO space tables are not coupled on a logical volume but on a partition/disc, so call them on that instead of logvol */
int udf_sync_space_tables(struct udf_log_vol *udf_log_vol) {
	struct udf_partition	 *udf_partition;
	struct udf_part_mapping  *part_mapping;
	struct part_hdr_desc	 *part_hdr_desc;
	struct udf_wrcallback	  wr_callback;
	union  dscrptr		 *dscrptr;
	uint64_t		  length;
	uint32_t		  sector;
	uint32_t		  lb_size, part_start, part_len;
	uint16_t		  dscr_ver;
	int			  part_num;
	int			  error;

	lb_size = udf_rw32(udf_log_vol->log_vol->lb_size);

	wr_callback.function = udf_sync_tables_callback;
	SLIST_FOREACH(part_mapping, &udf_log_vol->part_mappings, next_mapping) {
		part_num = part_mapping->udf_virt_part_num;
		UDF_VERBOSE_TABLES(printf("\tFor partition mapping %d->%d\n", part_num, part_mapping->udf_phys_part_num));

		/* retrieve and process unallocated- and freed-space information for all used partitions of the logvol */
		error = udf_logvol_vpart_to_partition(udf_log_vol, part_num, NULL, &udf_partition);
		assert(udf_partition);

		part_hdr_desc = (struct part_hdr_desc *) udf_partition->partition->contents_use;
		part_start    = udf_rw32(udf_partition->partition->start_loc);
		part_len      = udf_rw32(udf_partition->partition->part_len);
		dscr_ver      = udf_rw16(udf_partition->partition->tag.descriptor_ver);

		sector   = udf_rw32(part_hdr_desc->unalloc_space_table.lb_num);
		length   = udf_rw32(part_hdr_desc->unalloc_space_table.len);		/* needed? */
		if (length) {
			printf("UDF: Can't write space tables yet\n");
#if 0
			error = udf_read_logvol_descriptor(udf_log_vol, part_num, sector, "Unallocated space table", &dscrptr, NULL);
			UDF_VERBOSE_MAX(printf("\tUnalloced space table\n"));
			UDF_VERBOSE_MAX(udf_dump_descriptor(dscrptr));
			//udf_process_space_table(&udf_partition->unalloc_space, dscrptr);
			free(dscrptr);
#endif
		};

		sector   = udf_rw32(part_hdr_desc->unalloc_space_bitmap.lb_num);
		length   = udf_rw32(part_hdr_desc->unalloc_space_bitmap.len);
		if (length) {
			/* read it in and modify */
			error = udf_create_empty_space_bitmap(dscr_ver, part_len, (struct space_bitmap_desc **) &dscrptr);
			udf_sync_space_bitmap(&udf_partition->unalloc_space_queue, &dscrptr->sbd, lb_size);
			UDF_VERBOSE_MAX(printf("\tWriteout unallocated space bitmap\n"));
			UDF_VERBOSE_MAX(udf_validate_tag_sums((union dscrptr *) dscrptr); udf_dump_descriptor(dscrptr));
			udf_write_partition_descriptor(udf_partition, sector, "Unallocated space bitmap", dscrptr, length, &wr_callback);	/* SESSION descriptor!! */
			free(dscrptr);
		};

		sector    = udf_rw32(part_hdr_desc->freed_space_table.lb_num);
		length    = udf_rw32(part_hdr_desc->freed_space_table.len);
		if (length) {
			printf("UDF: Can't write space tables yet\n");
#if 0
			error = udf_read_logvol_descriptor(udf_log_vol, part_num, sector, "Freed space table", &dscrptr, NULL);
			UDF_VERBOSE_MAX(printf("\tFreed space table\n"));
			UDF_VERBOSE_MAX(udf_dump_descriptor(dscrptr));
			//udf_process_space_table(&udf_partition->freed_space, dscrptr);
			free(dscrptr);
#endif
		};

		sector    = udf_rw32(part_hdr_desc->freed_space_bitmap.lb_num);
		length    = udf_rw32(part_hdr_desc->freed_space_bitmap.len);
		if (length) {
			/* read it in and modify */
			error = udf_create_empty_space_bitmap(dscr_ver, part_len, (struct space_bitmap_desc **) &dscrptr);
			udf_sync_space_bitmap(&udf_partition->freed_space_queue, &dscrptr->sbd, lb_size);
			UDF_VERBOSE_MAX(printf("\tWriteout freed space bitmap\n"));
			UDF_VERBOSE_MAX(udf_validate_tag_sums((union dscrptr *) dscrptr); udf_dump_descriptor(dscrptr));
			udf_write_partition_descriptor(udf_partition, sector, "Freed space bitmap", dscrptr, length, &wr_callback);	/* SESSION descriptor!! */
			free(dscrptr);
		};
	};
	UDF_VERBOSE_TABLES(printf("\n"));


	return 0;
}


/* mark the logical volume `open'; for non-rewritables (CD-R/DVD+R/DVD-R) this is allmost a no-op */
int udf_open_logvol(struct udf_log_vol *udf_log_vol) {
	if (udf_log_vol->logvol_state == UDF_INTEGRITY_OPEN) return 0;

	DEBUG(printf("\nOpening logical volume\n"));
	if (!udf_log_vol->primary->udf_session->disc->sequential) {
		/*
		 * TODO non sequentials have a logicaL volume integrity descriptor
		 * that needs to be marked open
		 */
	} else {
		/* sequential recordable; any write just opens it; the descriptor is allready marked open */
	};

	/* mark it open */
	/* TODO on non recordables, the open integrity ought to be written to disc ! */
	udf_log_vol->logvol_state = UDF_INTEGRITY_OPEN;

	return 0;
}


/* mark the logical volume in a `closed' state; close the integrity when possible for recordables writeout VAT */
int udf_close_logvol(struct udf_log_vol *udf_log_vol) {
	struct logvol_int_desc  *intdesc;
	struct udf_logvol_info  *impl;
	struct udf_session      *session;
	struct udf_partition    *udf_partition;
	struct udf_part_mapping *part_mapping;
	struct udf_wrcallback    wr_callback;
	uint32_t                 sector, part_num, *free_space_pos, *size_pos, len;

	if (udf_log_vol->logvol_state == UDF_INTEGRITY_CLOSED) return 0;

	/* TODO: callback for LVID writeout */
	wr_callback.function = NULL;

	DEBUG(printf("\nClosing logical volume\n"));
	if (udf_log_vol->primary->udf_session->disc->sequential) {
		/* TODO: writeout virtual sectors, compile VAT and write out VAT */
	} else {
		/* TODO: search for a place to record the `new' descriptor in */
		/* for now, allways record over the current descriptor */
		sector  = udf_rw32(udf_log_vol->log_vol->integrity_seq_id.loc);
		session = udf_log_vol->primary->udf_session;

		/* create a new `fresh' logvol integrity */
		intdesc = calloc(1, UDF_SECTOR_SIZE);
		assert(intdesc);

		udf_init_desc_tag(&intdesc->tag, TAGID_LOGVOL_INTEGRITY, udf_rw16(udf_log_vol->log_vol->tag.descriptor_ver), udf_log_vol->integrity_serial);

		udf_set_timestamp_now(&intdesc->time);
		intdesc->integrity_type = udf_rw32(UDF_INTEGRITY_CLOSED);
		/* FIXME next extent? */
		intdesc->logvol_content_use.logvolhdr.next_unique_id = udf_rw64(udf_log_vol->next_unique_id);

		/* fill in free space */
		intdesc->num_part = udf_rw32(udf_log_vol->num_part_mappings);
		free_space_pos = intdesc->tables;
		size_pos       = intdesc->tables + udf_rw32(intdesc->num_part);
		SLIST_FOREACH(part_mapping, &udf_log_vol->part_mappings, next_mapping) {
			part_num = part_mapping->udf_virt_part_num;
			udf_logvol_vpart_to_partition(udf_log_vol, part_num, NULL, &udf_partition);
			assert(udf_partition);

			*size_pos++       = udf_rw32(udf_partition->partition->part_len);
			*free_space_pos++ = udf_rw32(udf_partition->free_unalloc_space / udf_rw32(udf_log_vol->log_vol->lb_size));
			printf("Freespace left on physical partition %d : %"PRIu64" (%"PRIu64" pages)\n", part_num,
					udf_partition->free_unalloc_space, udf_partition->free_unalloc_space / udf_rw32(udf_log_vol->log_vol->lb_size));
		};

		/* fill in UDF implementation use parameters */
		impl = (struct udf_logvol_info *) (intdesc->tables + 2*udf_log_vol->num_part_mappings);
		udf_set_imp_id(&impl->impl_id);
		impl->num_files        = udf_rw32(udf_log_vol->num_files);
		impl->num_directories  = udf_rw32(udf_log_vol->num_directories);
		impl->min_udf_readver  = udf_rw16(udf_log_vol->min_udf_readver);
		impl->min_udf_writever = udf_rw16(udf_log_vol->min_udf_writever);
		impl->max_udf_writever = udf_rw16(udf_log_vol->max_udf_writever);

		intdesc->l_iu = udf_rw32(sizeof(struct udf_logvol_info));		/* ECMA 3/10.10.7, UDF 2.2.6.4. */
		len  = sizeof(struct logvol_int_desc) - sizeof(uint32_t);		/* length of logvol_int_desc without the extra table entry */
		len += sizeof(uint32_t) * 2 * udf_log_vol->num_part_mappings;		/* size and free space */
		len += sizeof(struct udf_logvol_info);					/* extra implementation use area */
		len -= UDF_DESC_TAG_LENGTH;						/* without header */
		intdesc->tag.desc_crc_len = udf_rw16(len);

		udf_write_session_descriptor(session, sector, "Logvol integrity descriptor (LVID)", (union dscrptr *) intdesc, UDF_SECTOR_SIZE, &wr_callback);

		free(intdesc);
	};
	udf_writequeue_kick(udf_log_vol->primary->udf_session, "writeout LVID");
	sleep(1);

	/* mark it closed */
	udf_log_vol->logvol_state = UDF_INTEGRITY_CLOSED;
	return 0;
}


int udf_sync_logvol(struct udf_log_vol *udf_log_vol) {
	struct udf_mountpoint *udf_mountpoint;
	struct udf_node	      *udf_node;
	int error;

	/* sync all nodes */
	SLIST_FOREACH(udf_mountpoint, &udf_log_vol->mountpoints, logvol_next) {
		STAILQ_FOREACH(udf_node, &udf_mountpoint->dirty_nodes, next_dirty) {
			//printf("sync/writeout node %p\n", udf_node);
			udf_writeout_udf_node(udf_node, "Sync Disc");
		};
	};
	/* sync minicaches */
	udf_sync_caches(udf_log_vol);

	/* sync free and used space tables for writable volsets */
	UDF_VERBOSE(udf_dump_volume_name("Sync used/freed space tables for logical volume", udf_log_vol));
	error = udf_sync_space_tables(udf_log_vol);

	return error;
}


/* convenience routine */
int udf_sync_disc(struct udf_discinfo *disc) {
	struct udf_volumeset	*udf_volumeset;
	struct udf_pri_vol	*udf_pri_vol;
	struct udf_log_vol	*udf_log_vol;

	SLIST_FOREACH(udf_volumeset, &udf_volumeset_list, next_volumeset) {
		if (!udf_volumeset->obsolete) {
			STAILQ_FOREACH(udf_pri_vol, &udf_volumeset->primaries, next_primary) {
				if (udf_pri_vol->udf_session->disc == disc) {
					SLIST_FOREACH(udf_log_vol, &udf_pri_vol->log_vols, next_logvol) {
						udf_sync_logvol(udf_log_vol);
					}; /* logical */
				}; /* disc */
			}; /* primary */
		}; /* if */
	}; /* volumeset */

	return 0;
}


/******************************************************************************************
 *
 * UDF descriptor buildup and update functions
 *
 ******************************************************************************************/

static void udf_init_desc_tag(struct desc_tag *tag, uint16_t id, uint16_t dscr_ver, uint16_t serial_num) {
	bzero(tag, sizeof(struct desc_tag));
	tag->id			= udf_rw16(id);
	tag->descriptor_ver	= udf_rw16(dscr_ver);
	tag->serial_num		= udf_rw16(serial_num);
	/* the rest gets filled in when we write */
}


static void udf_osta_charset(struct charspec *charspec) {
	bzero(charspec, sizeof(struct charspec));
	charspec->type = 0;
	strcpy(charspec->inf, "OSTA Compressed Unicode");
}


static void udf_encode_osta_id(char *osta_id, uint16_t len, char *text) {
	uint16_t  u16_name[1024];
	uint8_t  *pos;
	uint16_t *pos16;

	bzero(osta_id, len);
	if (!text) return;

	bzero(u16_name, sizeof(uint16_t) * 1023);
	/* convert ascii to 16 bits unicode */
	pos = text; pos16 = u16_name;
	while (*pos) {
		*pos16 = *pos;
		pos++; pos16++;
	};
	*pos16 = 0;

	udf_CompressUnicode(len, 8, (unicode_t *) u16_name, (byte *) osta_id);

	/* Ecma 167/7.2.13 states that the length is recorded in the last byte */
	osta_id[len-1] = strlen(text)+1;
}


static void udf_set_app_id(struct regid *regid) {
	bzero(regid, sizeof(struct regid));
	regid->flags	= 0;						/* not dirty and not protected */
	strcpy(regid->id, APP_NAME);
	regid->id_suffix[0] = APP_VERSION_MAIN;
	regid->id_suffix[1] = APP_VERSION_SUB;
}


static void udf_set_imp_id(struct regid *regid) {
	bzero(regid, sizeof(struct regid));
	regid->flags	= 0;						/* not dirty and not protected */
	strcpy(regid->id, IMPL_NAME);
	regid->id_suffix[0] = 4;	/* unix */
	regid->id_suffix[1] = 0;	/* generic */
#if   defined(__NetBSD__)
	regid->id_suffix[1] = 8;	/* NetBSD */
#elif defined(__FreeBSD__)
	regid->id_suffix[1] = 7;	/* FreeBSD */
#elif defined(LINUX)
	regid->id_suffix[1] = 5;	/* Linux */
#endif
}


static void udf_set_entity_id(struct regid *regid, char *name, uint16_t UDF_version) {
	uint16_t *ver;

	bzero(regid, sizeof(struct regid));
	regid->flags    = 0;						/* not dirty and not protected */
	strcpy(regid->id, name);
	ver  = (uint16_t *) regid->id_suffix;
	*ver = udf_rw16(UDF_version);
	regid->id_suffix[2] = 4;	/* unix */
	regid->id_suffix[3] = 0;	/* generic */
#if   defined(__NetBSD__)
	regid->id_suffix[3] = 8;	/* NetBSD */
#elif defined(__FreeBSD__)
	regid->id_suffix[3] = 7;	/* FreeBSD */
#elif defined(LINUX)
	regid->id_suffix[3] = 5;	/* Linux */
#endif
}


void udf_set_contents_id(struct regid *regid, char *content_id) {
	bzero(regid, sizeof(struct regid));
	regid->flags    = 0;
	strcpy(regid->id, content_id);
}


/* checkout gettimeofday() and friends; problem is conversion to date/year etc. */
static void udf_set_timestamp_now(struct timestamp *timestamp) {
	struct tm *tm;
	time_t     now;

	bzero(timestamp, sizeof(struct timestamp));
	now = time(NULL);
	tm = gmtime(&now);
	timestamp->type_tz	= udf_rw16((1<<12) + 0);	/* has to be method 1 ? ECMA 1/7.3, UDF 2., 2.1.4.1, 3.1.1. */
	timestamp->year		= udf_rw16(tm->tm_year + 1900);
	timestamp->month	= tm->tm_mon + 1;		/* `tm' structure uses 0..11 for months */
	timestamp->day		= tm->tm_mday;
	timestamp->hour		= tm->tm_hour;
	timestamp->minute	= tm->tm_min;
	timestamp->second	= tm->tm_sec;
	/* FIXME use gettimeofday() for more acurate time */
}


/* XXX creators of empty descriptors could be externalised */

/*
 * result can be further processed using modify functions if demanded and then
 * processed trough udf_proc_pri_vol
 * [ int udf_proc_pri_vol(struct udf_session *udf_session, struct udf_pri_vol **current, struct pri_vol_desc *incomming); ]
 *
 */

int udf_create_empty_primary_volume_descriptor(uint16_t dscr_ver, uint16_t serial, char *volset_id, char *privol_name, int vds_num, int max_vol_seq, struct pri_vol_desc **dscrptr) {
	struct pri_vol_desc *dscr;

	assert(dscrptr);
	*dscrptr = NULL;

	/* allocate and populate an empty primary volume descriptor */
	dscr = malloc(UDF_SECTOR_SIZE);
	if (!dscr) return ENOMEM;
	bzero(dscr, UDF_SECTOR_SIZE);

	udf_init_desc_tag(&dscr->tag, TAGID_PRI_VOL, dscr_ver, 1);
	dscr->pvd_num		= udf_rw32(serial);
	udf_encode_osta_id(dscr->vol_id, 32, privol_name);
	dscr->vds_num		= udf_rw16(vds_num);
	dscr->max_vol_seq	= udf_rw16(max_vol_seq);
	if (max_vol_seq > 1) {
		dscr->ichg_lvl		= udf_rw16(3);			/* signal its a single volume intended to be in a set */
		dscr->max_ichg_lvl	= udf_rw16(3);			/* ,, */
		dscr->flags		= udf_rw16(1);			/* signal relevance volumeset id */
	} else {
		dscr->ichg_lvl		= udf_rw16(2);			/* signal its volume intended not to be in a set */
		dscr->max_ichg_lvl	= udf_rw16(2);			/* ,, */
		dscr->flags		= udf_rw16(0);			/* signal relevance volumeset id */
	};

	dscr->charset_list		= udf_rw16(1);			/* only CS0 */
	dscr->max_charset_list		= udf_rw16(1);
	udf_encode_osta_id(dscr->volset_id, 128, volset_id);
	udf_osta_charset(&dscr->desc_charset);
	udf_osta_charset(&dscr->explanatory_charset);
	udf_set_app_id(&dscr->app_id);
	udf_set_imp_id(&dscr->imp_id);
	udf_set_timestamp_now(&dscr->time);

	dscr->tag.desc_crc_len = udf_rw16(sizeof(struct pri_vol_desc) - UDF_DESC_TAG_LENGTH);

	*dscrptr = dscr;
	return 0;
}


int udf_create_empty_partition_descriptor(uint16_t dscr_ver, uint16_t serial, uint16_t part_num, uint32_t access_type, uint32_t start_loc, uint32_t part_len, uint32_t space_bitmap_size, uint32_t unalloc_space_bitmap, struct part_desc **dscrptr) {
	struct part_desc     *dscr;
	struct part_hdr_desc *part_hdr;

	assert(dscrptr);
	*dscrptr = NULL;

	/* allocate and populate empty partition descriptor */
	dscr = malloc(UDF_SECTOR_SIZE);					/* only descriptor, no bitmap! */
	if (!dscr) return ENOMEM;
	bzero(dscr, UDF_SECTOR_SIZE);

	udf_init_desc_tag(&dscr->tag, TAGID_PARTITION, dscr_ver, 1);
	dscr->seq_num  = udf_rw32(serial);
	dscr->flags    = udf_rw16(1);					/* bit 0 : space is allocated */
	dscr->part_num = udf_rw16(part_num);

	if (dscr_ver == 2) udf_set_contents_id(&dscr->contents, "+NSR02");
	if (dscr_ver == 3) udf_set_contents_id(&dscr->contents, "+NSR03");
	part_hdr = (struct part_hdr_desc *) dscr->contents_use;
	part_hdr->unalloc_space_bitmap.len    = udf_rw32(space_bitmap_size);
	part_hdr->unalloc_space_bitmap.lb_num = udf_rw32(unalloc_space_bitmap);

	dscr->access_type = udf_rw32(access_type);
	dscr->start_loc   = udf_rw32(start_loc);
	dscr->part_len    = udf_rw32(part_len);

	udf_set_imp_id(&dscr->imp_id);					/* why is this ignored? */

	dscr->tag.desc_crc_len = udf_rw16(sizeof(struct part_desc) - UDF_DESC_TAG_LENGTH);

	*dscrptr = dscr;
	return 0;
}


int udf_create_empty_unallocated_space_descriptor(uint16_t dscr_ver, uint16_t serial, struct unalloc_sp_desc **dscrptr) {
	struct unalloc_sp_desc *dscr;

	assert(dscrptr);
	*dscrptr = NULL;

	/* allocate and populate an empty unallocated space descriptor */
	dscr = malloc(UDF_SECTOR_SIZE);
	if (!dscr) return ENOMEM;
	bzero(dscr, UDF_SECTOR_SIZE);

	udf_init_desc_tag(&dscr->tag, TAGID_UNALLOC_SPACE, dscr_ver, 1);
	dscr->seq_num		= udf_rw32(serial);
	dscr->tag.desc_crc_len	= udf_rw16(sizeof(struct unalloc_sp_desc) - sizeof(struct extent_ad) - UDF_DESC_TAG_LENGTH);

	*dscrptr = dscr;

	return 0;
}


int udf_create_empty_implementation_use_volume_descriptor(uint16_t dscr_ver, uint16_t serial, char *logvol_name, struct impvol_desc **dscrptr) {
	struct impvol_desc *dscr;
	struct udf_lv_info *lv_info;

	assert(dscrptr);
	*dscrptr = NULL;

	/* allocate and populate an empty implementation use volume descriptor */
	dscr = malloc(UDF_SECTOR_SIZE);
	if (!dscr) return ENOMEM;
	bzero(dscr, UDF_SECTOR_SIZE);

	udf_init_desc_tag(&dscr->tag, TAGID_IMP_VOL, dscr_ver, 1);
	dscr->seq_num		= udf_rw32(serial);
	udf_set_entity_id(&dscr->impl_id, "*UDF LV Info", 0x102);	/* just pick one; it'll be modifed later */

	lv_info = &dscr->_impl_use.lv_info;
	udf_osta_charset(&lv_info->lvi_charset);
	udf_encode_osta_id(lv_info->logvol_id, 128, logvol_name);
	udf_encode_osta_id(lv_info->lvinfo1, 36, NULL);
	udf_encode_osta_id(lv_info->lvinfo2, 36, NULL);
	udf_encode_osta_id(lv_info->lvinfo3, 36, NULL);
	udf_set_imp_id(&lv_info->impl_id);

	dscr->tag.desc_crc_len	= udf_rw16(sizeof(struct impvol_desc) - UDF_DESC_TAG_LENGTH);

	*dscrptr = dscr;

	return 0;
}


int udf_create_empty_logical_volume_descriptor(uint16_t dscr_ver, uint16_t serial, char *logvol_name, uint32_t lb_size, uint32_t integrity_start, uint32_t integrity_length, struct logvol_desc **dscrptr) {
	struct logvol_desc *dscr;

	assert(dscrptr);
	*dscrptr = NULL;

	/* allocate and populate an empty logical volume descriptor */
	dscr = malloc(UDF_SECTOR_SIZE);
	if (!dscr) return ENOMEM;
	bzero(dscr, UDF_SECTOR_SIZE);

	udf_init_desc_tag(&dscr->tag, TAGID_LOGVOL, dscr_ver, 1);
	dscr->seq_num		= udf_rw32(serial);
	udf_osta_charset(&dscr->desc_charset);
	udf_encode_osta_id(dscr->logvol_id, 128, logvol_name);
	dscr->lb_size		= udf_rw32(lb_size);
	udf_set_contents_id(&dscr->domain_id, "*OSTA UDF Compliant");

	/* no fsd yet nor partition mapping */
	udf_set_imp_id(&dscr->imp_id);
	dscr->integrity_seq_id.loc = udf_rw32(integrity_start);
	dscr->integrity_seq_id.len = udf_rw32(integrity_length);

	dscr->tag.desc_crc_len = udf_rw16(sizeof(struct logvol_desc) - 1 - UDF_DESC_TAG_LENGTH);

	*dscrptr = dscr;
	return 0;
}


int udf_create_empty_space_bitmap(uint16_t dscr_ver, uint32_t part_length, struct space_bitmap_desc **dscrptr) {
	struct space_bitmap_desc *dscr;
	uint64_t bits;
	uint32_t bytes, space_bitmap_size;

	assert(dscrptr);
	*dscrptr = NULL;

	/* reserve space for unallocated space bitmap */
	bits  = part_length;
	bytes = (bits + 7)/8;
	space_bitmap_size = (bytes + sizeof(struct space_bitmap_desc)-1);

	/* round space bitmap size to sector size */
	space_bitmap_size = ((space_bitmap_size + UDF_SECTOR_SIZE - 1) / UDF_SECTOR_SIZE) * UDF_SECTOR_SIZE;

	/* allocate and populate an empty space bitmap descriptor */
	dscr = malloc(space_bitmap_size);
	if (!dscr) return ENOMEM;
	bzero(dscr, space_bitmap_size);

	udf_init_desc_tag(&dscr->tag, TAGID_SPACE_BITMAP, dscr_ver, 1);
	dscr->num_bits  = udf_rw32(bits);
	dscr->num_bytes = udf_rw32(bytes);

	*dscrptr = dscr;
	return 0;
}


/* FIXME: no rootdir setting yet */
int udf_create_empty_fileset_desc(uint16_t dscr_ver, uint32_t fileset_num, char *logvol_name, char *fileset_name, struct fileset_desc **dscrptr) {
	struct fileset_desc *dscr;

	assert(dscrptr);
	*dscrptr = NULL;

	/* allocate and populate an empty logical volume descriptor */
	dscr = malloc(UDF_SECTOR_SIZE);
	if (!dscr) return ENOMEM;
	bzero(dscr, UDF_SECTOR_SIZE);

	udf_init_desc_tag(&dscr->tag, TAGID_FSD, dscr_ver, 1);
	udf_set_timestamp_now(&dscr->time);
	dscr->ichg_lvl         = udf_rw16(3);	/* fixed? */
	dscr->max_ichg_lvl     = udf_rw16(3);	/* fixed? */
	dscr->charset_list     = udf_rw32(1);	/* only CS0 */
	dscr->max_charset_list = udf_rw32(1);	/* only CS0 */
	dscr->fileset_num      = udf_rw32(fileset_num);	/* key for fileset */
	dscr->fileset_desc_num = udf_rw32(0);		/* fileset descriptor number as in copy # */

	udf_osta_charset(&dscr->logvol_id_charset);
	udf_encode_osta_id(dscr->logvol_id, 128, logvol_name);

	udf_osta_charset(&dscr->fileset_charset);
	udf_encode_osta_id(dscr->fileset_id, 32, fileset_name);

	udf_encode_osta_id(dscr->copyright_file_id, 32, NULL);
	udf_encode_osta_id(dscr->abstract_file_id,  32, NULL);

	udf_set_contents_id(&dscr->domain_id, "*OSTA UDF Compliant");

	dscr->tag.desc_crc_len = udf_rw16(sizeof(struct fileset_desc) - UDF_DESC_TAG_LENGTH);

	*dscrptr = dscr;
	return 0;
}


int udf_create_empty_anchor_volume_descriptor(uint16_t dscr_ver, uint32_t main_vds_loc, uint32_t reserve_vds_loc, uint32_t length, struct anchor_vdp **vdp) {
	assert(vdp);
	assert(abs(main_vds_loc - reserve_vds_loc) >= length);

	*vdp = malloc(UDF_SECTOR_SIZE);						/* it needs to be 2028, but the crc len is fixed size 512 in Ecma/UDF? */
	if (!*vdp) return ENOMEM;
	bzero(*vdp, UDF_SECTOR_SIZE);

	udf_init_desc_tag(&(*vdp)->tag, TAGID_ANCHOR, dscr_ver, 1);
	(*vdp)->main_vds_ex.loc    = udf_rw32(main_vds_loc);
	(*vdp)->main_vds_ex.len    = udf_rw32(length * UDF_SECTOR_SIZE);
	(*vdp)->reserve_vds_ex.loc = udf_rw32(reserve_vds_loc);
	(*vdp)->reserve_vds_ex.len = udf_rw32(length * UDF_SECTOR_SIZE);

	(*vdp)->tag.desc_crc_len = udf_rw16(512-UDF_DESC_TAG_LENGTH);		/* fixed size in Ecma */
	return 0;
}


int udf_create_empty_terminator_descriptor(uint16_t dscr_ver, struct desc_tag **tag) {
	assert(tag);

	*tag = malloc(UDF_SECTOR_SIZE);
	if (!*tag) return ENOMEM;
	bzero(*tag, UDF_SECTOR_SIZE);

	udf_init_desc_tag(*tag, TAGID_TERM, dscr_ver, 1);

	(*tag)->desc_crc_len = udf_rw16(512-UDF_DESC_TAG_LENGTH);		/* fixed size in Ecma */
	return 0;
}


/******************************************************************************************
 *
 * Basic `open' and `close' disc functions
 *
 ******************************************************************************************/


static void udf_process_session_range(struct udf_discinfo *disc, int *enabled, int low, int high) {
	int session;

	if (!disc) return;

	high = MIN(high, disc->num_sessions-1);
	session = low;

	for (session = low; session <= high; session++) {
		enabled[session] = 1;
	};
}


/* range is specified in -3,5,7 or 5-6,8- etc */
static int udf_process_session_range_string(struct udf_discinfo *disc, char *range) {
	struct udf_session *udf_session, *next_udf_session, *last_ok;
	char *pos, *nop;
	int low, high, len, session;
	int enabled[MAX_SESSIONS];

	if (!range) return 0;
	DEBUG(printf("UDF range debugging string '%s'\n", range));

	if (disc) {
		/* disable all */
		for (session = 0; session < disc->num_sessions; session++) {
			enabled[session] = 0;
		};
	};

	/* parse string */
	nop = strdup(range);
	pos = range;
	if (sscanf(pos, "-%u%n%s", &high, &len, nop) >= 1) {
		DEBUG(printf("UDF range match till %d\n", high));
		udf_process_session_range(disc, enabled, 0, high);
		pos += len;
	};
	if (*pos && *pos == ',') pos++;
	while (*pos) {
		if (sscanf(pos, "%u%n%s", &low, &len, nop) >= 1) {
			pos += len;
			if (*pos == '-') {
				pos++;
				if (!*pos) {
					DEBUG(printf("UDF range match from %d\n", low));
					udf_process_session_range(disc, enabled, low, INT_MAX);
					free(nop);
					return 0;
				};
				if (sscanf(pos, "%u%n%s", &high, &len, nop) >= 1) {
					pos += len;
					DEBUG(printf("UDF range match from %d to %d\n", low, high));
					udf_process_session_range(disc, enabled, low, high);
				};
			} else {
				if (!*pos || (*pos == ',')) {
					DEBUG(printf("UDF range match %d\n", low));
					udf_process_session_range(disc, enabled, low, low);
				};
			};
			if (*pos && (*pos != ',')) {
				fprintf(stderr, "UDF range matching : ',' expected at %s\n", pos);
				free(nop);
				return ENOENT;
			};
			pos++;
		} else {
			fprintf(stderr, "UDF range matching : number expected at %s\n", pos);
			free(nop);
			return ENOENT;
		};
	};
	free(nop);

	DEBUG(printf("UDF range matching : all ok till the end\n"));
	if (!disc) return 0;

	last_ok = NULL;
	udf_session = STAILQ_FIRST(&disc->sessions);
	while (udf_session) {
		next_udf_session = STAILQ_NEXT(udf_session, next_session);
		session = udf_session->session_num;
		if (!enabled[session]) {
			/* remove this session */
			fprintf(stderr, "UDF: disabling UDF session %d on request\n", session);
			STAILQ_REMOVE(&disc->sessions, udf_session, udf_session, next_session);
			free(udf_session);

			disc->session_is_UDF[session] = 0;
		};
		udf_session = next_udf_session;
	};

	return 0;
}


int udf_check_session_range(char *range) {
	return udf_process_session_range_string(NULL, range);
}


int udf_mount_disc(char *devname, char *range, struct udf_discinfo **disc) {
	int error;

	error = udf_open_disc(devname, disc);
	if (error) return error;

	error = udf_get_anchors(*disc);
	UDF_VERBOSE(udf_dump_disc_anchors(*disc));

	if (range) {
		UDF_VERBOSE(printf("Selecting UDF sessions '%s' as specified\n", range));
		udf_process_session_range_string(*disc, range);
		UDF_VERBOSE(udf_dump_disc_anchors(*disc));
	};

	/* no UDF partitions so bail out */
	if ((*disc)->num_udf_sessions == 0) return 0;

	error = udf_get_volumeset_space(*disc);
	if (error) return error;

	UDF_VERBOSE(printf("Finding volumesets dependences\n"));
	udf_eliminate_predescessor_volumesets(*disc);

	UDF_VERBOSE_TABLES(udf_dump_alive_sets());

	UDF_VERBOSE(printf("Retrieving logical volumes their VAT's/Sparing tables\n"));
	error = udf_get_logical_volumes_supporting_tables(*disc);

	UDF_VERBOSE_TABLES(udf_dump_alive_sets());

	/* insert disc in the disc list */
	SLIST_INSERT_HEAD(&udf_discs_list, *disc, next_disc);

	return error;
}


int udf_dismount_disc(struct udf_discinfo *disc) {
	struct udf_volumeset	*udf_volumeset;
	struct udf_pri_vol	*udf_pri_vol;
	struct udf_log_vol	*udf_log_vol;

	if (!disc->recordable) {
		udf_close_disc(disc);
		return 0;
	};

	SLIST_FOREACH(udf_volumeset, &udf_volumeset_list, next_volumeset) {
		if (!udf_volumeset->obsolete) {
			STAILQ_FOREACH(udf_pri_vol, &udf_volumeset->primaries, next_primary) {
				if (udf_pri_vol->udf_session->disc == disc) {
					SLIST_FOREACH(udf_log_vol, &udf_pri_vol->log_vols, next_logvol) {
						udf_sync_logvol(udf_log_vol);

						udf_close_logvol(udf_log_vol);
					}; /* logical */
				}; /* disc */
			}; /* primary */
		}; /* if */
	}; /* volumeset */
	udf_sync_disc(disc);

	/* stop threads and finish writing to it */
	udf_stop_disc_thread(disc);
	udf_discinfo_finish_writing(disc);
	udf_close_disc(disc);

	return 0;
}


/******************************************************************************************
 *
 * Directory and other conversion UDF logic
 * Move to udf_unix.c / udf_vnops.c one day?
 *
 ******************************************************************************************/

static int udf_translate_icb_filetype_to_dirent_filetype(int udf_filetype) {
	int d_type;


	switch (udf_filetype) {
		case UDF_ICB_FILETYPE_DIRECTORY :
			d_type = DT_DIR;
			break;
		case UDF_ICB_FILETYPE_FIFO :
			d_type = DT_FIFO;
			break;
		case UDF_ICB_FILETYPE_CHARDEVICE :
			d_type = DT_CHR;
			break;
		case UDF_ICB_FILETYPE_BLOCKDEVICE :
			d_type = DT_BLK;
			break;
		case UDF_ICB_FILETYPE_RANDOMACCESS :
			d_type = DT_REG;
			break;
		case UDF_ICB_FILETYPE_SYMLINK :
			d_type = DT_LNK;
			break;
		case UDF_ICB_FILETYPE_SOCKET :
			d_type = DT_SOCK;
			break;
		default :
			d_type = DT_UNKNOWN;
			break;
	};
	return d_type;
}


/* VOP_GETATTR */
/* allmost NOOP since we remember the stat in the inode */
int udf_getattr(struct udf_node *udf_node, struct stat *stat) {
	*stat = udf_node->stat;
	return 0;
}


/* VOP_SETATTR */
/* allmost NOOP since we remember the stat in the inode */
/* note VOP_SETATTR can selectively set attrs		*/
int udf_setattr(struct udf_node *udf_node, struct stat *stat) {
	/* FIXME please dont just copy everything ... XXX */
	udf_node->stat = *stat;
	udf_node_mark_dirty(udf_node);
	return 0;
}


/* VOP_READDIR */
/* read in dirent's until the result_uio can't hold another */
int udf_readdir(struct udf_node *dir_node, struct uio *result_uio, int *eof_res /* int *cookies, int ncookies */) {
	struct fileid_desc *fid;
	struct uio     dir_uio;
	struct iovec   dir_iovec;
	struct dirent  entry;
	uint8_t       *buffer, *fid_name;
	uint32_t       pos, dir_part_offset, dir_part_size, entry_length;
	int            eof, enough, is_first;
	int            error;

	if (!dir_node) return EINVAL;

	buffer = malloc(DIRREAD_BUFFER_SIZE);
	if (!buffer) return ENOMEM;

	bzero(&dir_uio, sizeof(struct uio));

	/* initialise directory reading uio structure and start at the start of the directory */
	dir_uio.uio_rw        = UIO_WRITE;
	pos = dir_part_offset = result_uio->uio_offset; /* we remember our offset in the uio... is this sane? */

	is_first = (result_uio->uio_offset == 0);

	eof = (pos == dir_node->stat.st_size);
	while (!eof) {
		/* read in piece of directory */
		dir_iovec.iov_base  = buffer;
		dir_iovec.iov_len   = DIRREAD_BUFFER_SIZE;
		dir_uio.uio_iovcnt  = 1;
		dir_uio.uio_iov     = &dir_iovec;

		/* from offset `pos' for at most buffer size */
		dir_uio.uio_offset  = dir_part_offset = pos;
		dir_part_size       = MIN(DIRREAD_BUFFER_SIZE, dir_node->stat.st_size - pos);
		dir_uio.uio_resid   = dir_part_size;
		error = udf_read_file_part_uio(dir_node, "directory part", UDF_CACHEFLAG_METADATA, &dir_uio);
		if (error) {
			printf("Error while reading directory index file");
			free(buffer);
			return EIO;
		};

		/* process this chunk */
		do {
			eof = (pos == dir_node->stat.st_size);
			if (eof) break;

			/* check if there is enough room for another descriptor */
			if (result_uio->uio_resid <= sizeof(struct dirent)) {
				/* ask me again */
				free(buffer);
				result_uio->uio_offset = pos;	/* directory offset to call back with */
				if (eof_res) *eof_res = eof;
				return 0;
			};

			/* insert `.' entry when reading in first entry */
			if (is_first) {
				bzero(&entry, sizeof(struct dirent));
				strcpy(entry.d_name, ".");
				entry.d_type = DT_DIR;
				entry_length = 0;
				is_first = 0;
				enough = 1;
				uiomove(&entry, sizeof(struct dirent), result_uio);
				continue;
			};

			enough = (pos - dir_part_offset + UDF_FID_SIZE <= dir_part_size);
			if (!enough) break;

			fid = (struct fileid_desc *) (buffer + (pos - dir_part_offset));
			fid_name = fid->data + udf_rw16(fid->l_iu);
			entry_length = udf_calc_tag_malloc_size((union dscrptr *) fid);

			enough = (pos - dir_part_offset + entry_length <= dir_part_size);
			if (!enough) break;

			UDF_VERBOSE_MAX(udf_dump_descriptor((union dscrptr *) fid));

			error = udf_check_tag((union dscrptr *) fid);
			if (!error) error = udf_check_tag_payload((union dscrptr *) fid);
			if (error) {
				printf("BROKEN DIRECTORY ENTRY ?\n");
				// udf_dump_desc(&fid->tag);
				// udf_dump_fileid(fid);
				free(buffer);
				return EIO;
			};

			/* we got a whole descriptor; now check if the file is deleted */
			if (!error && ((fid->file_char & UDF_FILE_CHAR_DEL) == 0)) {
				/* create resulting dirent structure */
				bzero(&entry, sizeof(struct dirent));
				entry.d_fileno = udf_calc_hash(&fid->icb);
				entry.d_reclen = sizeof(struct dirent);
				entry.d_type   = DT_UNKNOWN;
				udf_to_unix_name(entry.d_name, fid_name, fid->l_fi, &dir_node->udf_log_vol->log_vol->desc_charset);
				if (fid->file_char & UDF_FILE_CHAR_DIR) entry.d_type = DT_DIR;
				if (fid->file_char & UDF_FILE_CHAR_PAR) strcpy(entry.d_name, "..");

				/* copy out the new dirent */
				uiomove(&entry, sizeof(struct dirent), result_uio);
			};
			pos += entry_length;
		} while (enough);
		eof = (pos >= dir_node->stat.st_size);	/* sanity */
	}; /* eof */

	/* end of directory reached, feeing buffer */
	free(buffer);
	if (eof_res) *eof_res = 1;

	return 0;
}


static void udf_glue_fid_on_node(struct udf_node *udf_node, struct fileid_desc *fid, char *entry_name, uint32_t dir_offset, uint32_t entry_length) {
	struct dirent      *entry;

	/* mark place */
	udf_node->dir_offset       = dir_offset;
	udf_node->dir_entrylen     = entry_length;

	/* create dirent descriptor */
	entry = &udf_node->dirent;
	bzero(entry, sizeof(struct dirent));

	entry->d_fileno = udf_node->hashkey;		/* FIXME when changing to use unique_id as hash, change this too XXX */
	entry->d_reclen = sizeof(struct dirent);
	entry->d_type   = DT_UNKNOWN;
	strcpy(entry->d_name, entry_name);

	if (fid->file_char & UDF_FILE_CHAR_DIR) entry->d_type = DT_DIR;
	if (fid->file_char & UDF_FILE_CHAR_PAR) strcpy(entry->d_name, "..");
	entry->d_type = udf_translate_icb_filetype_to_dirent_filetype(udf_node->udf_filetype);
}


int udf_build_udf_node(struct udf_node *dir_node, struct fileid_desc *fid, uint32_t dir_offset, struct udf_node **res_sub_node) {
	struct udf_node    *sub_node;
	uint8_t            *fid_name;
	uint8_t             entry_name[MAXNAMLEN];
	uint32_t            entry_length, bucket;
	ino_t               hashkey;
	int                 error;

	hashkey  = udf_calc_hash(&fid->icb);
	entry_length = udf_calc_tag_malloc_size((union dscrptr *) fid);
	fid_name = fid->data + udf_rw16(fid->l_iu);
	udf_to_unix_name(entry_name, fid_name, fid->l_fi, &dir_node->udf_log_vol->log_vol->desc_charset);

	/* build missing vnode */
	error = udf_readin_anon_udf_node(dir_node->udf_log_vol, NULL, &fid->icb, entry_name, &sub_node);

	/* de-anonymise this UDF node */
	sub_node->mountpoint       = dir_node->mountpoint;
	sub_node->hashkey          = hashkey;
	sub_node->file_version_num = udf_rw16(fid->file_version_num);	/* user set */
	sub_node->udf_filechar     = fid->file_char;

	udf_glue_fid_on_node(sub_node, fid, entry_name, dir_offset, entry_length);

	/* insert/replace in mountpoint's udfnode hashtable with optional check for doubles */
	bucket = sub_node->hashkey & UDF_INODE_HASHMASK;
	if (0) {
		struct udf_node *lookup;
		LIST_FOREACH(lookup, &dir_node->mountpoint->udf_nodes[bucket], next_node) {
			if (lookup->hashkey == sub_node->hashkey) printf("DOUBLE hashnode?\n");
		};
	};

	LIST_INSERT_HEAD(&dir_node->mountpoint->udf_nodes[bucket], sub_node, next_node);
	DEBUG(printf("inserting hash node for hash = %d\n", (int) hashkey));

	*res_sub_node = sub_node;
	return 0;
}



/* XXX
 * speedup posibility; remember where in the dir_node you ended and start
 * from that entry on; this greatly helps lookup's of directory entries
 * XXX
 */
/* VOP_LOOKUP (alike ... needs to be slightly adjusted for kernel) */
int  udf_lookup(struct udf_node *dir_node, struct udf_node **res_node, char *name) {
	struct fileid_desc *fid;
	struct uio          dir_uio;
	struct iovec        dir_iovec;
	struct udf_node    *sub_node;
	uint8_t            *buffer, *fid_name;
	uint8_t             entry_name[MAXNAMLEN];
	uint32_t            pos, dir_part_offset, dir_part_size, entry_length, bucket;
	ino_t               hashkey;
	int                 eof, enough, isdotdot;
	int                 error;

	if (!dir_node) return EINVAL;

	/* lookup name in name cache and if not found create a new udf_node */
	assert(dir_node);
	assert(res_node);
	*res_node = NULL;

	DEBUG(printf("udf_lookup node = %p, name = %s\n", dir_node, name));
	/* only lookup in directories XXX DT_COMP also possible XXX */
	if (dir_node->dirent.d_type != DT_DIR) return ENOTDIR;

	if (strcmp(name, ".") == 0) {
		*res_node = dir_node;
		return 0;
	};

	isdotdot = (strcmp(name, "..") == 0);

	/* XXX check name-cache when its implemented XXX */

	/* pitty we need a buffer */
	buffer = malloc(DIRREAD_BUFFER_SIZE);
	if (!buffer) return ENOMEM;

	/* initialise directory reading uio structure and start at the start of the directory */
	bzero(&dir_uio, sizeof(struct uio));
	dir_uio.uio_rw        = UIO_WRITE;
	pos = dir_part_offset = 0;

	eof = (pos == dir_node->stat.st_size);
	while (!eof) {
		/* read in piece of directory */
		dir_iovec.iov_base  = buffer;
		dir_iovec.iov_len   = DIRREAD_BUFFER_SIZE;
		dir_uio.uio_iovcnt  = 1;
		dir_uio.uio_iov     = &dir_iovec;

		/* from offset `pos' for at most buffer size */
		dir_uio.uio_offset  = dir_part_offset = pos;
		dir_part_size       = MIN(DIRREAD_BUFFER_SIZE, dir_node->stat.st_size - pos);
		dir_uio.uio_resid   = dir_part_size;
		error = udf_read_file_part_uio(dir_node, "directory part", UDF_CACHEFLAG_METADATA, &dir_uio);
		if (error) {
			printf("Error while reading directory index file");
			free(buffer);
			return EIO;
		};

		/* process this chunk */
		do {
			enough = (pos - dir_part_offset + UDF_FID_SIZE <= dir_part_size);
			if (!enough) break;	/* do-while */

			fid = (struct fileid_desc *) (buffer + (pos - dir_part_offset));
			fid_name = fid->data + udf_rw16(fid->l_iu);
			entry_length = udf_calc_tag_malloc_size((union dscrptr *) fid);

			enough = (pos - dir_part_offset + entry_length <= dir_part_size);
			if (!enough) break;	/* do-while */

			/* check if the file is deleted */
			if ((fid->file_char & UDF_FILE_CHAR_DEL) == 0) {
				udf_to_unix_name(entry_name, fid_name, fid->l_fi, &dir_node->udf_log_vol->log_vol->desc_charset);
				/* check if its the name we are looking for */
				if ((isdotdot && (fid->file_char & UDF_FILE_CHAR_PAR)) || (strcmp(name, entry_name) == 0)) {
					/* check if its allready in the mountpoint's udf_node cache (inodes) */
					hashkey = udf_calc_hash(&fid->icb);
					bucket  = hashkey & UDF_INODE_HASHMASK;
					LIST_FOREACH(sub_node, &dir_node->mountpoint->udf_nodes[bucket], next_node) {
						if (sub_node->hashkey == hashkey) {
							*res_node = sub_node;
							free(buffer);
							DEBUG(printf("reusing hash node for hash = %d\n", (int) hashkey));

							/* fill in FID info in udf_node like dir_offset, dir_entrylen and dirent structure */
							udf_glue_fid_on_node(sub_node, fid, entry_name, pos, entry_length);

							/* could polulate */
							return 0;
						};
					};
					error = udf_build_udf_node(dir_node, fid, dir_part_offset + pos, &sub_node);

					/* insert vnode in namecache */
					//printf("UDF would add vnode and insert a namecache entry for %s\n", name);

					*res_node = sub_node;
					free(buffer);

					/* could populate */
					return 0;
				};
			};
			pos += entry_length;
		} while (enough);
		eof = (pos >= dir_node->stat.st_size);	/* sanity */
	}; /* eof */

	/* not found .... */
	free(buffer);

	return ENOENT;
}


/* TODO use read logvol descriptor */
static int udf_readin_fid_info(struct udf_node *dir_node, struct udf_node *udf_node, struct fileid_desc *fid, int alloc_len) {
	struct uio          uio;
	struct iovec        iovec;

	bzero(&uio, sizeof(struct uio));
	uio.uio_rw     = UIO_WRITE;	/* write into this space */
	uio.uio_iovcnt = 1;
	uio.uio_iov    = &iovec;
	iovec.iov_base = fid;
	iovec.iov_len  = alloc_len;
	uio.uio_offset = udf_node->dir_offset;
	uio.uio_resid  = udf_node->dir_entrylen;

	return udf_read_file_part_uio(dir_node, "file id" /* udf_node->dirent.d_name */, UDF_CACHEFLAG_METADATA, &uio);
}


/* TODO use write logvol descriptor */
static int udf_writeout_fid_info(struct udf_node *dir_node, struct udf_node *udf_node, struct fileid_desc *fid) {
	struct uio     uio;
	struct iovec   iovec;
	uint32_t       fid_len;
	int flags;

	fid_len = udf_calc_tag_malloc_size((union dscrptr *) fid);

	bzero(&uio, sizeof(struct uio));
	uio.uio_rw     = UIO_READ;	/* read from this space */
	uio.uio_iovcnt = 1;
	uio.uio_iov    = &iovec;
	iovec.iov_base = fid;
	iovec.iov_len  = fid_len;
	uio.uio_offset = udf_node->dir_offset;
	uio.uio_resid  = udf_node->dir_entrylen;

	assert(fid_len == udf_node->dir_entrylen);

	flags = UDF_CACHEFLAG_METADATA;
	return udf_write_file_part_uio(dir_node, "file id" /* udf_node->dirent.d_name */, flags, &uio);
}


/* search for a space to record the fid in, not checking if it is allready in it ! */
/* ALERT: not to be used to move a descriptor ... use writeout_fid_info for that   */
static int udf_insert_fid_info(struct udf_node *dir_node, struct udf_node *udf_node, struct fileid_desc *fid, uint16_t fid_len) {
	struct fileid_desc *dir_fid;
	struct uio          dir_uio;
	struct iovec        dir_iovec;
	uint8_t            *buffer, *dir_fid_name;
	uint32_t            pos, dir_part_offset, dir_part_size, entry_length;
	uint32_t            chosen_length, chosen_offset, chosen_length_diff, padding;
	uint32_t            length_diff;
	int                 eof, enough;
	int                 error;

	/* pitty we need a buffer to read in directory info */
	buffer = malloc(DIRREAD_BUFFER_SIZE);
	if (!buffer) return ENOMEM;

	/* initialise directory reading uio structure and start at the start of the directory */
	bzero(&dir_uio, sizeof(struct uio));
	dir_uio.uio_rw        = UIO_WRITE;
	pos = dir_part_offset = 0;

	chosen_length = 0;
	chosen_offset = 0;
	chosen_length_diff = UINT_MAX;

	eof = (pos == dir_node->stat.st_size);
	while (!eof) {
		/* read in piece of directory */
		dir_iovec.iov_base  = buffer;
		dir_iovec.iov_len   = DIRREAD_BUFFER_SIZE;
		dir_uio.uio_iovcnt  = 1;
		dir_uio.uio_iov     = &dir_iovec;

		/* from offset `pos' for at most buffer size */
		dir_uio.uio_offset  = dir_part_offset = pos;
		dir_part_size       = MIN(DIRREAD_BUFFER_SIZE, dir_node->stat.st_size - pos);
		dir_uio.uio_resid   = dir_part_size;
		/* UDF_CACHEFLAG_METADATA */
		error = udf_read_file_part_uio(dir_node, "directory part", UDF_CACHEFLAG_METADATA, &dir_uio);
		if (error) {
			printf("Error while reading directory index file");
			free(buffer);
			return EIO;
		};

		/* process this chunk */
		do {
			enough = (pos - dir_part_offset + UDF_FID_SIZE <= dir_part_size);
			if (!enough) break;

			dir_fid = (struct fileid_desc *) (buffer + (pos - dir_part_offset));
			dir_fid_name = dir_fid->data + udf_rw16(dir_fid->l_iu);
			entry_length = udf_calc_tag_malloc_size((union dscrptr *) dir_fid);

			enough = (pos - dir_part_offset + entry_length <= dir_part_size);
			if (!enough) break;

			/* check this fid */
			if ((dir_fid->file_char & UDF_FILE_CHAR_DEL) && (entry_length >= fid_len)) {
				length_diff = entry_length - fid_len;
				if (length_diff < chosen_length_diff) {
					/* UDF 2.3.4.2+3 specifies rules for iu length */
					if ((length_diff == 0) || (length_diff >= 32)) {
						chosen_length_diff = length_diff;
						chosen_offset      = pos;
						chosen_length      = entry_length;
					};
				};
			};
			pos += entry_length;
		} while (enough);
		eof = (pos >= dir_node->stat.st_size);	/* sanity */
	}; /* eof */

	if (chosen_length == 0) {
		/* append */
		chosen_offset = pos;
		chosen_length = fid_len;
	};

	if (chosen_length > fid_len) {
		/* padding */
		padding = chosen_length - fid_len;
		fid->tag.desc_crc_len = udf_rw16(chosen_length - UDF_DESC_TAG_LENGTH);
		fid->l_iu = udf_rw16(padding);
		memmove(fid->data + padding, fid->data, fid->l_fi);
		bzero(fid->data, padding-1);
		fid_len = chosen_length;
	};

	udf_node->dir_offset   = chosen_offset;
	udf_node->dir_entrylen = chosen_length;

	udf_validate_tag_sums((union dscrptr *) fid);
	error = udf_writeout_fid_info(dir_node, udf_node, fid);

	if (buffer) free(buffer);
	return error;
}


/* create a file in the given directory with the given name and attributes */
/* use udf's file_char and udf'd filetype */
int udf_create_directory_entry(struct udf_node *dir_node, char *componentname, int filetype, int filechar, struct udf_node *refering, struct udf_node **new_node) {
	struct udf_allocentry *alloc_entry;
	struct udf_node       *udf_node;
	struct charspec        osta_charspec;
	struct fileid_desc    *fid;
	uint32_t     lb_num, lb_size;
	uint16_t     vpart_num, descr_ver, len;
	int          error;

	assert(dir_node);
	assert(componentname);
	assert(dir_node->udf_log_vol);
	lb_size   = udf_rw32(dir_node->udf_log_vol->log_vol->lb_size);
	descr_ver = udf_rw16(dir_node->udf_log_vol->log_vol->tag.descriptor_ver);

	/* lookup if it allready exists (sanity... not nessisary in-kernel) */
	error = udf_lookup(dir_node, new_node, componentname);
	if (!error || *new_node) {
		/* it existed! allready there */
		return 0;
	};

	if (!refering) {
		/* get ourselves an empty node and space to record file descriptor in */
		error = udf_init_udf_node(dir_node->mountpoint, dir_node->udf_log_vol, "New direntry", &udf_node);
		if (error) return error;

		error = udf_allocate_udf_node_on_disc(udf_node, filetype);
		if (error) {
			udf_dispose_udf_node(udf_node);
			return error;
		};
	} else {
		udf_node    = refering;
	};
	alloc_entry = TAILQ_FIRST(&udf_node->dscr_allocs);
	vpart_num   = alloc_entry->vpart_num;
	lb_num      = alloc_entry->lb_num;

	/* build up new directory entry */
	fid = calloc(1, lb_size);
	if (!fid) return ENOMEM;

	udf_osta_charset(&osta_charspec);
	udf_init_desc_tag(&fid->tag, TAGID_FID, descr_ver, 1);			/* serial number    */

	if (filechar & UDF_FILE_CHAR_PAR) {
		/* parent or `..' is not allowed to have a name length ... wierd but ok */
		fid->l_fi = 0;
	} else {
		unix_to_udf_name(fid->data, componentname, &fid->l_fi, &osta_charspec);
	};
	fid->file_version_num = udf_rw16(1);					/* new file/dir; version starts at 1 */
	fid->file_char        = filechar;					/* what is it                        */
	fid->l_iu             = udf_rw32(0);					/* no impl. use                      */
	fid->icb.len          = udf_rw32(lb_size);				/* fill in location                  */
	fid->icb.loc.part_num = udf_rw16(vpart_num);
	fid->icb.loc.lb_num   = udf_rw32(lb_num);

	/* fill in lower 32 bits of unique ID (UDF 3/3.2.2.1) in the impl use part of the FID's long_ad */
	fid->icb.impl.im_used.unique_id = udf_rw32(((udf_node->unique_id << 32) >> 32));
	memcpy(udf_node->dirent.d_name, componentname, strlen(componentname)+1);

	/* calculate minimum size needed for directory entry */
	len = UDF_FID_SIZE + fid->l_fi;
	len = (len + 3) & ~3;
	fid->tag.desc_crc_len = udf_rw16(len - UDF_DESC_TAG_LENGTH);

	error = udf_insert_fid_info(dir_node, udf_node, fid, len);
	if (error) printf("UDF: help! fid insertion failed\n");
	free(fid);	/* Ahum... easily forgotten here */

	if (udf_node) {
		/* only insert file in hashlist if its not an explicit reference */
		if (!refering) {
			udf_insert_node_in_hash(udf_node);
		} else {
			refering->link_cnt++;
			udf_node_mark_dirty(refering);
		};
		udf_node_mark_dirty(udf_node);
	};

	*new_node = udf_node;
	return error;
}


/* VOP_CREATE */
int udf_create_file(struct udf_node *dir_node, char *componentname, struct stat *stat, struct udf_node **new_node) {
	struct udf_node *udf_node;
	int error;

	if (!dir_node) return EINVAL;
	error = udf_create_directory_entry(dir_node, componentname, UDF_ICB_FILETYPE_RANDOMACCESS, 0, NULL, new_node);
	if (*new_node) {
		udf_node = *new_node;
		udf_node->stat = *stat;
		udf_node->stat.st_size    = 0;
		udf_node->stat.st_blksize = udf_rw32(dir_node->udf_log_vol->log_vol->lb_size);
		udf_node->stat.st_blocks  = 0;

		udf_node->udf_log_vol->num_files++;

		udf_node_mark_dirty(udf_node);
	};
	return error;
}


/* VOP_MKDIR */
int udf_create_directory(struct udf_node *dir_node, char *componentname, struct stat *stat, struct udf_node **new_node) {
	struct udf_node *udf_node, *dummy_node;
	int error;

	if (!dir_node) return EINVAL;
	error = udf_create_directory_entry(dir_node, componentname, UDF_ICB_FILETYPE_DIRECTORY, UDF_FILE_CHAR_DIR, NULL, new_node);
	if (*new_node) {
		udf_node = *new_node;
		udf_node->stat = *stat;
		udf_node->stat.st_size    = 0;
		udf_node->stat.st_blksize = udf_rw32(dir_node->udf_log_vol->log_vol->lb_size);
		udf_node->stat.st_blocks  = 0;

		udf_node->udf_log_vol->num_directories++;

		udf_node_mark_dirty(udf_node);

		/* create `..' directory entry */
		error = udf_create_directory_entry(udf_node, "..", UDF_ICB_FILETYPE_DIRECTORY, UDF_FILE_CHAR_DIR | UDF_FILE_CHAR_PAR, dir_node, &dummy_node);
		assert(error == 0);

	};
	return error;
}



/* really deletes all space referenced to this udf node including descriptor spaces */
int udf_delete_udf_node(struct udf_node *udf_node) {
	struct udf_allocentry *alloc_entry;

	udf_node->link_cnt--;
	udf_node_mark_dirty(udf_node);
	if (udf_node->link_cnt > 0) return 0;

	/* trunc node */
	udf_truncate_udf_node(udf_node, (uint64_t) 0);	/* get rid of file contents	*/
	udf_sync_udf_node(udf_node, "deleting");	/* just in case			*/

	if (udf_node->dirty) {
		udf_node_unmark_dirty(udf_node);	/* forget changes		*/
		udf_sync_udf_node(udf_node, "deleting");
	};

	if (udf_node->udf_log_vol) {
		/* now follow all the descriptor allocs and free them */
		TAILQ_FOREACH(alloc_entry, &udf_node->dscr_allocs, next_alloc) {
			if (alloc_entry->flags != UDF_SPACE_FREE) {
				udf_partmap_free_lbs(udf_node->udf_log_vol, alloc_entry->vpart_num, alloc_entry->lb_num, alloc_entry->len);
			};
		};
	};

	return 0;
}


/* NOTE: Dont use the EXTENT erased part; its for non sequential WORM only */
/*       UDF 2.3.10.1, ECMA 4/48.14.1.1 */
/*       fid->icb.impl.im_used.flags = udf_rw16(UDF_ADIMP_FLAGS_EXTENT_ERASED); */

static int udf_remove_file_entry(struct udf_node *dir_node, struct udf_node *udf_node, char *componentname) {
	struct fileid_desc *fid;
	struct udf_node    *instance_node;
	uint32_t            lb_size;
	int                 error;

	assert(dir_node);
	assert(udf_node);
	assert(udf_node->udf_log_vol);
	lb_size     = udf_rw32(udf_node->udf_log_vol->log_vol->lb_size);

	/* use VOP_LOOKUP to get an de-anonymised udf_node for this instance; all other info preserved */
	error = udf_lookup(dir_node, &instance_node, componentname);
	assert(error == 0);
	assert(instance_node == udf_node);

	if (instance_node->dir_offset == 0) {
		printf("Asked to remove `..' parent directory; not allowed!\n");
		return ENOENT;
	};

	/* get a fid to read/write into */
	fid = calloc(1, lb_size);
	if (!fid) return ENOMEM;

	/* mark udf_node entry `deleted' */
	udf_readin_fid_info(dir_node, udf_node, fid, lb_size);
	fid->file_char |= UDF_FILE_CHAR_DEL;
	bzero(&fid->icb, sizeof(struct long_ad));
	udf_validate_tag_sums((union dscrptr *) fid);
	udf_writeout_fid_info(dir_node, udf_node, fid);

	free(fid);

	/* delete contents and administration */
	udf_delete_udf_node(udf_node);
	udf_dispose_udf_node(udf_node);

	return 0;
}


static int udf_count_direntries(struct udf_node *dir_node, int count_dotdot, uint32_t *dir_entries) {
	struct fileid_desc *dir_fid;
	struct uio          dir_uio;
	struct iovec        dir_iovec;
	uint8_t            *buffer, *dir_fid_name;
	uint32_t            pos, dir_part_offset, dir_part_size, entry_length;
	uint32_t            chosen_length, chosen_offset, chosen_length_diff;
	int                 eof, enough;
	int                 error;

	*dir_entries = 0;

	/* pitty we need a buffer to read in directory info */
	buffer = malloc(DIRREAD_BUFFER_SIZE);
	if (!buffer) return ENOMEM;

	/* initialise directory reading uio structure and start at the start of the directory */
	bzero(&dir_uio, sizeof(struct uio));
	dir_uio.uio_rw        = UIO_WRITE;
	pos = dir_part_offset = 0;

	chosen_length = 0;
	chosen_offset = 0;
	chosen_length_diff = UINT_MAX;

	eof = (pos == dir_node->stat.st_size);
	while (!eof) {
		/* read in piece of directory */
		dir_iovec.iov_base  = buffer;
		dir_iovec.iov_len   = DIRREAD_BUFFER_SIZE;
		dir_uio.uio_iovcnt  = 1;
		dir_uio.uio_iov     = &dir_iovec;

		/* from offset `pos' for at most buffer size */
		dir_uio.uio_offset  = dir_part_offset = pos;
		dir_part_size       = MIN(DIRREAD_BUFFER_SIZE, dir_node->stat.st_size - pos);
		dir_uio.uio_resid   = dir_part_size;
		/* UDF_CACHEFLAG_METADATA */
		error = udf_read_file_part_uio(dir_node, "directory part", UDF_CACHEFLAG_METADATA, &dir_uio);
		if (error) {
			printf("Error while reading directory index file");
			free(buffer);
			return error;
		};

		/* process this chunk */
		do {
			enough = (pos - dir_part_offset + UDF_FID_SIZE <= dir_part_size);
			if (!enough) break;

			dir_fid = (struct fileid_desc *) (buffer + (pos - dir_part_offset));
			dir_fid_name = dir_fid->data + udf_rw16(dir_fid->l_iu);
			entry_length = udf_calc_tag_malloc_size((union dscrptr *) dir_fid);

			enough = (pos - dir_part_offset + entry_length <= dir_part_size);
			if (!enough) break;

			if (!(dir_fid->file_char & UDF_FILE_CHAR_DEL)) {
				if (dir_fid->file_char & UDF_FILE_CHAR_PAR) {
					if (count_dotdot) *dir_entries = *dir_entries + 1;
				} else {
					*dir_entries = *dir_entries + 1;
				};
			};
			pos += entry_length;
		} while (enough);
		eof = (pos >= dir_node->stat.st_size);	/* sanity */
	}; /* eof */

	if (buffer) free(buffer);

	return 0;
}


/* VOP_REMOVE */
int udf_remove_file(struct udf_node *dir_node, struct udf_node *udf_node, char *componentname) {
	int error;

	if (udf_node->udf_filetype == UDF_ICB_FILETYPE_DIRECTORY) {
		/* only remove files with this call */
		return EISDIR;
	};

	error = udf_remove_file_entry(dir_node, udf_node, componentname);
	if (!error) udf_node->udf_log_vol->num_files--;

	return error;
}


/* VOP_RMDIR */
int udf_remove_directory(struct udf_node *dir_node, struct udf_node *udf_node, char *componentname) {
	uint32_t num_nodes;
	int error;

	if (udf_node->udf_filetype != UDF_ICB_FILETYPE_DIRECTORY) {
		/* only remove directories with this call */
		return ENOTDIR;
	};

	error = udf_count_direntries(udf_node, 0, &num_nodes);
	if (error) return error;

	if (num_nodes != 0) return ENOTEMPTY;

	/* remove the entry */
	error = udf_remove_file_entry(dir_node, udf_node, componentname);
	if (!error) {
		dir_node->link_cnt--;
		udf_node_mark_dirty(dir_node);
		udf_node->udf_log_vol->num_directories--;
	} else {
		/* whoah! something went wrong, mark the .. as present again */
		printf("UDF warning: filesystem might by in compromised state\n");
		udf_node->link_cnt++;
	};

	return error;
}


/* end of udf.c */

